# -*- coding: utf-8 -*-
# pylint: disable=C0326
"""
Created on Tue Sep 30 12:31:04 2014

This module contains the definition of the P wave abstraction pattern.

@author: T. Teijeiro
"""

import construe.knowledge.observables as o
import construe.utils.signal_processing.Douglas_Peucker as DP
import construe.knowledge.constants as C
import construe.acquisition.signal_buffer as sig_buf
from construe.utils.units_helper import (samples2msec as sp2ms,
                                            phys2digital as ph2dg,
                                            digital2phys as dg2ph)
from construe.model import verify, Interval as Iv
from construe.model.automata import (PatternAutomata, ABSTRACTED,
                                        ENVIRONMENT, BASIC_TCONST)
import numpy as np
import pickle
import sklearn.preprocessing as preproc

####################################################
### Definition of the P Wave abstraction pattern ###
####################################################

#Auxiliary functions
def _delimit_p(signal, lead, es_lim, ls_lim, ee_lim):
    """
    Performs the delimitation of a P wave in a signal fragment. If a waveform
    compatible with a P wave cannot be found, returns None, else return an
    Interval within signal length.
    """
    #shape simplification (ignoring the environment signal)
    delta = ph2dg(0.02)
    points = DP.arrayRDP(signal[int(es_lim):], delta, 6) + int(es_lim)
    #If no relevant disturbances are detected, there is no a P wave
    if len(points) == 2:
        return None
    #Now we look for the shorter limits that satisfy the P-Wave classifier.
    cand = None
    i = next(k for k in xrange(len(points)-1, -1, -1) if points[k] <= ls_lim)
    while i >= 0:
        j = next(k for k in xrange(i+1, len(points)) if points[k] >= ee_lim)
        while j < len(points):
            sigfr = signal[points[i]:points[j]+1]
            #We consider a good P wave environment if the signal has no
            #amplitude variations
            beg = int(max(0, points[i]-C.PWAVE_ENV))
            plainenv = not np.any(signal[beg:points[i]+1]-signal[beg])
            #The minimum threshold varies with the environment quality
            ampthres = C.PWAVE_MIN_AMP if not plainenv else delta
            if (seems_pwave(sigfr, lead) and np.ptp(sigfr) >= ampthres):
                cand = (points[i], points[j])
                break
            j += 1
        if cand is not None:
            break
        i -= 1
    return None if cand is None else Iv(int(cand[0]-es_lim),
                                        int(cand[1]-es_lim))

def delineate_pwave(es_lim, ls_lim, ee_lim, le_lim, pwave):
    """
    Performs the delineation of a possible P-wave contained in the given limits.

    Parameters
    ----------
    es_lim:
        earliest possible time for the beginning of the P-wave.
    ls_lim:
        latest possible time for the beginning of the P-wave.
    ee_lim:
        earliest possible time for the ending of the P-wave.
    le_lim:
        latest possible time for the ending of the P-wave.
    pwave:
        PWave instance, **which is modified** to establish the amplitude in all
        those leads in which the identification was correct.

    Returns
    -------
    out:
        Interval with the delineation of the p-wave, relative to *es_lim*. If
        a p-wave cannot be delineated, returns None.
    """
    start = finish = None
    for lead in (l for l in C.PWAVE_LEADS if sig_buf.is_available(l)):
        #We take some environment signal for the P wave.
        beg = int(es_lim-C.PWAVE_ENV)
        beg = 0 if beg < 0 else beg
        sig = sig_buf.get_signal_fragment(beg, le_lim, lead=lead)[0]
        endpoints = _delimit_p(sig, lead, es_lim-beg, ls_lim-beg, ee_lim-beg)
        if endpoints is None:
            continue
        elif start is None:
            start, finish = endpoints.start, endpoints.end
            if finish > start:
                pwave.amplitude[lead] = np.ptp(sig[start:finish+1])
        else:
            if abs(start - endpoints.start) < C.TMARGIN:
                start = min(start, endpoints.start)
            if abs(finish - endpoints.end) < C.TMARGIN:
                finish = max(finish, endpoints.end)
            if finish > start:
                pwave.amplitude[lead] = np.ptp(sig[start:finish+1])
    return None if start is None else Iv(start, finish)


def _p_qrs_tconst(pattern, qrs):
    """
    Adds the temporal constraints wrt the qrs environment observation.
    """
    BASIC_TCONST(pattern, qrs)
    pwave = pattern.hypothesis
    #Temporal constraints
    tnet = pattern.last_tnet
    #P wave duration constraint
    tnet.add_constraint(pwave.start, pwave.end, C.PW_DURATION)
    #Relations between P wave and QRS complex
    tnet.add_constraint(pwave.end, qrs.start, C.PQ_INTERVAL)
    tnet.add_constraint(pwave.start, qrs.start, C.PR_INTERVAL)

def _p_defl_tconst(pattern, defl):
    """
    Temporal constraints definition wrt the abstracted energy interval.
    """
    BASIC_TCONST(pattern, defl)
    pwave = pattern.hypothesis
    qrs = None
    if pattern.evidence[o.QRS]:
        qrs = pattern.evidence[o.QRS][0]
    #Temporal constraints
    tnet = pattern.last_tnet
    #P wave duration constraint
    tnet.add_constraint(pwave.start, pwave.end, C.PW_DURATION)
    #Constraints with the energy interval
    tnet.add_constraint(defl.start, defl.end, C.PW_DEF_DUR)
    tnet.add_constraint(pwave.start, defl.start, Iv(-C.PW_DEF_OVER,
                                                    C.PW_DEF_OVER))
    tnet.add_constraint(pwave.end, defl.end,
                        Iv(-C.PW_DEF_OVER, C.PW_DEF_OVER))
    tnet.set_before(defl.start, pwave.end)
    tnet.set_before(pwave.start, defl.end)
    if qrs is not None:
        tnet.add_constraint(defl.start, qrs.start, C.PR_DEF_SEP)
        tnet.add_constraint(defl.end, qrs.start, C.PQ_DEF_SEP)
        tnet.set_before(defl.end, qrs.start)

def _p_gconst(pattern, defl):
    """
    General constraints of the P Wave abstraction pattern, once all the
    evidence has been observed.
    """
    pwave = pattern.hypothesis
    if ((defl is not None and defl.earlystart != defl.latestart)
                                               or not pattern.evidence[o.QRS]):
        return
    qrs = pattern.evidence[o.QRS][0]
    beg = pwave.earlystart
    if beg < 0:
        beg = 0
    #We try the delineation of the P-Wave
    endpoints = delineate_pwave(beg, int(pwave.latestart),
                                int(pwave.earlyend), int(pwave.lateend), pwave)
    verify(endpoints is not None)
    #Now we obtain the amplitudes, and we ensure the QRS amplitude is at
    #least twice the P Wave amplitude in each lead
    pwave.start.value = Iv(beg + endpoints.start, beg + endpoints.start)
    pwave.end.value = Iv(beg + endpoints.end, beg + endpoints.end)
    for lead in pwave.amplitude:
        verify(pwave.amplitude[lead] <= C.PWAVE_AMP[lead])
        verify(lead not in qrs.shape or
               pwave.amplitude[lead] < qrs.shape[lead].amplitude)


#########################
## Automata definition ##
#########################

PWAVE_PATTERN = PatternAutomata()
PWAVE_PATTERN.name = 'P Wave'
PWAVE_PATTERN.Hypothesis = o.PWave
PWAVE_PATTERN.add_transition(0, 1, o.QRS, ENVIRONMENT, _p_qrs_tconst)
PWAVE_PATTERN.add_transition(1, 2, o.Deflection, ABSTRACTED, _p_defl_tconst,
                             _p_gconst)
#PWAVE_PATTERN.add_transition(1, 2, gconst=_p_gconst)
PWAVE_PATTERN.final_states.add(2)
PWAVE_PATTERN.freeze()


##################################################
### Statistical knowledge stored as histograms ###
##################################################

def _check_histogram(hist, value):
    """
    Obtains a score of a value according to an histogram, between 0.0 and 1.0
    """
    i = 0
    while i < len(hist[1]) and value > hist[1][i]:
        i += 1
    if i == 0 or i == len(hist[1]):
        return 0.0
    else:
        return hist[0][i-1]


#Static definition of the PR histogram
_PR_HIST = (np.array(
             [0.20300752,  0.22932331,  0.32330827,  0.57142857,  0.66541353,
              0.92857143,  0.84962406,  0.84210526,  0.87969925,  0.93984963,
              1.        ,  0.77443609,  0.63533835,  0.51879699,  0.43609023,
              0.48496241,  0.28947368,  0.29699248,  0.29323308,  0.27443609,
              0.2556391 ,  0.19172932,  0.20676692,  0.16165414,  0.17293233,
              0.17293233,  0.10150376,  0.07518797,  0.07142857,  0.04323308,
              0.01503759,  0.0112782 ,  0.01503759,  0.02255639,  0.0075188 ,
              0.0112782 ,  0.0112782 ,  0.01503759,  0.01503759,  0.01691729,
              0.01879699,  0.0037594 ,  0.03383459,  0.0075188 ,  0.02631579,
              0.03759398,  0.01879699,  0.04887218,  0.0112782 ,  0.04511278]),
            np.array(
             [20. ,   23.6,   27.2,   30.8,   34.4,   38. ,   41.6,   45.2,
              48.8,   52.4,   56. ,   59.6,   63.2,   66.8,   70.4,   74. ,
              77.6,   81.2,   84.8,   88.4,   92. ,   95.6,   99.2,  102.8,
              106.4,  110. ,  113.6,  117.2,  120.8,  124.4,  128. ,  131.6,
              135.2,  138.8,  142.4,  146. ,  149.6,  153.2,  156.8,  160.4,
              164. ,  167.6,  171.2,  174.8,  178.4,  182. ,  185.6,  189.2,
              192.8,  196.4,  200. ]))

#Static definition of the PQ histogram
_PQ_HIST = (np.array(
             [0.07231405,  0.15495868,  0.4338843 ,  0.58677686,  0.92768595,
              1.        ,  0.94214876,  0.6714876 ,  0.52066116,  0.39669421,
              0.23760331,  0.1446281 ,  0.14876033,  0.10330579,  0.08057851,
              0.03512397,  0.02479339,  0.01239669,  0.01859504,  0.0268595 ,
              0.01239669,  0.03099174,  0.00826446,  0.00413223,  0.00413223]),
        np.array([ 80.,  112.,  124.,  136.,  148.,  160.,  172.,  184.,  196.,
                  208.,  220.,  232.,  244.,  256.,  268.,  280.,  292.,  304.,
                  316.,  328.,  340.,  352.,  364.,  376.,  388.,  400.]))

#Static definition of the P wave duration histogram
_PDUR_HIST = (np.array(
             [0.00410678,  0.0164271 ,  0.05749487,  0.13552361,  0.24640657,
              0.40246407,  0.6899384 ,  0.78234086,  1.        ,  0.87474333,
              0.69609856,  0.58316222,  0.33880903,  0.23819302,  0.17453799,
              0.12936345,  0.07597536,  0.05338809,  0.01848049,  0.01232033,
              0.0164271 ,  0.00410678,  0.00410678,  0.        ,  0.00410678]),
       np.array([44.  ,   51.84,   59.68,   67.52,   75.36,   83.2 ,   91.04,
                 98.88,  106.72,  114.56,  122.4 ,  130.24,  138.08,  145.92,
                 153.76,  161.6 ,  169.44,  177.28,  185.12,  192.96,  200.8 ,
                 208.64,  216.48,  224.32,  232.16,  240.  ]))


#################################
### P Wave signal classifiers ###
#################################

#We have one classifier for the limb leads, and other for precordial leads.
#The classifiers are one-class SVM trained with 73 clean records from the QT
#database. The features used are the 4 points in coordinates (X,Y) obtained
#from a RDP simplification of the signal delimiting the P wave, using 5 points
#for the simplification and assuming the first point is always (0,0). The
#units of the coordinates are in msec and mV.

def _scale_sample(signal, lead):
    if not lead in _CL_MAP:
        raise ValueError('Invalid lead.')
    scaler = _SCALERS[_CL_MAP[lead]]
    #The signal is converted to physical units, and the first point is (0,0)
    sig = dg2ph(signal-signal[0])
    #The features are the 4 points better representing the signal shape.
    points = DP.arrayRDP(sig, 0.001, 5)[1:]
    if len(points) < 4:
        raise ValueError('Not enough points after path simplification')
    sample = np.concatenate((sp2ms(points), sig[points])).reshape(1, -1)
    return scaler.transform(sample)

def seems_pwave(signal, lead):
    """
    Checks if a signal fragment looks like a P wave. It is assumed the signal
    is in raw signal units at the record sampling frequency.

    Parameters
    ----------
    signal:
        Raw signal array with the delimited P wave.
    lead:
        Lead where the signal has been obtained. It must be a limb
        or a precordial lead.
    """
    if not lead in _CL_MAP:
        raise ValueError('Invalid lead.')
    classifier = _CLASSIFIERS[_CL_MAP[lead]]
    try:
        sample = _scale_sample(signal, lead)
    except ValueError:
        return False
    return classifier.predict(sample)[0] == 1

def pwave_distance(signal, lead):
    """
    Obtains a distance measure of a signal fragment to a recognized P wave
    morphology.
    """
    if not lead in _CL_MAP:
        raise ValueError('Invalid lead.')
    classifier = _CLASSIFIERS[_CL_MAP[lead]]
    try:
        sample = _scale_sample(signal, lead)
    except ValueError:
        #Unknown values are located in the decision boundary
        return 0.0
    return classifier.decision_function(sample)[0][0]

#Mapping for each lead with the corresponding classifier.
_CL_MAP = {sig_buf.Leads.UNIQUE: 0, sig_buf.Leads.MLI:   0,
           sig_buf.Leads.MLII:   0, sig_buf.Leads.MLIII: 0,
           sig_buf.Leads.V1:     1, sig_buf.Leads.V2:    1,
           sig_buf.Leads.V3:     1, sig_buf.Leads.V4:    1,
           sig_buf.Leads.V5:     1, sig_buf.Leads.V6:    1}

#Each classifier has a scaler to preprocess the feature set.

_SCALERS = [preproc.StandardScaler(), preproc.StandardScaler()]
_SCALERS[0].mean_ = np.array(
        [3.64100119e+01,   6.27794994e+01,   8.74159714e+01,   1.18181168e+02,
         6.97318236e-02,   1.17550656e-01,   8.52205006e-02,  -1.42669845e-02])
_SCALERS[1].mean_ = np.array(
        [3.52004505e+01,   6.13603604e+01,   8.69031532e+01,   1.17867117e+02,
         5.26295045e-02,   8.14245495e-02,   3.52759009e-02,  -1.85416667e-02])
_std0 = np.array([18.28456548,  21.80939775,  25.19837448,  27.46293336,
                  0.06650826,   0.09788993,   0.09393059,   0.04712977])
_std1 = np.array([16.8934232 ,  19.45625391,  23.22422215,  26.39513332,
                  0.05968984,   0.07814591,   0.08662095,   0.05071906])
#Fix for scikit-learn > 0.16
try:
    _SCALERS[0].std_ = _std0
    _SCALERS[1].std_ = _std1
except AttributeError:
    _SCALERS[0].scale_ = _std0
    _SCALERS[1].scale_ = _std1

#Trained classifiers. These classifiers were serialized using the pickle module
# into the following two (long) strings.
#They are instances of sklearn.svm.OneClassSVM.

_LIMB_STR = ('ccopy_reg\n_reconstructor\np0\n(csklearn.svm.classes\nOneClassSV'
"M\np1\nc__builtin__\nobject\np2\nNtp3\nRp4\n(dp5\nS'_impl'\np6\nS'one_"
"class'\np7\nsS'kernel'\np8\nS'rbf'\np9\nsS'verbose'\np10\nI00\nsS'pro"
"bability'\np11\nI00\nsS'support_'\np12\ncnumpy.core.multiarray\n_r"
"econstruct\np13\n(cnumpy\nndarray\np14\n(I0\ntp15\nS'b'\np16\ntp17\nRp"
"18\n(I1\n(I186\ntp19\ncnumpy\ndtype\np20\n(S'i4'\np21\nI0\nI1\ntp22\nRp2"
"3\n(I3\nS'<'\np24\nNNNI-1\nI-1\nI0\ntp25\nbI00\nS'\\x0e\\x00\\x00\\x00\\x1"
'5\\x00\\x00\\x00Z\\x00\\x00\\x00[\\x00\\x00\\x00\\\\\\x00\\x00\\x00]\\x00\\x'
'00\\x00^\\x00\\x00\\x00`\\x00\\x00\\x00h\\x00\\x00\\x00i\\x00\\x00\\x00l\\'
'x00\\x00\\x00n\\x00\\x00\\x00o\\x00\\x00\\x00q\\x00\\x00\\x00r\\x00\\x00\\'
'x00v\\x00\\x00\\x00z\\x00\\x00\\x00\\x80\\x00\\x00\\x00\\x82\\x00\\x00\\x0'
'0\\x83\\x00\\x00\\x00\\x84\\x00\\x00\\x00\\x85\\x00\\x00\\x00\\x8a\\x00\\x0'
'0\\x00\\x8d\\x00\\x00\\x00\\x8e\\x00\\x00\\x00\\x91\\x00\\x00\\x00\\x94\\x0'
'0\\x00\\x00\\x9d\\x00\\x00\\x00\\xb2\\x00\\x00\\x00\\xce\\x00\\x00\\x00?\\x'
'01\\x00\\x00A\\x01\\x00\\x00C\\x01\\x00\\x00D\\x01\\x00\\x00E\\x01\\x00\\x'
'00F\\x01\\x00\\x00J\\x01\\x00\\x00K\\x01\\x00\\x00P\\x01\\x00\\x00X\\x01\\'
'x00\\x00Z\\x01\\x00\\x00]\\x01\\x00\\x00b\\x01\\x00\\x00g\\x01\\x00\\x00\\'
'x92\\x01\\x00\\x00\\xf4\\x01\\x00\\x00\\xf5\\x01\\x00\\x00\\xf9\\x01\\x00\\'
'x00\\xfa\\x01\\x00\\x00\\xfe\\x01\\x00\\x00\\xff\\x01\\x00\\x00\\x00\\x02\\'
'x00\\x00\\x01\\x02\\x00\\x00\\x06\\x02\\x00\\x00@\\x02\\x00\\x00K\\x02\\x0'
'0\\x00b\\x02\\x00\\x00\\x91\\x02\\x00\\x00\\xd9\\x02\\x00\\x00\\xe4\\x02\\x'
'00\\x00\\xf8\\x02\\x00\\x004\\x03\\x00\\x00\\x8c\\x03\\x00\\x00\\x8f\\x03\\'
'x00\\x00\\x93\\x03\\x00\\x00\\x96\\x03\\x00\\x00\\x9c\\x03\\x00\\x00\\xa2\\'
'x03\\x00\\x00\\xa6\\x03\\x00\\x00\\xc6\\x03\\x00\\x00\\xdc\\x03\\x00\\x00\\'
'xe3\\x03\\x00\\x00\\xe4\\x03\\x00\\x00\\xe8\\x03\\x00\\x00\\xe9\\x03\\x00\\'
'x00\\xeb\\x03\\x00\\x00\\xee\\x03\\x00\\x00\\xef\\x03\\x00\\x00\\xf7\\x03\\'
'x00\\x00\\xfb\\x03\\x00\\x00\\xff\\x03\\x00\\x00\\x03\\x04\\x00\\x00O\\x04'
'\\x00\\x00W\\x04\\x00\\x00[\\x04\\x00\\x00_\\x04\\x00\\x00a\\x04\\x00\\x00'
'b\\x04\\x00\\x00c\\x04\\x00\\x00e\\x04\\x00\\x00g\\x04\\x00\\x00h\\x04\\x0'
'0\\x00i\\x04\\x00\\x00k\\x04\\x00\\x00l\\x04\\x00\\x00m\\x04\\x00\\x00q\\x'
'04\\x00\\x00r\\x04\\x00\\x00t\\x04\\x00\\x00v\\x04\\x00\\x00y\\x04\\x00\\x'
'00z\\x04\\x00\\x00\\x80\\x04\\x00\\x00\\x87\\x04\\x00\\x00\\x88\\x04\\x00\\'
'x00\\x89\\x04\\x00\\x00\\x8a\\x04\\x00\\x00\\x8b\\x04\\x00\\x00\\x8d\\x04\\'
'x00\\x00\\x8f\\x04\\x00\\x00\\x92\\x04\\x00\\x00\\x93\\x04\\x00\\x00\\x94\\'
'x04\\x00\\x00\\x96\\x04\\x00\\x00\\x97\\x04\\x00\\x00\\x98\\x04\\x00\\x00\\'
'x99\\x04\\x00\\x00\\x9b\\x04\\x00\\x00\\x9e\\x04\\x00\\x00\\x9f\\x04\\x00\\'
'x00\\xa7\\x04\\x00\\x00\\xb2\\x04\\x00\\x00\\xb9\\x04\\x00\\x00\\xbc\\x04\\'
'x00\\x00\\xbd\\x04\\x00\\x00\\xc7\\x04\\x00\\x00\\xcb\\x04\\x00\\x00\\xcc\\'
'x04\\x00\\x00\\xce\\x04\\x00\\x00\\xcf\\x04\\x00\\x00\\xd0\\x04\\x00\\x00\\'
'xd6\\x04\\x00\\x00\\xd7\\x04\\x00\\x00\\xda\\x04\\x00\\x00\\xdf\\x04\\x00\\'
'x00\\xe3\\x04\\x00\\x00\\xe5\\x04\\x00\\x00\\xe6\\x04\\x00\\x00\\xeb\\x04\\'
'x00\\x00-\\x05\\x00\\x002\\x05\\x00\\x00:\\x05\\x00\\x00=\\x05\\x00\\x00E'
'\\x05\\x00\\x00J\\x05\\x00\\x00f\\x05\\x00\\x00g\\x05\\x00\\x00j\\x05\\x00'
'\\x00q\\x05\\x00\\x00z\\x05\\x00\\x00}\\x05\\x00\\x00\\x87\\x05\\x00\\x00\\'
'x8b\\x05\\x00\\x00\\x8d\\x05\\x00\\x00\\x8e\\x05\\x00\\x00\\x8f\\x05\\x00\\'
'x00\\x90\\x05\\x00\\x00\\x93\\x05\\x00\\x00\\x94\\x05\\x00\\x00\\x95\\x05\\'
'x00\\x00\\x97\\x05\\x00\\x00\\x98\\x05\\x00\\x00\\x99\\x05\\x00\\x00\\x9d\\'
'x05\\x00\\x00\\xa2\\x05\\x00\\x00\\xa4\\x05\\x00\\x00\\xa8\\x05\\x00\\x00\\'
'xab\\x05\\x00\\x00\\xae\\x05\\x00\\x00\\xb2\\x05\\x00\\x00\\xb4\\x05\\x00\\'
'x00\\xb8\\x05\\x00\\x00\\xbb\\x05\\x00\\x00\\xbc\\x05\\x00\\x00\\xdf\\x05\\'
'x00\\x00\\xe0\\x05\\x00\\x00\\xe2\\x05\\x00\\x00\\xe8\\x05\\x00\\x00\\xed\\'
'x05\\x00\\x00\\xee\\x05\\x00\\x00\\xef\\x05\\x00\\x00\\x1c\\x06\\x00\\x00-'
"\\x06\\x00\\x00A\\x06\\x00\\x00E\\x06\\x00\\x00q\\x06\\x00\\x00'\np26\ntp2"
"7\nbsS'dual_coef_'\np28\ng13\n(g14\n(I0\ntp29\ng16\ntp30\nRp31\n(I1\n(I"
"1\nI186\ntp32\ng20\n(S'f8'\np33\nI0\nI1\ntp34\nRp35\n(I3\nS'<'\np36\nNNNI"
"-1\nI-1\nI0\ntp37\nbI00\nS'\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\"
'x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x0'
'0\\x00\\x00\\x00\\xf0?T\\xbbI\\xde\\xa7\\xfb\\xdf?\\x00\\x00\\x00\\x00\\x0'
'0\\x00\\xf0?.\\x90?H\\xbf\\x92\\xea?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\'
'x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\xd7V\\xd4\\x066\\xf5\\xe7?\\x00\\x00\\'
'x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x0'
'0\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\'
'x00\\x00\\x00\\xf0?8\\xba0\\xb5\\x1fL\\xd7?\\x00\\x00\\x00\\x00\\x00\\x00'
'\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\x'
'f0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0'
'?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\'
'x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x0'
'0\\x00\\x00\\x00\\x00\\x00\\xf0?\\x9e</\\xc5\\rK\\xd3?\\x00\\x00\\x00\\x00'
'\\x00\\x00\\xf0?\\x1b\\xd7\\xda\\xff`\\x8c\\xe9?\\x00\\x00\\x00\\x00\\x00\\'
'x00\\xf0?GR\\x1d"\\xed\\xf9i?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x'
'00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00'
'\\x00\\x00\\x00\\x00\\xf0?|\\x9f\\xd4\\xce\\xf9\\xdb\\xd4?\\x00\\x00\\x00\\'
'x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x0'
'0\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\'
'x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x0'
'0\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\'
"x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?]\\'\\x93\\xfd\\xb22\\xd8?\\x"
'00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00'
'\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x'
'00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?E\\xd6T.d'
'\\x93\\xd1?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x'
"00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\xf5\\x9f\\xf2\\x1e\\xc5\\'\\x"
'ec?tU\\x9c\\xb3\\xdb\\x1e\\xe1?\\x11\\n\\xffm?a\\xe1?\\x00\\x00\\x00\\x00'
'\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x'
'00\\x00\\xf0?\\x80\\x94G\\xb7F$\\xd6?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?'
'\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?d\\xd4\\xffA\\xb7\\xc3\\xdb?\\x00\\x00'
'\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x15b\\xe2\\'
'x9f\\x1eS\\xd7?6\\xb7\\xdeF:\\x08\\xb8?\\x00\\x00\\x00\\x00\\x00\\x00\\xf'
'0?\\x86-\\x08\\x07\\xd7\\x07\\xea?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x0'
'0\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\'
'x00\\x00\\x00\\x00\\x00\\xf0?\\x98\\xb4\\xbe\\x0f\\xd5\\x12\\xcf?\\x00\\x0'
'0\\x00\\x00\\x00\\x00\\xf0?\\xc8W&\\n\\x195\\xe2?\\x00\\x00\\x00\\x00\\x00'
'\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x'
'00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\xcd\\x10\\xa7\\xbe\\x9d\\xb7'
'\\xec?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\x'
'f0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0'
'?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\'
'x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x0'
'0\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\xf8Q'
'\\x80\\xf3\\x94\\x15\\xd6?\\xaf\\x9dY\\xeeB\\xb1\\xe3?\\x00\\x00\\x00\\x00'
'\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?k\\xd1\\x0e\\xf9\\xd2\\'
'xe1\\xd3?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x0'
'0\\xf0?\\x0cw\\x0b\\x99\\x14\\x9d\\xd4?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0'
'?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\'
'x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x0'
'0\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\'
'x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x0'
'0\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\'
'x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x0'
'0\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\'
'x00\\x00\\x00\\xf0?\\xa7S\\xcf\\xfc\\x1c\\x81\\xe3?\\x00\\x00\\x00\\x00\\x'
'00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00'
'\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x'
'00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00'
'\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\x'
'f0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0'
'?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\'
'x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x9'
'9I\\x9f\\xc8\\x87\\x13\\xe1?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?[\\x1b\\xf'
'd2\\xa0\\x92\\xdf?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00'
'\\x00\\x00\\xf0??}?\\xe3\\xb5o\\xde?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\'
'x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x0'
'0\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\'
'x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x9bX\\x'
'e2\\xc8\\xae\\xfe\\xe6?p\\xc2cD\\x9c\\\\\\xdf?\\x00\\x00\\x00\\x00\\x00\\x0'
'0\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\xcdk.\\xca\\xd7B\\xee?\\x00\\'
'x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x0'
'0\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\'
'x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x0'
'0\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\'
'x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x0'
'0\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\'
'x00\\x00\\xf0?\\xd7x\\x0cK\\x06\\x1d\\xe8?\\x00\\x00\\x00\\x00\\x00\\x00\\'
'xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?N=\\xb6\\x91p&\\xec?\\x00\\x00\\x'
'00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00'
'\\x00\\x00\\x00\\xf0?\\x93\\xdf\\x0eI\\x91c\\xe2?\\xc2\\xf0<T\\x1b\\x8f\\x'
'c3?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?-Y,u=\\xa7\\xe4?\\x00\\x00\\x00\\x'
'00\\x00\\x00\\xf0?\\xa49\\r#\\xeb\\x1d\\xe6?\\x00\\x00\\x00\\x00\\x00\\x00'
'\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\x'
'f0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0'
"?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?'"
"\np38\ntp39\nbsS'shrinking'\np40\nI01\nsS'class_weight'\np41\nNsS'_g"
"amma'\np42\nF0.125\nsS'probA_'\np43\ng13\n(g14\n(I0\ntp44\ng16\ntp45\nR"
"p46\n(I1\n(I0\ntp47\ng35\nI00\nS''\np48\ntp49\nbsS'_sparse'\np50\nI00\ns"
"S'class_weight_'\np51\ng13\n(g14\n(I0\ntp52\ng16\ntp53\nRp54\n(I1\n(I0"
"\ntp55\ng35\nI00\ng48\ntp56\nbsS'random_state'\np57\nNsS'_sklearn_ve"
"rsion'\np58\nS'0.18'\np59\nsS'tol'\np60\nF0.001\nsS'coef0'\np61\nF0.0"
"\nsS'nu'\np62\nF0.1\nsS'n_support_'\np63\ng13\n(g14\n(I0\ntp64\ng16\ntp"
"65\nRp66\n(I1\n(I2\ntp67\ng23\nI00\nS'\\x88\\xf7\\xac\\x1d\\xa1\\x7f\\x00\\"
"x00'\np68\ntp69\nbsS'shape_fit_'\np70\n(I1678\nI8\ntp71\nsS'C'\np72\nF"
"0.0\nsS'support_vectors_'\np73\ng13\n(g14\n(I0\ntp74\ng16\ntp75\nRp76"
"\n(I1\n(I186\nI8\ntp77\ng35\nI00\nS'\\x9ci B!\\xdc\\xf1\\xbf\\xb1\\x97\\xc"
'b\\xaf\\xe6\\x83\\xf9\\xbf\\xb9OX\\xe6\\xdf#\\xea?\\xd6\\xa7D\\xbf!]\\xf8'
'?\\x89\\xa1E\\xf4|b\\xcd?\\xb3\\x85\\xf2Z\\xe0\\x07\\xf4\\xbf\\xd8\\x1c\\x'
'8c\\xd4&\\x8d\\xf3?\\xd4K\\xba@\\x02$\\xbf\\xbf\\xcc)V\\xd8L\\xf6\\x96\\x'
'bf\\x8a|\\x9b\\xc4q\\xaf\\xe5\\xbf\\xc6p$\\xd3$\\xf6\\xdf?\\xe3nn\\x16\\x'
'f1\\xb9\\xd7\\xbfKN\\x9e\\xbc\\xd7\\xfc\\xd2\\xbf\\x0e\\xdd,\\x00\\xe4Y\\x'
'e9\\xbf\\xa5\\x10\\xf4\\xd5\\x8e.\\x03\\xc0\\xdc_\\x06\\x8f5\\n\\x07\\xc0\\'
'xa3\\xff\\x8d\\xb1%\\xb8\\xec\\xbf\\\\\\xe7\\xf1\\x1d\\xc5\\x0e\\xdb?2\\x84'
'0\\xb3\\x13I\\xc7?\\x18&\\x89{\\x93\\x15\\xee?\\xb7\\xf8e\\x90\\x0c2\\xf9'
'\\xbfyhC\\x1a\\r\\xd1\\xc6?\\xec\\xfe\\xf7\\x00\\x1cg\\x07@4;\\xae\\xb4 \\'
'xdf\\xe3?\\x1b\\x89\\xc0\\xaf\\xa9!\\xc9?\\x0e\\x8fw\\xdc\\x11\\xa7\\xac?'
'K\\x19#\\xcb\\xbc\\x93\\xe3\\xbf\\xea\\x06Iy\\x0f\\x1a\\xe0?\\x1a#\\x88\\n'
'A\\x96\\xf5\\xbf \\x13\\xfdt?i\\xc0\\xbfP\\x12\\x01 \\xfek\\x04@d\\xf1\\x'
'c8E\\xf3\\xdd\\xe4\\xbf\\x9ci B!\\xdc\\xf1\\xbf\\x99\\x03\\xd4\\xa6\\x00\\'
'xe5\\xd3\\xbf@\\x1b\\x80\\x19\\xac\\xfe\\xdc\\xbf?\\x98$\\xbe__\\xf1?-\\\\'
'q\\x0e\\x1e\\xfe\\xf7\\xbf\\x89\\xfe\\r\\xadUG\\xc0?X1l\\xa9\\x06F\\x05@}'
'LV\\x8e\\\\]\\xd5\\xbff\\xd3y\\xab/\\\\\\xf5\\xbf\\x10Udh_b\\xff\\xbf\\xea\\'
'x03\\xba\\x9c\\xde\\xd5\\xd2\\xbf;L\\xb7\\x8cP/\\xe5\\xbf\\x9f=\\x99"\\x1'
'a\\xd6\\x0c\\xc0k]\\xe9\\x94\\x0eH\\xd5\\xbf@\\xb0\\xfc\\xce\\x95\\x9f\\xf'
'e?\\xee\\xc35\\xebie\\x03@1=\\xd3\\x14>\\xdc\\xf8\\xbf\\x99\\x03\\xd4\\xa'
'6\\x00\\xe5\\xd3\\xbf\\xa20\\xe9G\\x8a\\xbc\\xed\\xbfB\\x15\\xb9M\\x04U\\x'
'b4\\xbf-\\\\q\\x0e\\x1e\\xfe\\xf7\\xbfOU\\x19\\x956\\x9f\\x99?P\\x12\\x01 '
'\\xfek\\x04@\\x03\\x03?F\\xdcx\\xe1\\xbf\\x9ci B!\\xdc\\xf1\\xbf\\xe894}'
'\\xea\\x8d\\xeb\\xbf\\xfc\\xa3W\\xe2\\xab\\xf2\\xf3\\xbf4\\xb4\\xdc)2\\xcf'
'\\xcc\\xbf.\\xa7\\xf47lO\\xf3?\\x1f\\xd3\\xf1\\xb0\\x197\\xd5?x\\xad\\x18'
'\\xcf(\\xae\\x08@}LV\\x8e\\\\]\\xd5\\xbf1=\\xd3\\x14>\\xdc\\xf8\\xbf\\x99\\'
'x03\\xd4\\xa6\\x00\\xe5\\xd3\\xbf@\\x1b\\x80\\x19\\xac\\xfe\\xdc\\xbf\\xa0'
'Vw\\x8d|\\xd8\\xe9\\xbf\\x0c\\x0b\\xfa\\xd1\\xae\\x9c\\x02\\xc0 s\\xf7\\x9'
'2,\\xd0\\xe2\\xbfH\\xf3\\x95\\x96\\xf5\\x91\\x03@\\x86?\\x18\\xab\\xaf>\\x'
'04@x_\\xc3\\x80\\xa8$\\xff?\\xf3\\xb1Fq\\x1b_\\xf8?\\x08\\xc5#\\xcf\\xf0'
'\\xc4\\xf9?\\xd6\\xa7D\\xbf!]\\xf8?\\x11\\xb5\\xc7\\x89\\xb2P\\xd8?\\x81\\'
'x1c\\x12!\\xa4t\\xff?\\xc4O\\xa8\\x02\\xb9\\xbd\\xc4\\xbf;2\\x0c\\xa1U\\x'
'03\\xf8\\xbf\\x9ci B!\\xdc\\xf1\\xbfa\\xf6\\x17\\x0c#s\\xfc\\xbf\\xa20\\x'
'e9G\\x8a\\xbc\\xed\\xbfB\\x15\\xb9M\\x04U\\xb4\\xbf\\xf2\\xee\\xa4\\xc5\\x'
'f1\\xba\\xfb?\\x81\\xb4\\xa2cL\\x01\\x03\\xc04\\x16\\xbc\\xd5h\\x11\\x0f@'
'UHS+$\\x8c\\x02@\\xf1\\xb0P\\x18\\xd8o\\xdd\\xbf\\xe894}\\xea\\x8d\\xeb\\'
'xbf\\xf6$\\x86\\x89#\\xa8\\xe8\\xbf\\x0b\\xf9\\x11\\xf1\\xc6\\xe1\\xd6?B\\'
'xe0\\xdd;I\\xb7\\xf5?\\x06\\xfa\\x91\\xd4\\x84\\xae\\x9a\\xbfx\\xad\\x18\\'
'xcf(\\xae\\x08@*\\xce\\xdcD!\\xa8\\xeb\\xbf\\xad\\xf5i\\x17\\x9a\\xa4\\xf'
'b?DS\\xfa\\x14\\xdfo\\xf5?\\x883\\x8f\\xb1V&\\xf2?\\x84\\xc8#\\xdf\\xdd\\'
'x1e\\xcb?\\x82\\r}\\xb1\\xf5\\xe2\\x06@)V\\xdb#\\x0c%\\x00\\xc0\\xf9y\\xc'
'6\\x0c%\\xf4\\xca?\\x0eW\\xcb\\x0e\\xc1\\xbc\\x08\\xc0x_\\xc3\\x80\\xa8$\\'
'xff?\\xca+\\xc3\\xb8\\xcc"\\xef?2\\x840\\xb3\\x13I\\xc7?\\xb4\\x1b\\xc9z'
'gl\\xe9?\\xcc|\\xd2\\xbd7\\xeb\\xf6?\\xfa\\x15}\\xd9\\xf6-\\x02@\\xe3\\xd'
'f\\x8cw\\x13\\x8d\\x06@S\\xceC\\xd8w6\\xf9?\\xb8\\xebE}\\x0e\\x91\\xda?>'
'\\xd9\\x80Y\\xa7\\xa3\\xce?2\\x840\\xb3\\x13I\\xc7?4\\xb4\\xdc)2\\xcf\\xc'
'c\\xbf\\x10jD`d\\xff\\xec?b\\x12\\x84K\\xea\\x8c\\xd8\\xbf8\\xb5\\xbf\\x8'
'3\\xe4\\xdd\\x01@\\x80\\xa9\\x84j\\x0c}\\x08@1=\\xd3\\x14>\\xdc\\xf8\\xbf'
'\\x99\\x03\\xd4\\xa6\\x00\\xe5\\xd3\\xbf\\xea\\x03\\xba\\x9c\\xde\\xd5\\xd2'
"\\xbfO\\x11\\tz;\\xc3\\xe4?8\\'\\x9a\\x91l \\xdd?\\xfb\\xe6gO\\xae|\\xcd\\"
'xbfp\\x8e\\xadE \\xd4\\x07@s\\xdf\\x84\\x1e]&\\xcd\\xbf\\xad\\xf5i\\x17\\'
'x9a\\xa4\\xfb?\\xca+\\xc3\\xb8\\xcc"\\xef?2\\x840\\xb3\\x13I\\xc7?\\xe3='
'G\\xb8[\\xf4\\xb0?\\xa8\\x7fO\\xb9\\xaf\\xb2\\x0b@\\x9c\\xfdA\\xfd\\x10\\x'
'a6\\xf5?\\xc4O\\xa8\\x02\\xb9\\xbd\\xc4\\xbf[\\x06L\\xb3e\\x0e\\xee?\\xad'
'\\xf5i\\x17\\x9a\\xa4\\xfb?\\xca+\\xc3\\xb8\\xcc"\\xef?\\xb9OX\\xe6\\xdf#'
'\\xea?\\x08\\xad\\xa4\\xbf\\xb7\\xb1\\xfa?\\xfc0[\\\\\\x87\\x97\\xea? s\\xf'
'7\\x92,\\xd0\\xe2\\xbf\\x1f4P\\xa9s7\\xfb?\\xa8L\\xbd!\\xb3\\xeb\\xf2\\xb'
'f\\xad\\xf5i\\x17\\x9a\\xa4\\xfb?\\\\\\xe7\\xf1\\x1d\\xc5\\x0e\\xdb?\\xea\\x'
'03\\xba\\x9c\\xde\\xd5\\xd2\\xbf\\xea\\x06Iy\\x0f\\x1a\\xe0?\\x90\\xc4\\x8'
'2K\\xbdV\\xff?yhC\\x1a\\r\\xd1\\xc6?P\\x12\\x01 \\xfek\\x04@\\xae\\x06\\x'
'1a\\xda\\x12+\\xc9?N\\xb8]\\xdbn$\\xf1?\\xf3\\xb1Fq\\x1b_\\xf8?\\x0eD\\x'
"f5\\'y\\x0f\\xe5?\\xea\\x06Iy\\x0f\\x1a\\xe0?@\\x95Z\\x12\\xfbe\\xfa\\xbf"
'\\x10\\r*\\x8fZ\\x8d\\xfa?\\xb9\\xf2_\\x02\\x84\\x08\\xf7\\xbfHz\\xb7\\x0f'
'\\x93\\xf2\\x01\\xc0q\\xc9U\\x11\\xa4H\\xe4?\\x99\\x03\\xd4\\xa6\\x00\\xe5'
'\\xd3\\xbf\\xc6p$\\xd3$\\xf6\\xdf?\\xb4\\x1b\\xc9zgl\\xe9?\\xb8C\\xe9\\xb'
'9Z\\x83\\xf4?\\x05\\x92\\xc7\\xb6\\xbf\\x9e\\xec\\xbf\\xdd\\xc0!\\xee\\n\\x'
'b3\\x05@\\x03\\x03?F\\xdcx\\xe1\\xbf\\xf1\\xb0P\\x18\\xd8o\\xdd\\xbfV~\\x'
'05\\x18\\xf2\\xa1\\xdf\\xbf@\\x1b\\x80\\x19\\xac\\xfe\\xdc\\xbfO\\x11\\tz;'
'\\xc3\\xe4?$\\xa3-dAg\\xef?u\\xa8N\\xde2\\x03\\xd2\\xbf\\xb3%\\n?\\xe0p\\'
'x01@zq|\\x8f\\x1e\\xa5\\x03\\xc0\\x01\\xa2\\x1a\\xb1\\x86\\xd2\\x0b@\\x0e'
'\\x03\\xfbW\\x1dd\\t@\\x1a\\xb1\\x8d\\xd5\\xf8;\\x03@O^b\\xe0\\xbc\\xd7\\x'
'00@I\\xb8\\x97i;Z\\xcc\\xbf\\x05\\xc2\\xc4E6\\xd2\\xfd?$ \\xad\\xcc /\\x'
'd1\\xbf\\xf8\\x17\\xc2\\xb3N\\xa9\\xea?\\x06\\x9d\\x08\\xe4\\xc0H\\xeb?\\r'
'\\xb1\\x91G\\xdbe\\xe3?\\xb9OX\\xe6\\xdf#\\xea?B\\x15\\xb9M\\x04U\\xb4\\x'
'bf$\\xa3-dAg\\xef?7\\xc3#\\xc7-\\xff\\x02@\\x84\\x94\\xb2=\\x87g\\xdb\\x'
'bf\\xee\\xc35\\xebie\\x03@1=\\xd3\\x14>\\xdc\\xf8\\xbfa\\xf6\\x17\\x0c#s'
'\\xfc\\xbf\\xc6p$\\xd3$\\xf6\\xdf?\\xa3\\xa2\\xe4\\xbe\\x8b\\x08\\xf6?:\\x'
'bd\\xa0\\xe4\\x08\\xc3\\xc3?\\xfa}\\xfe\\x91;\\x94\\xf1\\xbf@\\xb0\\xfc\\x'
'ce\\x95\\x9f\\xfe?;2\\x0c\\xa1U\\x03\\xf8\\xbf\\xa3\\xff\\x8d\\xb1%\\xb8\\'
'xec\\xbf\\xb7\\x11Ek\\x1eP\\xc0\\xbf\\\'\\xd9\\xe7?"Z\\xc1\\xbf\\xd6A\\xf7'
'\\x8b$\\x86\\xe0\\xbf\\xcc|\\xd2\\xbd7\\xeb\\xf6?K\\xba\\xd0|\\x91^\\xfb?'
'\\xf9y\\xc6\\x0c%\\xf4\\xca?\\xee\\xc35\\xebie\\x03@\\x86\\x19\\xbb\\xa9b'
'\\x12\\x03@\\xa2\\x10\\x93\\xcdWN\\xfb?\\xdeJU.$O\\xfc?\\xb4\\x1b\\xc9zg'
'l\\xe9?n\\xd4\\x93\\xad\\x18{\\x04@u\\xa8N\\xde2\\x03\\xd2\\xbfr\\x04\\\\\\'
'xd7\\xf6\\xcf\\xe2\\xbf\\t;G!\\xcaP\\xf6\\xbf1=\\xd3\\x14>\\xdc\\xf8\\xbf'
'a\\xf6\\x17\\x0c#s\\xfc\\xbf\\xf6$\\x86\\x89#\\xa8\\xe8\\xbf\\xea\\x06Iy\\'
'x0f\\x1a\\xe0?\\x90\\x86\\x93\\x88Rb\\xf4\\xbfOU\\x19\\x956\\x9f\\x99?\\x'
'00\\xb8\\xa3\\x83Q\\xcf\\xf7?\\x18\\x83\\xf2\\x8f\\x07@\\x00\\xc0\\xfc\\xa'
"6,~L\\\\\\xfc\\xbf\\x10Udh_b\\xff\\xbf\\x94\\xa0\\'_\\xd6R\\x00\\xc0\\n\\xe"
'd\\r%\\xc2\\xf2\\x00\\xc0\\xf3\\xb0\\xb5\\x02\\x87\\xc6\\xf0\\xbf3)\\xb1\\x'
'7f<{\\xb3?\\x84\\x94\\xb2=\\x87g\\xdb\\xbf\\xd2L$\\xb5\\tz\\xe0?68\\xc1G'
'xR\\x08@\\xf3\\xb1Fq\\x1b_\\xf8?\\xb9OX\\xe6\\xdf#\\xea?B\\x15\\xb9M\\x0'
'4U\\xb4\\xbf\\xfc0[\\\\\\x87\\x97\\xea?\\x07\\xf2\\xc1\\xd4\\xac\\x05\\xdf?'
'\\xa3\\x10_c\\xa9\\xcf\\xde\\xbf\\xf8\\x17\\xc2\\xb3N\\xa9\\xea?\\x8e\\x13'
'\\xd6\\xe5<\\xdf\\xce\\xbf\\x99\\x03\\xd4\\xa6\\x00\\xe5\\xd3\\xbf&\\x1e&\\'
'x83xh\\xf1\\xbf\\xb0\\xcf[IX<\\xfd\\xbf\\x06(\\x8e\\xc9\\xce"\\xfe?\\x0f'
"=\\'\\x1e\\xd1\\xc0\\xdb?\\x84\\x94\\xb2=\\x87g\\xdb\\xbf\\xe0\\xbc4k\\xe5"
')\\xda?\\xa3\\xff\\x8d\\xb1%\\xb8\\xec\\xbf\\xb1\\x97\\xcb\\xaf\\xe6\\x83\\'
"xf9\\xbf\\x94\\xa0\\'_\\xd6R\\x00\\xc0\\n\\xed\\r%\\xc2\\xf2\\x00\\xc0\\xa4"
'\\n\\x00\\xb6}\\x1b\\xf2?\\x17\\x88\\x8cg\\xf5{\\xd8?}-\\x15\\x1d\\x9b\\x1'
'a\\xdb?d\\xf1\\xc8E\\xf3\\xdd\\xe4\\xbff\\xd3y\\xab/\\\\\\xf5\\xbf\\x10Udh'
"_b\\xff\\xbf\\x94\\xa0\\'_\\xd6R\\x00\\xc0\\xd6\\xf4\\x9d%\\xa3q\\x04\\xc0"
'\\x10jD`d\\xff\\xec?\\x17\\x88\\x8cg\\xf5{\\xd8?\\xe3D\\xee\\xaa`\\r\\xaa'
'?\\xc7\\xdfRE\\nC\\xe8\\xbf\\xfc\\xa6,~L\\\\\\xfc\\xbf7\\x89~\\x10l\\xa0\\x'
'02\\xc0?\\xac\\x8a\\x1d=g\\x05\\xc0\\n\\xed\\r%\\xc2\\xf2\\x00\\xc0\\xd3(\\'
'x82\\x011%\\xef\\xbf\\x89\\xfe\\r\\xadUG\\xc0?/If\\x12\\xc5y\\xe9?\\t;G!'
'\\xcaP\\xf6\\xbf1=\\xd3\\x14>\\xdc\\xf8\\xbf\\x10Udh_b\\xff\\xbfj&Y\\xbe'
'\\t\\xdd\\x02\\xc0\\xa4\\xef=%\\r\\x1d\\x02\\xc0\\x86D\\xdd\\xf1\\xbc\\x85\\'
'xe5\\xbf\\x07\\xf2\\xc1\\xd4\\xac\\x05\\xdf?\\xb7\\xa0\\xdf\\xae\\x04%\\xf'
'0?\\x14\\xc6V#R\\xda\\x8f\\xbf\\xf1\\xb0P\\x18\\xd8o\\xdd\\xbf\\xa3{\\xe6'
'\\x9a1\\xb6\\xf0\\xbf}5\\xec\\xffE\\x91\\xfb\\xbf\\xe2\\xd4\\xbbI\\xee\\x9'
'0\\xff\\xbfB\\xe0\\xdd;I\\xb7\\xf5?dr~i\\xd7\\xf3\\xea?\\xd0\\xfd K\\x1e'
'\\xb3\\xf2?\\xd2L$\\xb5\\tz\\xe0?\\xfc\\xa6,~L\\\\\\xfc\\xbf\\x10Udh_b\\xf'
"f\\xbf\\x94\\xa0\\'_\\xd6R\\x00\\xc0\\xb0\\xcf[IX<\\xfd\\xbf\\xca1O\\x94\\"
'xe9\\x99\\xfb\\xbfaR\\x8f\\x0f\\x10\\xbf\\xb3\\xbf\\xe0\\x127\\xb4o\\xf5\\'
'xe0?;2\\x0c\\xa1U\\x03\\xf8\\xbf\\xfc\\xa6,~L\\\\\\xfc\\xbf\\x10Udh_b\\xf'
"f\\xbf\\x94\\xa0\\'_\\xd6R\\x00\\xc0\\xe2\\xd4\\xbbI\\xee\\x90\\xff\\xbf\\x"
'1a#\\x88\\nA\\x96\\xf5\\xbf\\x89\\xfe\\r\\xadUG\\xc0?\\xb7\\xa0\\xdf\\xae\\'
'x04%\\xf0?_\\xfajY\\xbe\\xb9\\xf0?\\xfc\\xa6,~L\\\\\\xfc\\xbf\\xe0YX\\xe2'
'\\xcd(\\x01\\xc0j&Y\\xbe\\t\\xdd\\x02\\xc0\\xa4\\xef=%\\r\\x1d\\x02\\xc0\\x'
'ad\\xb6\\xaf\\xf9vU\\xea\\xbf3)\\xb1\\x7f<{\\xb3?\\xfe\\x8e\\xe3\\xd9\\x9'
'1]\\xe4?d\\xf1\\xc8E\\xf3\\xdd\\xe4\\xbf68\\xc1GxR\\x08@\\x00\\xe7\\x15C'
'h\\x96\\x00@]\\xb9\\xc0\\x10\\x8a\\xb0\\xf4?\\x0b\\xf9\\x11\\xf1\\xc6\\xe1'
'\\xd6?[\\xdf\\x9b\\x94\\xa66\\xb2\\xbf\\x05\\x92\\xc7\\xb6\\xbf\\x9e\\xec\\'
'xbf\\x81\\x19r@H\\x12\\xf1\\xbf\\xd4K\\xba@\\x02$\\xbf\\xbf\\x1c\\xedm|\\'
'x7f\\x12\\n@_\\xa4\\xae\\xfb\\xe0t\\x06@\\x08\\xc5#\\xcf\\xf0\\xc4\\xf9?\\'
'xa3\\xa2\\xe4\\xbe\\x8b\\x08\\xf6?=\\x94\\xde\\xaa\\x18\\x84p?\\xf02\\x99'
'H\\x17\\xd9\\xf4\\xbf\\xe2\\xb67[n\\xbc\\xee\\xbfs\\xdf\\x84\\x1e]&\\xcd\\'
'xbf\\xfc\\xa6,~L\\\\\\xfc\\xbf\\xf3\\xb1Fq\\x1b_\\xf8?d[\\xbb\\xa4F8\\xef'
'?\\x84\\xc8#\\xdf\\xdd\\x1e\\xcb?}M\\xaa\\x84u\\xfa\\xf1\\xbfb\\x12\\x84K'
'\\xea\\x8c\\xd8\\xbf\\xc4O\\xa8\\x02\\xb9\\xbd\\xc4\\xbf_\\xfajY\\xbe\\xb9'
'\\xf0?\\x01\\xa2\\x1a\\xb1\\x86\\xd2\\x0b@X\\x16<q\\x06\\x0e\\x02@\\x08\\x'
'c5#\\xcf\\xf0\\xc4\\xf9?\\x18&\\x89{\\x93\\x15\\xee?\\x89\\xa1E\\xf4|b\\x'
'cd?\\xa4\\x18\\xaa\\xb7\\xbe-\\xe1\\xbf\\xa2\\xbe\\xde\\x0f*\\xec\\xe7\\xb'
'f\\xae\\x06\\x1a\\xda\\x12+\\xc9?\\x1c\\xedm|\\x7f\\x12\\n@X\\x16<q\\x06\\'
'x0e\\x02@3?\\xf2o\\xbd:\\xf7?\\xb4\\x1b\\xc9zgl\\xe9?r\\xc0p\\xc4\\x91\\'
'xcc\\xd7\\xbf\\xbc\\xd0W\\xa4\\x04\\xc3\\xf0\\xbf\\xa1\\x95\\x1efjz\\xf4\\'
'xbf*\\xce\\xdcD!\\xa8\\xeb\\xbf\\xf1\\xb0P\\x18\\xd8o\\xdd\\xbf\\x94\\xf4'
'\\xad\\xb8\\xa2\\x80\\xf2?\\xf06\\xbf4,\\xc6\\x05@}}\\xa2\\xe2@\\xd3\\x0e'
'@[\\xdf\\x9b\\x94\\xa66\\xb2\\xbfOU\\x19\\x956\\x9f\\x99?\\x81\\x19r@H\\x'
'12\\xf1\\xbfs\\xdf\\x84\\x1e]&\\xcd\\xbf\\xcc)V\\xd8L\\xf6\\x96\\xbf\\xca'
'+\\xc3\\xb8\\xcc"\\xef?oY^VW\\xcd\\xd5?\\xb4\\x1b\\xc9zgl\\xe9?\\xbc\\x1'
'9\\xc1[W\\xa0\\x08\\xc0K\\xba\\xd0|\\x91^\\xfb?\\xe2\\x8dw\\xb1\\xaeJ\\xf'
'b\\xbf\\x9c \\x96\\xa0lh\\xfb\\xbf\\xf1\\xb0P\\x18\\xd8o\\xdd\\xbfkn*\\x0'
'0TD\\xe9?2\\x840\\xb3\\x13I\\xc7?;L\\xb7\\x8cP/\\xe5\\xbf:\\xbd\\xa0\\xe'
'4\\x08\\xc3\\xc3?.\\xe0?6N\\xaa\\xf5\\xbf\\xf2\\xcbM\\xc4\\xbf\\xfe\\xfc\\'
'xbf;2\\x0c\\xa1U\\x03\\xf8\\xbf\\xa3\\xff\\x8d\\xb1%\\xb8\\xec\\xbf\\xb7\\'
'x11Ek\\x1eP\\xc0\\xbfK\\x19#\\xcb\\xbc\\x93\\xe3\\xbf\\xe3nn\\x16\\xf1\\x'
'b9\\xd7\\xbfF\\xb6\\xb5\\xddE\\xd4\\t\\xc0\\xb3\\x1dq\\xa2\\x9b\\xa1\\x04@'
'c\\x18\\x06\\x18e\\xff\\xd7\\xbf4;\\xae\\xb4 \\xdf\\xe3?1=\\xd3\\x14>\\xd'
'c\\xf8\\xbf\\x10Udh_b\\xff\\xbf\\xd2)\\x89A\\xdf|\\xf6\\xbf\\xe3nn\\x16\\'
'xf1\\xb9\\xd7\\xbf\\x11\\xb5\\xc7\\x89\\xb2P\\xd8?P|\\xb9\\xb8\\xa1\\x16\\'
'xdf\\xbf\\x0f\\xf6y\\x96b\\x83\\xf9?\\xb2\\xa0I\\xea\\x97/\\n@1=\\xd3\\x1'
'4>\\xdc\\xf8\\xbfa\\xf6\\x17\\x0c#s\\xfc\\xbf\\xf6$\\x86\\x89#\\xa8\\xe8\\'
'xbf4\\xb4\\xdc)2\\xcf\\xcc\\xbf\\xa4\\n\\x00\\xb6}\\x1b\\xf2?\\x16(\\x92I'
'\\x08\\x15\\xe6\\xbf8\\xb5\\xbf\\x83\\xe4\\xdd\\x01@\\xee\\xc35\\xebie\\x0'
'3@f\\xd3y\\xab/\\\\\\xf5\\xbfa\\xf6\\x17\\x0c#s\\xfc\\xbf+\\xac"\\xcd\\xc5'
'\\xbb\\x97?\\x04a7\\x8e\\xa8\\x81\\xee\\xbf\\x90\\x86\\x93\\x88Rb\\xf4\\xb'
'f\\x06\\xfa\\x91\\xd4\\x84\\xae\\x9a\\xbf\\x08$\\xfd\\xf4p3\\x00\\xc0vU\\x'
"f8\\xa1\\'9\\xf1\\xbf\\xf1\\xb0P\\x18\\xd8o\\xdd\\xbf\\xe894}\\xea\\x8d\\x"
'eb\\xbf\\xd2)\\x89A\\xdf|\\xf6\\xbfK\\xc5\\x9bH,\\x93\\xf8\\xbf.\\xa7\\xf'
'47lO\\xf3?\\x89\\xfe\\r\\xadUG\\xc0?a\\xc6\\x85\\xc4\\xe5\\x1b\\xe1\\xbf\\'
'x85\\xc5\\x08X\\x03\\xe9\\xfa?1=\\xd3\\x14>\\xdc\\xf8\\xbf\\xe0YX\\xe2\\x'
'cd(\\x01\\xc0\\xd2)\\x89A\\xdf|\\xf6\\xbf4\\xb4\\xdc)2\\xcf\\xcc\\xbf\\x8'
'2n\\x05P\\xc0h\\x01\\xc0\\x7f\\xec\\x14\\x92-A\\xee\\xbf\\x9e\\xa9\\xc1B\\'
'xbd\\x82\\xde?[\\x06L\\xb3e\\x0e\\xee?\\x9ci B!\\xdc\\xf1\\xbfkn*\\x00T'
'D\\xe9?2\\x840\\xb3\\x13I\\xc7?\\xa0Vw\\x8d|\\xd8\\xe9\\xbf$\\xa3-dAg\\x'
'ef?\\xfa}\\xfe\\x91;\\x94\\xf1\\xbf\\x16}\\xd9\\xaf\\xc1t\\xac\\xbf\\xe0\\'
'xbc4k\\xe5)\\xda?\\xf1\\xb0P\\x18\\xd8o\\xdd\\xbf\\x8a|\\x9b\\xc4q\\xaf\\'
'xe5\\xbf&\\x1e&\\x83xh\\xf1\\xbf\\x19\\xc0;H\\x96>\\xf6\\xbfKN\\x9e\\xbc'
'\\xd7\\xfc\\xd2\\xbfl\\x8d\\xe6#\\x85{\\xf6\\xbf\\x91WHSY\\xc6\\xf2\\xbfz'
'q|\\x8f\\x1e\\xa5\\x03\\xc0\\xf1\\xb0P\\x18\\xd8o\\xdd\\xbf\\x8a|\\x9b\\xc'
'4q\\xaf\\xe5\\xbf\\xa20\\xe9G\\x8a\\xbc\\xed\\xbfK\\xc5\\x9bH,\\x93\\xf8\\'
'xbf_\\xd2\\n\\xea\\x02\\xb6\\xe0\\xbfg\\x03\\x82\\x11C\\xf6\\x00\\xc0!\\x8'
'1>\\x91\\x8a\\xc1\\x02\\xc0C\\xe4#\\xcf\\xef0\\x06\\xc0N\\xb8]\\xdbn$\\xf'
'1?\\x0e\\x8fw\\xdc\\x11\\xa7\\xac?K\\x19#\\xcb\\xbc\\x93\\xe3\\xbf\\x04a7'
'\\x8e\\xa8\\x81\\xee\\xbf\\x96Ff\\xb5\\xd2J\\t@\\x10\\r*\\x8fZ\\x8d\\xfa?\\'
'xcc\\x82K\\xdb\\xf9\\xfe\\x03@\\x7f\\x9b\\xc2\\xf3\\x10G\\x17@\\xa3\\xff\\'
'x8d\\xb1%\\xb8\\xec\\xbfS\\xda2\\xf7m\\xa5\\xf3\\xbf}5\\xec\\xffE\\x91\\x'
'fb\\xbf\\n\\xed\\r%\\xc2\\xf2\\x00\\xc0\\xc2\\x85\\x9fP\\xf0_\\xe3?\\x7f\\x'
'ec\\x14\\x92-A\\xee\\xbfr\\x04\\\\\\xd7\\xf6\\xcf\\xe2\\xbf\\x1b\\xe0 l\\xb'
'7_\\xd3?\\x86\\x19\\xbb\\xa9b\\x12\\x03@DS\\xfa\\x14\\xdfo\\xf5?\\x0eD\\x'
"f5\\'y\\x0f\\xe5?\\x0b\\xf9\\x11\\xf1\\xc6\\xe1\\xd6?_\\xd2\\n\\xea\\x02\\x"
'b6\\xe0\\xbf\\xbc\\xd0W\\xa4\\x04\\xc3\\xf0\\xbf\\xa3\\x10_c\\xa9\\xcf\\xd'
'e\\xbf\\xe0\\xbc4k\\xe5)\\xda?\\x18"\\xb7D}\\xa4\\xf4?\\x0e\\x8fw\\xdc\\x'
'11\\xa7\\xac?\\xea\\x03\\xba\\x9c\\xde\\xd5\\xd2\\xbf\\xe3nn\\x16\\xf1\\xb'
'9\\xd7\\xbf\\x86D\\xdd\\xf1\\xbc\\x85\\xe5\\xbf\\x9a\\xcdDn\\x9ar\\xe4\\xb'
'f\\x82B2\\xea\\x07\\x84\\xe4\\xbf_\\xfajY\\xbe\\xb9\\xf0?\\x9ci B!\\xdc\\'
'xf1\\xbfRo\\xdf)\\x94=\\xfe?\\xdeJU.$O\\xfc?q\\x9d\\x84\\xbe\\xf5\\xb3\\'
'xf3?\\x1a#\\x88\\nA\\x96\\xf5\\xbf\\x0b}2\\xe2\\xf6\\xf2\\xc6\\xbf\\xe2\\x'
'b67[n\\xbc\\xee\\xbfM\\x9a\\xe4\\xb7m-\\xb7?\\x18"\\xb7D}\\xa4\\xf4?>\\x'
'd9\\x80Y\\xa7\\xa3\\xce?\\xc6p$\\xd3$\\xf6\\xdf?\\xd6A\\xf7\\x8b$\\x86\\x'
'e0\\xbf=\\x94\\xde\\xaa\\x18\\x84p?i\\xd2x\\x87\\xc4Z\\xcd?\\x9d\\xf1\\x8'
'8L\\x86T\\x02\\xc0Hz\\xb7\\x0f\\x93\\xf2\\x01\\xc0q\\xc9U\\x11\\xa4H\\xe4'
'?\\x0e\\x03\\xfbW\\x1dd\\t@\\x0ej[\\x88\\x96<\\x11@}}\\xa2\\xe2@\\xd3\\x0'
'e@\\xd3(\\x82\\x011%\\xef\\xbf3)\\xb1\\x7f<{\\xb3?\\xa9\\xb4\\x89\\xefrT'
'\\xf5\\xbf\\xa8L\\xbd!\\xb3\\xeb\\xf2\\xbf68\\xc1GxR\\x08@\\xb7\\xd3\\xd4'
')\\x7f\\xec\\x07@E+\\\\v\\xc5\\xb1\\x00@\\x08\\xad\\xa4\\xbf\\xb7\\xb1\\xfa'
'?\\xa4\\n\\x00\\xb6}\\x1b\\xf2?k]\\xe9\\x94\\x0eH\\xd5\\xbfr\\x04\\\\\\xd7\\'
'xf6\\xcf\\xe2\\xbf\\x14\\xc6V#R\\xda\\x8f\\xbf\\xf1\\xb0P\\x18\\xd8o\\xdd'
'\\xbf\\\\\\xe7\\xf1\\x1d\\xc5\\x0e\\xdb?3?\\xf2o\\xbd:\\xf7?\\x82c\\xc2\\xe'
'0R,\\x03@\\xd3(\\x82\\x011%\\xef\\xbf\\xa4\\x18\\xaa\\xb7\\xbe-\\xe1\\xbf'
'\\xb7\\xa0\\xdf\\xae\\x04%\\xf0?\\x96)8\\xb47D\\xe7?\\x86\\x19\\xbb\\xa9b'
'\\x12\\x03@\\xb0Eb\\x9f\\xa4\\x85\\x03@\\x9bB"\\xf3\\x92\\xda\\n@\\x80p\\x'
'b2\\xe1\\xc9\\xff\\x08@\\x11\\xb5\\xc7\\x89\\xb2P\\xd8?\\xe8\\x171\\x8eiQ'
'\\xe9?a\\xc6\\x85\\xc4\\xe5\\x1b\\xe1\\xbf4;\\xae\\xb4 \\xdf\\xe3?\\xa1d\\'
'x0eu[R\\x01@\\x00\\xe7\\x15Ch\\x96\\x00@\\xc5\\xbc\\xf0\\x93_P\\x08@\\xb'
'4h"\\xe1\\xe8\\x80\\x05@:\\xbd\\xa0\\xe4\\x08\\xc3\\xc3?\\x0f=\\\'\\x1e\\xd'
'1\\xc0\\xdb?\\xb2\\xfc\\xb4";\\xa0\\xe9\\xbf\\x14\\xc6V#R\\xda\\x8f\\xbf\\'
'xf1\\xb0P\\x18\\xd8o\\xdd\\xbf\\x00\\xe7\\x15Ch\\x96\\x00@\\xc5\\xbc\\xf0'
'\\x93_P\\x08@\\xe9`\\x92\\xe0\\x07\\x02\\x02@\\xd3(\\x82\\x011%\\xef\\xbf'
'\\x1f\\xd3\\xf1\\xb0\\x197\\xd5?y\\xfa\\x06\\xb7?8\\xf0\\xbf\\xd4K\\xba@\\'
'x02$\\xbf\\xbf\\xe3\\x8b\\x10\\xae\\x8b$\\xf8?\\x00\\xe7\\x15Ch\\x96\\x00'
'@\\x9bB"\\xf3\\x92\\xda\\n@\\x80p\\xb2\\xe1\\xc9\\xff\\x08@=\\x94\\xde\\xa'
'a\\x18\\x84p?u\\x08I\\xfc\\x1fj\\xe4?\\xb2\\xfc\\xb4";\\xa0\\xe9\\xbfM\\x'
'9a\\xe4\\xb7m-\\xb7?f\\xd3y\\xab/\\\\\\xf5\\xbfa\\xf6\\x17\\x0c#s\\xfc\\xb'
'f\\xfc\\xa3W\\xe2\\xab\\xf2\\xf3\\xbf;L\\xb7\\x8cP/\\xe5\\xbf\\x89\\xa1E\\'
'xf4|b\\xcd?Y\\xf7\\x1b\\x91<\\x05\\xfd\\xbf\\\\\\xb1h\\xf7x\\xb2\\xd7?\\xc'
'7\\xdfRE\\nC\\xe8\\xbf\\x86\\x19\\xbb\\xa9b\\x12\\x03@\\x94\\xf4\\xad\\xb8'
'\\xa2\\x80\\xf2?oY^VW\\xcd\\xd5?\\xe3=G\\xb8[\\xf4\\xb0?.\\xa7\\xf47lO\\'
'xf3?\\x0e\\xdd,\\x00\\xe4Y\\xe9\\xbf$ \\xad\\xcc /\\xd1\\xbf\\xfe\\x0e \\'
'xa0\\x83\\xcd\\xfe\\xbf\\xcc)V\\xd8L\\xf6\\x96\\xbf\\x8a|\\x9b\\xc4q\\xaf'
'\\xe5\\xbf\\xf6$\\x86\\x89#\\xa8\\xe8\\xbf\\xd6A\\xf7\\x8b$\\x86\\xe0\\xbf'
'\\xf3\\xb0\\xb5\\x02\\x87\\xc6\\xf0\\xbf\\xf1b\\x96\\xd7\\x8d\\x0c\\xe6?\\x'
'fe\\x8e\\xe3\\xd9\\x91]\\xe4?\\xdc_\\x06\\x8f5\\n\\x07\\xc0\\x06\\x9d\\x08'
'\\xe4\\xc0H\\xeb?\\xb7\\x11Ek\\x1eP\\xc0\\xbf\\\'\\xd9\\xe7?"Z\\xc1\\xbf\\x'
'd6A\\xf7\\x8b$\\x86\\xe0\\xbf\\xde\\xb5\\xbb\\xc1\\x14S\\xf9?\\xa4\\x18\\x'
'aa\\xb7\\xbe-\\xe1\\xbf\\xc4O\\xa8\\x02\\xb9\\xbd\\xc4\\xbfC\\xe4#\\xcf\\x'
'ef0\\x06\\xc0\\xcc)V\\xd8L\\xf6\\x96\\xbf\\x94\\xf4\\xad\\xb8\\xa2\\x80\\x'
'f2?oY^VW\\xcd\\xd5?\\xe3nn\\x16\\xf1\\xb9\\xd7\\xbf\\xb6\\xe3\\x7f\\xa3\\'
'x1a\\x0f\\x12@\\xfa}\\xfe\\x91;\\x94\\xf1\\xbf@\\x87<%\\xd6-\\xeb?\\x9c '
'\\x96\\xa0lh\\xfb\\xbf1=\\xd3\\x14>\\xdc\\xf8\\xbf\\xb7\\x11Ek\\x1eP\\xc0'
'\\xbf\\xc6p$\\xd3$\\xf6\\xdf?4\\xb4\\xdc)2\\xcf\\xcc\\xbf\\xa9\\xe0\\xd7W'
'z8\\x06\\xc0\\xd2_\\x83\\xa1#\\xbc\\xf9?}-\\x15\\x1d\\x9b\\x1a\\xdb?\\x18'
'\\x83\\xf2\\x8f\\x07@\\x00\\xc0\\xb8\\xebE}\\x0e\\x91\\xda?V~\\x05\\x18\\x'
'f2\\xa1\\xdf\\xbf\\xea\\x03\\xba\\x9c\\xde\\xd5\\xd2\\xbf\\xb4\\xb5{Gj\\x9'
'5\\xf1\\xbf\\x82\\r}\\xb1\\xf5\\xe2\\x06@\\xe8\\x171\\x8eiQ\\xe9?@\\x87<%'
'\\xd6-\\xeb?\\x9c \\x96\\xa0lh\\xfb\\xbfq\\xc9U\\x11\\xa4H\\xe4?kn*\\x00'
'TD\\xe9?+\\xac"\\xcd\\xc5\\xbb\\x97?\\xa0Vw\\x8d|\\xd8\\xe9\\xbf=\\xbf\\x'
'82p\\xfeH\\x06@\\x89\\xfe\\r\\xadUG\\xc0?\\xb0\\xd3\\xf4x{.\\xf6\\xbf\\x9'
'c \\x96\\xa0lh\\xfb\\xbf\\xf1\\xb0P\\x18\\xd8o\\xdd\\xbf\\xb7\\x11Ek\\x1e'
'P\\xc0\\xbf@\\x1b\\x80\\x19\\xac\\xfe\\xdc\\xbf\\xb4\\xb5{Gj\\x95\\xf1\\xb'
'f|\\x8b\\x99G\\xe0\\xee\\xfc?\\xa5H\\xa7F5a\\xf2?\\xfe\\x8e\\xe3\\xd9\\x9'
'1]\\xe4?\\xb0\\xfe\\xd4OM\\x19\\x01\\xc0\\x06\\x9d\\x08\\xe4\\xc0H\\xeb?k'
'n*\\x00TD\\xe9?oY^VW\\xcd\\xd5?;L\\xb7\\x8cP/\\xe5\\xbf\\xd9\\x94`\\xf6'
'\\xc9\\xe4\\t@\\xf1b\\x96\\xd7\\x8d\\x0c\\xe6?\\xc1\\x11\\xcb\\x8b\\x8c\\xe'
'2\\xf7\\xbfd\\xf1\\xc8E\\xf3\\xdd\\xe4\\xbf\\xa3\\xff\\x8d\\xb1%\\xb8\\xec'
'\\xbf\\xe894}\\xea\\x8d\\xeb\\xbf@\\x1b\\x80\\x19\\xac\\xfe\\xdc\\xbf\\xb4'
'\\xb5{Gj\\x95\\xf1\\xbf\\xf2\\xee\\xa4\\xc5\\xf1\\xba\\xfb?\\x10\\r*\\x8fZ'
"\\x8d\\xfa?\\xd8\\x1c\\x8c\\xd4&\\x8d\\xf3?vU\\xf8\\xa1\\'9\\xf1\\xbff\\xd"
'3y\\xab/\\\\\\xf5\\xbf\\xe894}\\xea\\x8d\\xeb\\xbfK\\x19#\\xcb\\xbc\\x93\\x'
'e3\\xbf\\xe7\\xba\\xdbG\\x00\\xea\\xf3\\xbf\\x86D\\xdd\\xf1\\xbc\\x85\\xe5'
'\\xbf\\xd2_\\x83\\xa1#\\xbc\\xf9?\\xe7Zb\\xe77A\\xf5?\\x86?\\x18\\xab\\xa'
'f>\\x04@\\xb8\\xebE}\\x0e\\x91\\xda?\\r\\xb1\\x91G\\xdbe\\xe3?\\xea\\x03\\'
'xba\\x9c\\xde\\xd5\\xd2\\xbf\\xb4\\xb5{Gj\\x95\\xf1\\xbf\\x82\\r}\\xb1\\xf'
"5\\xe2\\x06@\\'\\x1eW\\xfa=\\xf2\\xd1?\\xf9y\\xc6\\x0c%\\xf4\\xca?\\xfe\\x"
'0e \\xa0\\x83\\xcd\\xfe\\xbf\\xad\\xf5i\\x17\\x9a\\xa4\\xfb?\\xb7\\xd3\\xd'
'4)\\x7f\\xec\\x07@\\x0ej[\\x88\\x96<\\x11@\\xa4B\\x99\\xf1\\x10)\\x11@}M'
'\\xaa\\x84u\\xfa\\xf1\\xbf)\\xeeYk\\xc7\\xbe\\xf0?\\xc906\\x15\\x95\\xbc\\'
'xf8\\xbfk)\\xd1 \\xe1\\xb5\\xf9\\xbf\\xa3\\xff\\x8d\\xb1%\\xb8\\xec\\xbf>'
'\\xd9\\x80Y\\xa7\\xa3\\xce?/\\xee\\xf4%\\xdf\\xf6\\x01@m\\xb7d\\xc0\\xe3Z'
'\\xff?@\\x95Z\\x12\\xfbe\\xfa\\xbf\\x7fS\\xaeED%\\xe1?\\x01\\n$\\xd7\\xd0'
'\\xb2\\xfe\\xbf;2\\x0c\\xa1U\\x03\\xf8\\xbff\\xd3y\\xab/\\\\\\xf5\\xbfDS\\x'
'fa\\x14\\xdfo\\xf5?3?\\xf2o\\xbd:\\xf7?\\x18&\\x89{\\x93\\x15\\xee?T\\xc'
'eC\\x16\\xd8\\xcd\\xfc\\xbf\\x7fS\\xaeED%\\xe1?\\xc4O\\xa8\\x02\\xb9\\xbd'
'\\xc4\\xbf\\t;G!\\xcaP\\xf6\\xbf\\xa3\\xff\\x8d\\xb1%\\xb8\\xec\\xbf\\x94\\'
'xf4\\xad\\xb8\\xa2\\x80\\xf2?\\xc6p$\\xd3$\\xf6\\xdf?\\xea\\x06Iy\\x0f\\x'
'1a\\xe0?\\x86D\\xdd\\xf1\\xbc\\x85\\xe5\\xbf\\xbb8\\xc7a\\\\6\\r@$\\xd8\\xe'
'5\\xc2W]\\r@\\x80\\xa9\\x84j\\x0c}\\x08@l\\xceg\\xdei\\xd2\\x04@\\x00\\xe'
'7\\x15Ch\\x96\\x00@d[\\xbb\\xa4F8\\xef?O\\x11\\tz;\\xc3\\xe4?\\x0ekc\\xe'
'8\\xee\\x10\\x15@\\xe1\\xf5M4l2\\xf3?\\x06H\\x01N\\xfd\\x8d\\xcb\\xbf}LV'
'\\x8e\\\\]\\xd5\\xbfq\\xc9U\\x11\\xa4H\\xe4?kn*\\x00TD\\xe9?\\xb9OX\\xe6\\'
'xdf#\\xea?\\x18&\\x89{\\x93\\x15\\xee?.\\xa7\\xf47lO\\xf3?\\xaa\\xd2\\x0'
'bYw\\xe6\\x07@:\\x87\\x8a\\x90\\xd6\\xf0\\x12@\\xac\\n\\xb6\\xa9\\xf4m\\x0'
'e@\\x06\\x9d\\x08\\xe4\\xc0H\\xeb?\\\\\\xe7\\xf1\\x1d\\xc5\\x0e\\xdb?\\xc6p'
'$\\xd3$\\xf6\\xdf?B\\x15\\xb9M\\x04U\\xb4\\xbf\\xe8\\xf7qX\\xaa/\\xe8?\\x'
'05\\xc2\\xc4E6\\xd2\\xfd?\\x88\\xeb\\xee\\xe19b\\n@\\xe9\\xea)\\xd4\\x9cN'
'\\x14@\\x0e,\\xdb\\xde\\x08\\xb8\\xe5\\xbf\\r\\xb1\\x91G\\xdbe\\xe3?\\xc6p'
'$\\xd3$\\xf6\\xdf?O\\x11\\tz;\\xc3\\xe4?_\\xd2\\n\\xea\\x02\\xb6\\xe0\\xbf'
'\\x90!\\xeb\\x06n\\xdb\\x05@h\\x81Y[C\\x06\\x11@I\\x1c,\\xaa\\xdd\\x08\\x'
'0b@\\x06\\x9d\\x08\\xe4\\xc0H\\xeb?\\xa2\\x10\\x93\\xcdWN\\xfb?3?\\xf2o\\'
'xbd:\\xf7??\\x98$\\xbe__\\xf1?B\\xe0\\xdd;I\\xb7\\xf5?Z\\x8f\\x9a\\xd8\\'
'xf7\\x9e\\r@@\\x87<%\\xd6-\\xeb?\\x8f\\xf1/\\xd9Il\\xf2?N\\xb8]\\xdbn$\\'
'xf1?\\xa2\\x10\\x93\\xcdWN\\xfb?\\xdeJU.$O\\xfc??\\x98$\\xbe__\\xf1?\\x'
"de\\xb5\\xbb\\xc1\\x14S\\xf9?\\xc5\\x83,\\xab\\x80\\xf1\\t@\\xf3\\xc0Y\\'\\"
'x08;c\\xbf[\\x06L\\xb3e\\x0e\\xee?\\x8e\\x13\\xd6\\xe5<\\xdf\\xce\\xbf\\x'
'0e\\x8fw\\xdc\\x11\\xa7\\xac?+\\xac"\\xcd\\xc5\\xbb\\x97?\\xe3nn\\x16\\xf'
'1\\xb9\\xd7\\xbf|\\x8b\\x99G\\xe0\\xee\\xfc?\\x89gwj\\xc8/\\xfc?\\x81\\x1'
'9r@H\\x12\\xf1\\xbfHz\\xb7\\x0f\\x93\\xf2\\x01\\xc0\\x1b\\x89\\xc0\\xaf\\x'
'a9!\\xc9?\\x99\\x03\\xd4\\xa6\\x00\\xe5\\xd3\\xbf\\xea\\x03\\xba\\x9c\\xde'
'\\xd5\\xd2\\xbf4\\xb4\\xdc)2\\xcf\\xcc\\xbf.\\xa7\\xf47lO\\xf3?\\x7f\\xbb'
'/\\xfe\\x88\\x8b\\x00@\\xa8g\\x9b\\x07\\\\\\xca\\r@\\xb86\\xdd*;\\xf1\\x05@'
"N\\xb8]\\xdbn$\\xf1?kn*\\x00TD\\xe9?\\x0eD\\xf5\\'y\\x0f\\xe5?\\x18&\\x8"
'9{\\x93\\x15\\xee?B\\xe0\\xdd;I\\xb7\\xf5?\\x14\\xc7\\x9d+\\x009\\x04@`b'
'\\xee\\xd1:,\\x10@4;\\xae\\xb4 \\xdf\\xe3?N\\xb8]\\xdbn$\\xf1?\\\\\\xe7\\x'
'f1\\x1d\\xc5\\x0e\\xdb?oY^VW\\xcd\\xd5?\\x0b\\xf9\\x11\\xf1\\xc6\\xe1\\xd'
'6?hR\\xb0C\\x03\\x87\\xfa?[\\xbf\\xa9b[\\xc5\\x01@\\xa0H0~S\\xf0\\x0c@U'
'HS+$\\x8c\\x02@\\x1b\\x89\\xc0\\xaf\\xa9!\\xc9?\\x94\\xf4\\xad\\xb8\\xa2\\'
'x80\\xf2?\\x08\\xc5#\\xcf\\xf0\\xc4\\xf9?\\x18&\\x89{\\x93\\x15\\xee?\\xf'
'c0[\\\\\\x87\\x97\\xea?\\x0c|8\\xe2\\xdb}\\x07@\\xd2xaH]\\x08\\xed\\xbf\\x'
'03\\x03?F\\xdcx\\xe1\\xbf\\x0e,\\xdb\\xde\\x08\\xb8\\xe5\\xbf\\\\\\xe7\\xf1'
'\\x1d\\xc5\\x0e\\xdb?oY^VW\\xcd\\xd5?\\x18&\\x89{\\x93\\x15\\xee?\\x99}\\'
'xc6\\xf5\\x99\\xed\\xe7\\xbf\\x1e\\x12\\x03u$\\xf4\\x00@\\x88\\xeb\\xee\\x'
'e19b\\n@\\xae\\x06\\x1a\\xda\\x12+\\xc9?N\\xb8]\\xdbn$\\xf1?\\xca+\\xc3\\'
'xb8\\xcc"\\xef?\\x883\\x8f\\xb1V&\\xf2?\\xea\\x06Iy\\x0f\\x1a\\xe0?wj>|'
'\\x95\\x80\\r@\\xb3\\x1dq\\xa2\\x9b\\xa1\\x04@\\x92\\x80\\x08\\xfd\\x188\\x'
'e6\\xbf\\xc7\\xdfRE\\nC\\xe8\\xbf\\x0e,\\xdb\\xde\\x08\\xb8\\xe5\\xbf\\x8a'
'|\\x9b\\xc4q\\xaf\\xe5\\xbfK\\x19#\\xcb\\xbc\\x93\\xe3\\xbfB\\x15\\xb9M\\x'
'04U\\xb4\\xbf\\x89\\xa1E\\xf4|b\\xcd?\\xc7\\x14\\x1eX\\xff\\x00\\xfd?\\xf'
'4\\x1dc\\x8a$A\\x08@\\xc7\\xdfRE\\nC\\xe8\\xbf\\x8e\\x13\\xd6\\xe5<\\xdf\\'
"xce\\xbfkn*\\x00TD\\xe9?\\x0eD\\xf5\\'y\\x0f\\xe5?O\\x11\\tz;\\xc3\\xe4?"
'\\xad\\xb6\\xaf\\xf9vU\\xea\\xbf[\\xbf\\xa9b[\\xc5\\x01@\\x13\\x9a\\x0f\\x'
'b0F\\xa9\\x0b@\\xb6\\xbc\\xcd\\xd7\\x8e\\x9b\\xfc?N\\xb8]\\xdbn$\\xf1?\\\\'
'\\xe7\\xf1\\x1d\\xc5\\x0e\\xdb?\\xc6p$\\xd3$\\xf6\\xdf?\\xea\\x06Iy\\x0f\\'
'x1a\\xe0?\\x10jD`d\\xff\\xec?\\x05\\xc2\\xc4E6\\xd2\\xfd?,\\xf7PL`7\\x0'
'e@\\xe8-\\xa2\\xaa\\xc6\\xa3\\x07@\\x8e\\x13\\xd6\\xe5<\\xdf\\xce\\xbf>\\x'
'd9\\x80Y\\xa7\\xa3\\xce?2\\x840\\xb3\\x13I\\xc7?\\xe3nn\\x16\\xf1\\xb9\\x'
'd7\\xbf\\xfd\\xd3\\xf2Y\\xc7\\xba\\xc2\\xbf\\x90!\\xeb\\x06n\\xdb\\x05@\\x'
'c4c\\xe0Q\\xf1$\\x03@O\\xb2\\xbf\\xea\\x80\\xca\\x06@\\x06\\x9d\\x08\\xe4'
'\\xc0H\\xeb?\\xf3\\xb1Fq\\x1b_\\xf8?d[\\xbb\\xa4F8\\xef?:\\xb2\\x04\\xc0'
'M\\x06\\xfd?\\xc2\\x85\\x9fP\\xf0_\\xe3?I)\\xdf\\xcf\\x12O\\x08@p\\x8e\\x'
'adE \\xd4\\x07@d\\xf1\\xc8E\\xf3\\xdd\\xe4\\xbf1=\\xd3\\x14>\\xdc\\xf8\\x'
'bf\\xe894}\\xea\\x8d\\xeb\\xbf\\xa20\\xe9G\\x8a\\xbc\\xed\\xbf\\x84\\xc8#'
'\\xdf\\xdd\\x1e\\xcb?\\xa4\\xbf|\\x8c/\\xca\\xf6\\xbf\\x0b}2\\xe2\\xf6\\xf'
'2\\xc6\\xbf\\xad\\x06\\x9f\\xb5\\xd7\\x96\\x00@\\x9c \\x96\\xa0lh\\xfb\\xb'
'f\\x1b\\x89\\xc0\\xaf\\xa9!\\xc9?kn*\\x00TD\\xe9?\\xb4\\xd0\\x86\\x8dW\\x'
'd9\\xfe?\\xd6\\xa7D\\xbf!]\\xf8?\\x11\\xb5\\xc7\\x89\\xb2P\\xd8?\\xc8\\xf'
'8m>\\x85\\xda\\x10@\\x06H\\x01N\\xfd\\x8d\\xcb\\xbf\\xf8\\x17\\xc2\\xb3N\\'
'xa9\\xea?\\xb8\\xebE}\\x0e\\x91\\xda?\\\\\\xe7\\xf1\\x1d\\xc5\\x0e\\xdb?+\\'
'xac"\\xcd\\xc5\\xbb\\x97?\\xea\\x06Iy\\x0f\\x1a\\xe0?\\xde\\xb5\\xbb\\xc1'
'\\x14S\\xf9?\\xa5\\xfc\\xe7\\xa2W\\x14\\x12@\\xe3\\xdf\\x8cw\\x13\\x8d\\x0'
'6@\\x85\\xc5\\x08X\\x03\\xe9\\xfa?N\\xb8]\\xdbn$\\xf1?DS\\xfa\\x14\\xdfo'
"\\xf5?\\xdeJU.$O\\xfc?\\xa3\\xa2\\xe4\\xbe\\x8b\\x08\\xf6?8\\'\\x9a\\x91l"
' \\xdd?u@\\xbb*\\x01\\xaa\\x0f@8\\x91\\x91E\\x8d\\xc5\\xfd?\\x96)8\\xb47'
'D\\xe7?\\x1b\\x89\\xc0\\xaf\\xa9!\\xc9?\\x99\\x03\\xd4\\xa6\\x00\\xe5\\xd3'
'\\xbfK\\x19#\\xcb\\xbc\\x93\\xe3\\xbfB\\x15\\xb9M\\x04U\\xb4\\xbf\\xe47\\x'
'9f+*G\\x03@Z\\x8f\\x9a\\xd8\\xf7\\x9e\\r@8\\xb5\\xbf\\x83\\xe4\\xdd\\x01@'
'\\xe0\\xbc4k\\xe5)\\xda?\\xcc)V\\xd8L\\xf6\\x96\\xbf>\\xd9\\x80Y\\xa7\\xa'
'3\\xce?\\xea\\x03\\xba\\x9c\\xde\\xd5\\xd2\\xbfO\\x11\\tz;\\xc3\\xe4?\\xc2'
'\\x85\\x9fP\\xf0_\\xe3?Z\\x8f\\x9a\\xd8\\xf7\\x9e\\r@\\xd4\\xa1\\xb6d\\x02'
'\\xd9\\x04@\\xae\\x06\\x1a\\xda\\x12+\\xc9?\\x1b\\x89\\xc0\\xaf\\xa9!\\xc9'
'?\\x0e\\x8fw\\xdc\\x11\\xa7\\xac?K\\x19#\\xcb\\xbc\\x93\\xe3\\xbf\\x0b\\xf'
'9\\x11\\xf1\\xc6\\xe1\\xd6?\\xe8\\xf7qX\\xaa/\\xe8?\\x87\\xd6\\x85\\xbdI '
'\\t@\\xb8\\xa5q\\x1am~\\x0f@\\xb6\\xbc\\xcd\\xd7\\x8e\\x9b\\xfc?\\x1b\\x89'
"\\xc0\\xaf\\xa9!\\xc9?\\x0e\\x8fw\\xdc\\x11\\xa7\\xac?\\x0eD\\xf5\\'y\\x0f"
'\\xe5?q\\x9d\\x84\\xbe\\xf5\\xb3\\xf3?\\x1cn\\x0b4\\x8f\\xe7\\xf0?up\\xca'
'\\xb4d\\xd0\\x03@\\xd0\\xfd K\\x1e\\xb3\\xf2?Hz\\xb7\\x0f\\x93\\xf2\\x01\\'
'xc0\\xf1\\xb0P\\x18\\xd8o\\xdd\\xbf\\r\\xb1\\x91G\\xdbe\\xe3?\\x883\\x8f\\'
'xb1V&\\xf2??\\x98$\\xbe__\\xf1?\\xad\\xb6\\xaf\\xf9vU\\xea\\xbf\\x87\\xd'
'6\\x85\\xbdI \\t@0r&\\xbc\\x84\\xeb\\xfc?\\xe0\\xbc4k\\xe5)\\xda?\\x06\\x'
'9d\\x08\\xe4\\xc0H\\xeb?kn*\\x00TD\\xe9?]\\xb9\\xc0\\x10\\x8a\\xb0\\xf4?'
'\\x18&\\x89{\\x93\\x15\\xee?\\xcc|\\xd2\\xbd7\\xeb\\xf6?\\xd6\\xe9\\xe7\\x'
'b3eA\\x0f@X1l\\xa9\\x06F\\x05@\\x8c\\xd5\\xab\\xebR\\x00\\x00@\\x9ci B!'
'\\xdc\\xf1\\xbf\\x0e\\x8fw\\xdc\\x11\\xa7\\xac?\\xb9OX\\xe6\\xdf#\\xea?O\\'
'x11\\tz;\\xc3\\xe4?\\x1a#\\x88\\nA\\x96\\xf5\\xbf\\x98lPP\\x92\\x96\\x02@'
'\\xf0P\\r\\xc7\\x80\\xa9\\xe2?Hz\\xb7\\x0f\\x93\\xf2\\x01\\xc0\\x8e\\x13\\x'
'd6\\xe5<\\xdf\\xce\\xbf>\\xd9\\x80Y\\xa7\\xa3\\xce?\\x08\\xc5#\\xcf\\xf0\\'
'xc4\\xf9?\\x18&\\x89{\\x93\\x15\\xee?\\x11\\xb5\\xc7\\x89\\xb2P\\xd8?Z\\x'
'8f\\x9a\\xd8\\xf7\\x9e\\r@C\\x9cY\\xf2B\\x97\\xd4\\xbf\\xd2L$\\xb5\\tz\\xe'
'0?\\x1b\\x89\\xc0\\xaf\\xa9!\\xc9?\\xb7\\x11Ek\\x1eP\\xc0\\xbf\\xa20\\xe9'
'G\\x8a\\xbc\\xed\\xbf\\xea\\x06Iy\\x0f\\x1a\\xe0?B\\xe0\\xdd;I\\xb7\\xf5?'
'Z\\x8f\\x9a\\xd8\\xf7\\x9e\\r@*I\\xb4}\\xc5<\\x11@"\\xd7~X\\xec\\x83\\xf7'
'?\\xb8\\xebE}\\x0e\\x91\\xda?\\xca+\\xc3\\xb8\\xcc"\\xef?\\x05t&\\x85\\x1'
'2\\x81\\x04@\\x82c\\xc2\\xe0R,\\x03@I\\xb8\\x97i;Z\\xcc\\xbf[\\xbf\\xa9b'
'[\\xc5\\x01@\\xcd\\xab\\x0b\\x85\\xb9p\\x07\\xc0\\x1a\\x91\\xb4\\x06\\x03v'
'\\x11\\xc0\\x8e\\x13\\xd6\\xe5<\\xdf\\xce\\xbf\\x94\\xf4\\xad\\xb8\\xa2\\x8'
'0\\xf2?\\xb4\\xd0\\x86\\x8dW\\xd9\\xfe?\\xd6\\xa7D\\xbf!]\\xf8?\\x982C\\x'
'ccK\\x9c\\xdc\\xbf\\xce\\xce\\x91\\xf4\\xa4\\xac\\x06@\\xe0\\x127\\xb4o\\x'
'f5\\xe0?;2\\x0c\\xa1U\\x03\\xf8\\xbf\\x1b\\x89\\xc0\\xaf\\xa9!\\xc9?\\x0e'
'\\x8fw\\xdc\\x11\\xa7\\xac?+\\xac"\\xcd\\xc5\\xbb\\x97?\\xd6A\\xf7\\x8b$\\'
'x86\\xe0\\xbf$\\xa3-dAg\\xef?&-Y4\\xe5\\x88\\t@/If\\x12\\xc5y\\xe9?%Q\\'
'x8e\\xab\\x98\\xd9\\x00@\\x9ci B!\\xdc\\xf1\\xbfkn*\\x00TD\\xe9?+\\xac"'
'\\xcd\\xc5\\xbb\\x97?O^b\\xe0\\xbc\\xd7\\x00@\\x0c\\x0b\\xfa\\xd1\\xae\\x9'
"c\\x02\\xc0\\'\\x1eW\\xfa=\\xf2\\xd1?\\x92\\x80\\x08\\xfd\\x188\\xe6\\xbfo"
'EU\\x0e\\xd8!\\x0c\\xc0x_\\xc3\\x80\\xa8$\\xff?\\xb0Eb\\x9f\\xa4\\x85\\x0'
'3@\\xdeJU.$O\\xfc?\\x08\\xad\\xa4\\xbf\\xb7\\xb1\\xfa?\\xde\\xb5\\xbb\\xc'
'1\\x14S\\xf9?\\x7f\\xec\\x14\\x92-A\\xee\\xbf}-\\x15\\x1d\\x9b\\x1a\\xdb?'
'Hz\\xb7\\x0f\\x93\\xf2\\x01\\xc0\\xe3\\x8b\\x10\\xae\\x8b$\\xf8?\\x00\\xe7'
'\\x15Ch\\x96\\x00@3?\\xf2o\\xbd:\\xf7?\\x0b\\xf9\\x11\\xf1\\xc6\\xe1\\xd6'
'?.\\xa7\\xf47lO\\xf3?\\xb3\\x85\\xf2Z\\xe0\\x07\\xf4\\xbf\\xbc\\x81m\\xc1'
'\\xe0#\\xc4?\\xf2\\xdf\\xb9\\xd8`\\xd1\\xf5?x_\\xc3\\x80\\xa8$\\xff?Ro\\x'
'df)\\x94=\\xfe?]\\xb9\\xc0\\x10\\x8a\\xb0\\xf4?\\x18&\\x89{\\x93\\x15\\xe'
'e?\\xcc|\\xd2\\xbd7\\xeb\\xf6?yhC\\x1a\\r\\xd1\\xc6?\\xd2xaH]\\x08\\xed\\'
'xbf\\x8c\\xd5\\xab\\xebR\\x00\\x00@\\x0e,\\xdb\\xde\\x08\\xb8\\xe5\\xbf\\x'
'0e\\x8fw\\xdc\\x11\\xa7\\xac?d[\\xbb\\xa4F8\\xef?O\\x11\\tz;\\xc3\\xe4?\\'
"xc7\\xbc\\xff\\x90\\xb7\\x02\\x02\\xc0\\'\\x1eW\\xfa=\\xf2\\xd1?\\x84\\x94"
'\\xb2=\\x87g\\xdb\\xbf\\xe6\\xb3\\x92W\\x1aN\\xfe?\\x06\\x9d\\x08\\xe4\\xc'
'0H\\xeb?\\x00\\xe7\\x15Ch\\x96\\x00@\\xdeJU.$O\\xfc?\\xd6\\xa7D\\xbf!]\\'
"xf8?\\xcc|\\xd2\\xbd7\\xeb\\xf6?\\x0f=\\'\\x1e\\xd1\\xc0\\xdb?\\xc1\\x11\\"
'xcb\\x8b\\x8c\\xe2\\xf7\\xbf\\xdc_\\x06\\x8f5\\n\\x07\\xc0\\xcc)V\\xd8L\\x'
'f6\\x96\\xbf\\xb7\\x11Ek\\x1eP\\xc0\\xbf\\\'\\xd9\\xe7?"Z\\xc1\\xbf\\xa3\\x'
'a2\\xe4\\xbe\\x8b\\x08\\xf6?\\xb8C\\xe9\\xb9Z\\x83\\xf4?\\xfb\\xe6gO\\xae'
'|\\xcd\\xbfP\\xc5\\x128\\xe7\\xe1\\xec?oEU\\x0e\\xd8!\\x0c\\xc0\\x1b\\x89'
'\\xc0\\xaf\\xa9!\\xc9?DS\\xfa\\x14\\xdfo\\xf5?\\x1a\\xb1\\x8d\\xd5\\xf8;\\'
'x03@:\\xb2\\x04\\xc0M\\x06\\xfd?\\x06(\\x8e\\xc9\\xce"\\xfe?\\x1f\\xa3\\x'
'f4!\\xa3\\x03\\xf4?\\x81\\x19r@H\\x12\\xf1\\xbf[\\x06L\\xb3e\\x0e\\xee?\\'
'x8e\\x13\\xd6\\xe5<\\xdf\\xce\\xbfDS\\xfa\\x14\\xdfo\\xf5?\\x883\\x8f\\xb'
"1V&\\xf2?q\\x9d\\x84\\xbe\\xf5\\xb3\\xf3?.\\xa7\\xf47lO\\xf3?\\x0f=\\'\\x"
'1e\\xd1\\xc0\\xdb?\\xa9\\xb4\\x89\\xefrT\\xf5\\xbf\\x08\\xc17\\xce\\x1d\\x'
'fb\\x0c\\xc068\\xc1GxR\\x08@\\x00\\xe7\\x15Ch\\x96\\x00@\\xdeJU.$O\\xfc'
'?O^b\\xe0\\xbc\\xd7\\x00@n\\xd4\\x93\\xad\\x18{\\x04@\\xd6\\x81f\\xfb \\x'
'db\\xef?C\\x9cY\\xf2B\\x97\\xd4\\xbfM\\x9a\\xe4\\xb7m-\\xb7?\\x1c\\xedm|'
'\\x7f\\x12\\n@X\\x16<q\\x06\\x0e\\x02@\\xb4\\xd0\\x86\\x8dW\\xd9\\xfe?\\x0'
'8\\xad\\xa4\\xbf\\xb7\\xb1\\xfa?c1Ux\\xb8\\x18\\x0b@Cok3m\\xa3\\xfe?^\\x'
'03\\xe9J\\xf8\\x95\\xee?\\xb86\\xdd*;\\xf1\\x05@q\\xc9U\\x11\\xa4H\\xe4?'
'\\xf3\\xb1Fq\\x1b_\\xf8?\\x08\\xc5#\\xcf\\xf0\\xc4\\xf9?q\\x9d\\x84\\xbe\\'
'xf5\\xb3\\xf3?\\xb8C\\xe9\\xb9Z\\x83\\xf4? \\x13\\xfdt?i\\xc0\\xbf\\x19b'
'\\xd3\\x07\\x82\\xe7\\x01\\xc0\\xe6\\xce\\xa5\\xe6\\xa5\\xe2\\x11\\xc0q\\xc'
'9U\\x11\\xa4H\\xe4?\\xa2\\x10\\x93\\xcdWN\\xfb?\\xdeJU.$O\\xfc?\\x08\\xa'
'd\\xa4\\xbf\\xb7\\xb1\\xfa?\\xf2\\xee\\xa4\\xc5\\xf1\\xba\\xfb?\\x05\\xc2\\'
'xc4E6\\xd2\\xfd?\\x1f\\x0b\\x90\\xff\\xb3\\xc5\\xe7?\\xb86\\xdd*;\\xf1\\x'
'05@\\x1b\\x89\\xc0\\xaf\\xa9!\\xc9?DS\\xfa\\x14\\xdfo\\xf5?]\\xb9\\xc0\\x'
'10\\x8a\\xb0\\xf4?\\x08\\xad\\xa4\\xbf\\xb7\\xb1\\xfa?\\xde\\xb5\\xbb\\xc1'
'\\x14S\\xf9?m\\xbd\\xe3\\xb2\\xfb\\xae\\xe7?\\xd9n\\x0c(\\xa6p\\xfa\\xbfC'
'\\xe4#\\xcf\\xef0\\x06\\xc0q\\xc9U\\x11\\xa4H\\xe4?\\x99\\x03\\xd4\\xa6\\x'
'00\\xe5\\xd3\\xbfK\\x19#\\xcb\\xbc\\x93\\xe3\\xbf?\\x98$\\xbe__\\xf1?r\\x'
'c0p\\xc4\\x91\\xcc\\xd7\\xbf\\xe8\\x171\\x8eiQ\\xe9?\\x00\\xb8\\xa3\\x83Q'
'\\xcf\\xf7?\\xachA\\x0f\\xaaW\\x05\\xc0\\xb8\\xebE}\\x0e\\x91\\xda?\\xca+'
'\\xc3\\xb8\\xcc"\\xef?d[\\xbb\\xa4F8\\xef?\\xea\\x06Iy\\x0f\\x1a\\xe0?=\\'
'x94\\xde\\xaa\\x18\\x84p?\\x10\\r*\\x8fZ\\x8d\\xfa?\\x0f\\xf6y\\x96b\\x83'
'\\xf9?\\x80\\xa9\\x84j\\x0c}\\x08@\\xad\\xf5i\\x17\\x9a\\xa4\\xfb?\\xca+\\'
'xc3\\xb8\\xcc"\\xef?2\\x840\\xb3\\x13I\\xc7?;L\\xb7\\x8cP/\\xe5\\xbf)\\x'
"86\\x99l!\\xe1\\x03@\\'\\x1eW\\xfa=\\xf2\\xd1?\\xbf\\xbfJ8\\r\\xff\\xf0?U"
'HS+$\\x8c\\x02@\\xad\\xf5i\\x17\\x9a\\xa4\\xfb?\\xf3\\xb1Fq\\x1b_\\xf8?\\'
"x0eD\\xf5\\'y\\x0f\\xe5??\\x98$\\xbe__\\xf1?\\xc7[w\\xf2\\xec|\\x07@\\xd"
'e\\xcc\\xcbDE\\x96\\xec?\\x16}\\xd9\\xaf\\xc1t\\xac\\xbfk)\\xd1 \\xe1\\xb'
'5\\xf9\\xbfq\\xc9U\\x11\\xa4H\\xe4?kn*\\x00TD\\xe9?+\\xac"\\xcd\\xc5\\xb'
'b\\x97?\\xa0Vw\\x8d|\\xd8\\xe9\\xbf\\x8b\\xb0\\xbb\\xe6UE\\x00@u\\x08I\\x'
'fc\\x1fj\\xe4?c\\x18\\x06\\x18e\\xff\\xd7\\xbfUHS+$\\x8c\\x02@\\xfc\\xa6'
',~L\\\\\\xfc\\xbf7\\x89~\\x10l\\xa0\\x02\\xc0\\\'\\xd9\\xe7?"Z\\xc1\\xbf\\xb'
'4\\x1b\\xc9zgl\\xe9?\\xb7\\xf8e\\x90\\x0c2\\xf9\\xbf\\x05\\x92\\xc7\\xb6\\'
'xbf\\x9e\\xec\\xbf\\x0f\\xaf\\x9en\\xe9\\xda\\xbb\\xbf\\x14\\xc6V#R\\xda\\'
'x8f\\xbf\\xad\\xf5i\\x17\\x9a\\xa4\\xfb?\\xca+\\xc3\\xb8\\xcc"\\xef?2\\x8'
'40\\xb3\\x13I\\xc7?\\xd6A\\xf7\\x8b$\\x86\\xe0\\xbf\\x982C\\xccK\\x9c\\xd'
'c\\xbf\\x05\\x92\\xc7\\xb6\\xbf\\x9e\\xec\\xbf\\x99v\\xb3\\xdca\\xa0\\xf3\\'
'xbf\\xd4K\\xba@\\x02$\\xbf\\xbfl\\xceg\\xdei\\xd2\\x04@_\\xa4\\xae\\xfb\\'
'xe0t\\x06@\\xdeJU.$O\\xfc?\\xb4\\x1b\\xc9zgl\\xe9?\\xfd\\xd3\\xf2Y\\xc7'
'\\xba\\xc2\\xbf\\xaa:\\x8d\\x11\\xbcL\\xf7\\xbf\\xb2\\xfc\\xb4";\\xa0\\xe9'
'\\xbf\\xc7\\xdfRE\\nC\\xe8\\xbf1=\\xd3\\x14>\\xdc\\xf8\\xbf\\xe0YX\\xe2\\x'
'cd(\\x01\\xc0@\\x1b\\x80\\x19\\xac\\xfe\\xdc\\xbf\\xd6\\xa7D\\xbf!]\\xf8?'
'KN\\x9e\\xbc\\xd7\\xfc\\xd2\\xbf\\xfa}\\xfe\\x91;\\x94\\xf1\\xbf\\xf9y\\xc'
'6\\x0c%\\xf4\\xca?\\xae\\x06\\x1a\\xda\\x12+\\xc9?1=\\xd3\\x14>\\xdc\\xf8'
'\\xbf\\x94\\xf4\\xad\\xb8\\xa2\\x80\\xf2?oY^VW\\xcd\\xd5?\\x0b\\xf9\\x11\\'
'xf1\\xc6\\xe1\\xd6?\\xca1O\\x94\\xe9\\x99\\xfb\\xbf s\\xf7\\x92,\\xd0\\xe'
'2\\xbf\\xd2xaH]\\x08\\xed\\xbfs\\xdf\\x84\\x1e]&\\xcd\\xbf\\xa3\\xff\\x8d'
'\\xb1%\\xb8\\xec\\xbf\\xb1\\x97\\xcb\\xaf\\xe6\\x83\\xf9\\xbf\\xf6$\\x86\\x'
'89#\\xa8\\xe8\\xbfq\\x9d\\x84\\xbe\\xf5\\xb3\\xf3?_\\xd2\\n\\xea\\x02\\xb6'
'\\xe0\\xbf\\xaa:\\x8d\\x11\\xbcL\\xf7\\xbf\\x0f\\xaf\\x9en\\xe9\\xda\\xbb\\'
'xbf\\xd2L$\\xb5\\tz\\xe0?\\xf1\\xb0P\\x18\\xd8o\\xdd\\xbf\\x8a|\\x9b\\xc4'
'q\\xaf\\xe5\\xbf\\xf6$\\x86\\x89#\\xa8\\xe8\\xbf~\\xca\\xfbH\\xc2\\xe7\\xf'
'a\\xbf\\x11\\xb5\\xc7\\x89\\xb2P\\xd8?\\xf1b\\x96\\xd7\\x8d\\x0c\\xe6?@\\x'
'87<%\\xd6-\\xeb?I\\x1c,\\xaa\\xdd\\x08\\x0b@\\x0e,\\xdb\\xde\\x08\\xb8\\x'
'e5\\xbf\\x0e\\x8fw\\xdc\\x11\\xa7\\xac?K\\x19#\\xcb\\xbc\\x93\\xe3\\xbf\\x'
'b4\\xb5{Gj\\x95\\xf1\\xbfs\\x0b\\xf4\\xed\\xdf\\x1d\\xe3\\xbfdr~i\\xd7\\x'
'f3\\xea?\\xdf;\\xf7]/g\\xf4?\\xe2\\x97\\x0ej#\\xe2\\x0b@\\x9ci B!\\xdc\\'
'xf1\\xbf\\x029\\x7fS\\xaa\\x94\\xf6\\xbfR\\xbb\\x1d_y\\x1b\\xfe\\xbf\\xb0'
'\\xcf[IX<\\xfd\\xbfT\\xceC\\x16\\xd8\\xcd\\xfc\\xbf\\x13\\xff\\x0fZ\\xe1x'
"\\xff\\xbf\\xc906\\x15\\x95\\xbc\\xf8\\xbfvU\\xf8\\xa1\\'9\\xf1\\xbf\\x1b\\"
'x89\\xc0\\xaf\\xa9!\\xc9?\\x8a|\\x9b\\xc4q\\xaf\\xe5\\xbf\\xfc\\xa3W\\xe2'
'\\xab\\xf2\\xf3\\xbfK\\xc5\\x9bH,\\x93\\xf8\\xbf\\xd3(\\x82\\x011%\\xef\\x'
'bfl\\x8d\\xe6#\\x85{\\xf6\\xbf\\x92\\x80\\x08\\xfd\\x188\\xe6\\xbfd\\xf1\\'
'xc8E\\xf3\\xdd\\xe4\\xbff\\xd3y\\xab/\\\\\\xf5\\xbf\\xe0YX\\xe2\\xcd(\\x01'
'\\xc0}5\\xec\\xffE\\x91\\xfb\\xbf\\x19\\xc0;H\\x96>\\xf6\\xbf\\xf3\\xb0\\x'
'b5\\x02\\x87\\xc6\\xf0\\xbf\\x05\\x92\\xc7\\xb6\\xbf\\x9e\\xec\\xbf\\x99v\\'
'xb3\\xdca\\xa0\\xf3\\xbf\\x03\\x03?F\\xdcx\\xe1\\xbf\\x18"\\xb7D}\\xa4\\x'
'f4?>\\xd9\\x80Y\\xa7\\xa3\\xce?@\\x1b\\x80\\x19\\xac\\xfe\\xdc\\xbf\\xe7\\'
'xba\\xdbG\\x00\\xea\\xf3\\xbf\\x982C\\xccK\\x9c\\xdc\\xbfu\\xd8Km\\xa96\\'
'xf3\\xbf\\xb2\\xfc\\xb4";\\xa0\\xe9\\xbf\\xd4K\\xba@\\x02$\\xbf\\xbff\\xd'
'3y\\xab/\\\\\\xf5\\xbf\\x8a|\\x9b\\xc4q\\xaf\\xe5\\xbf\\xb9OX\\xe6\\xdf#\\x'
'ea?O^b\\xe0\\xbc\\xd7\\x00@\\x82n\\x05P\\xc0h\\x01\\xc0\\x89\\xfe\\r\\xad'
'UG\\xc0?\\x10Ch~y\\r\\x01\\xc0\\xachA\\x0f\\xaaW\\x05\\xc0\\xe3\\x8b\\x10'
'\\xae\\x8b$\\xf8?\\xb0Eb\\x9f\\xa4\\x85\\x03@\\x1a\\xb1\\x8d\\xd5\\xf8;\\x'
'03@\\x19s\\xe2\\xe1\\x14*\\n@\\xde\\xb5\\xbb\\xc1\\x14S\\xf9?.\\xe0?6N\\x'
'aa\\xf5\\xbf\\xa3\\x10_c\\xa9\\xcf\\xde\\xbfs\\xdf\\x84\\x1e]&\\xcd\\xbf\\'
'xb8\\xebE}\\x0e\\x91\\xda?\\x0e\\x8fw\\xdc\\x11\\xa7\\xac?\\xb4\\xd0\\x86'
'\\x8dW\\xd9\\xfe?\\x82c\\xc2\\xe0R,\\x03@_\\xd2\\n\\xea\\x02\\xb6\\xe0\\xb'
'fm\\xbd\\xe3\\xb2\\xfb\\xae\\xe7?\\xe2\\x8dw\\xb1\\xaeJ\\xfb\\xbf\\xf8\\x1'
'7\\xc2\\xb3N\\xa9\\xea?\\x8e\\x13\\xd6\\xe5<\\xdf\\xce\\xbfDS\\xfa\\x14\\x'
'dfo\\xf5?3?\\xf2o\\xbd:\\xf7?q\\x9d\\x84\\xbe\\xf5\\xb3\\xf3?\\x06(\\x8e'
'\\xc9\\xce"\\xfe?)V\\xdb#\\x0c%\\x00\\xc0\\x84\\x94\\xb2=\\x87g\\xdb\\xbf'
"C)j\\x8d\\x8a\\'\\xdc\\xbf\\xe3\\x8b\\x10\\xae\\x8b$\\xf8?\\x07u\\x88\\xcd"
'B\\xfd\\x04@\\x05t&\\x85\\x12\\x81\\x04@\\x19s\\xe2\\xe1\\x14*\\n@B\\xe0\\'
'xdd;I\\xb7\\xf5?\\xc8\\xac\\xae\\x9a\\xa7\\x8d\\x00\\xc0\\xe2\\xb67[n\\xb'
'c\\xee\\xbf\\x8a\\xbcfD8\\r\\xef\\xbfN\\xb8]\\xdbn$\\xf1?\\xb0Eb\\x9f\\xa'
'4\\x85\\x03@/\\xee\\xf4%\\xdf\\xf6\\x01@\\x82c\\xc2\\xe0R,\\x03@:\\xbd\\x'
'a0\\xe4\\x08\\xc3\\xc3?Um\\xb7~\\xfa\\x7f\\x07\\xc0\\xd9n\\x0c(\\xa6p\\xf'
'a\\xbf\\xe1\\xf5\\x99\\xcf\\xd8\\xcb\\x02\\xc0\\xa1d\\x0eu[R\\x01@\\xb7\\x'
'd3\\xd4)\\x7f\\xec\\x07@\\xf06\\xbf4,\\xc6\\x05@MkR\\xe13\\xab\\x06@\\x1'
'cn\\x0b4\\x8f\\xe7\\xf0?\\x1f\\x0bv\\xda\\xe7i\\x03\\xc0\\xc1\\x11\\xcb\\x'
'8b\\x8c\\xe2\\xf7\\xbf\\xa8L\\xbd!\\xb3\\xeb\\xf2\\xbf\\xa3\\xff\\x8d\\xb1'
'%\\xb8\\xec\\xbf\\xa3{\\xe6\\x9a1\\xb6\\xf0\\xbf\\xea\\x03\\xba\\x9c\\xde\\'
'xd5\\xd2\\xbfq\\x9d\\x84\\xbe\\xf5\\xb3\\xf3?\\xd6\\xbe\\x88T\\xcd\\xc7\\x'
'e5?3)\\xb1\\x7f<{\\xb3?\\xd9n\\x0c(\\xa6p\\xfa\\xbf\\x9c \\x96\\xa0lh\\x'
'fb\\xbf\\xf1\\xb0P\\x18\\xd8o\\xdd\\xbfDS\\xfa\\x14\\xdfo\\xf5?\\x08\\xc5'
'#\\xcf\\xf0\\xc4\\xf9?q\\x9d\\x84\\xbe\\xf5\\xb3\\xf3?\\x11\\xb5\\xc7\\x89'
'\\xb2P\\xd8?\\xd9\\x12j\\xa3\\x8c\\xdd\\x05\\xc0\\xd1O\\xa1\\x9e\\x9d\\x96'
'\\xf9\\xbf;2\\x0c\\xa1U\\x03\\xf8\\xbf1=\\xd3\\x14>\\xdc\\xf8\\xbf\\xa3{\\'
"xe6\\x9a1\\xb6\\xf0\\xbf\\x0eD\\xf5\\'y\\x0f\\xe5?\\xd6\\xa7D\\xbf!]\\xf8"
'?@\\x95Z\\x12\\xfbe\\xfa\\xbfi\\xd2x\\x87\\xc4Z\\xcd?\\n)\\x8f`\\xd9\\x8c'
'\\xff\\xbf\\x18\\x83\\xf2\\x8f\\x07@\\x00\\xc0\\xb8\\xebE}\\x0e\\x91\\xda?'
'\\x00\\xe7\\x15Ch\\x96\\x00@\\xdeJU.$O\\xfc?\\x1bf\\xf2\\xe0\\x9dV\\x04@'
"\\xd0\\xfe\\xb5\\'M\\xdf\\x00@)V\\xdb#\\x0c%\\x00\\xc0a\\xc6\\x85\\xc4\\xe"
'5\\x1b\\xe1\\xbf\\x9c \\x96\\xa0lh\\xfb\\xbff\\xd3y\\xab/\\\\\\xf5\\xbf\\x0'
'e\\x8fw\\xdc\\x11\\xa7\\xac?d[\\xbb\\xa4F8\\xef?m\\xb7d\\xc0\\xe3Z\\xff?'
'\\xa4\\xbf|\\x8c/\\xca\\xf6\\xbf\\x1f\\xd3\\xf1\\xb0\\x197\\xd5?\\x898\\xd'
'd\\xc9P\\xec\\xf1\\xbf\\xf8\\x17\\xc2\\xb3N\\xa9\\xea?\\xcc)V\\xd8L\\xf6\\'
'x96\\xbfDS\\xfa\\x14\\xdfo\\xf5?\\xb9OX\\xe6\\xdf#\\xea?q\\x9d\\x84\\xbe'
'\\xf5\\xb3\\xf3?\\xfd\\xd3\\xf2Y\\xc7\\xba\\xc2\\xbf\\xf3\\xc3\\x8a\\xf5\\x'
'95\\xe8\\x07\\xc0\\x95\\xd2\\x1d\\xc3}z\\x01\\xc09\\xb8\\xfcM\\xa9\\xad\\x'
'0e\\xc0\\x0e,\\xdb\\xde\\x08\\xb8\\xe5\\xbf\\xf3\\xb1Fq\\x1b_\\xf8?3?\\xf'
'2o\\xbd:\\xf7?\\xa3\\xa2\\xe4\\xbe\\x8b\\x08\\xf6?\\x10jD`d\\xff\\xec?\\x'
'fc\\x0e\\xf0>\\xba\\xa3\\x04\\xc0\\x81\\x19r@H\\x12\\xf1\\xbfzq|\\x8f\\x1'
'e\\xa5\\x03\\xc0x_\\xc3\\x80\\xa8$\\xff?\\x07u\\x88\\xcdB\\xfd\\x04@/\\xe'
'e\\xf4%\\xdf\\xf6\\x01@MkR\\xe13\\xab\\x06@|\\x8b\\x99G\\xe0\\xee\\xfc?Y'
'\\xf7\\x1b\\x91<\\x05\\xfd\\xbfc\\x18\\x06\\x18e\\xff\\xd7\\xbfC)j\\x8d\\x'
"8a\\'\\xdc\\xbf\\xf1\\xb0P\\x18\\xd8o\\xdd\\xbf>\\xd9\\x80Y\\xa7\\xa3\\xce"
'?3?\\xf2o\\xbd:\\xf7?\\x82c\\xc2\\xe0R,\\x03@\\x06\\xea\\x9e\\x06d.\\xf3'
'\\xbf\\xf1b\\x96\\xd7\\x8d\\x0c\\xe6?\\n)\\x8f`\\xd9\\x8c\\xff\\xbf\\xc7\\x'
'dfRE\\nC\\xe8\\xbfN\\xb8]\\xdbn$\\xf1?\\xb7\\xd3\\xd4)\\x7f\\xec\\x07@q\\'
'xc8SR\\xc6d\\r@\\xe6m\\x82\\xe1~\\xd5\\x07@$\\xa3-dAg\\xef?\\xd5Qil\\xa'
"a\\xa7\\xfe\\xbf\\x84\\x94\\xb2=\\x87g\\xdb\\xbfC)j\\x8d\\x8a\\'\\xdc\\xbf"
'\\xb8\\xebE}\\x0e\\x91\\xda?X\\x16<q\\x06\\x0e\\x02@\\xb4\\xd0\\x86\\x8dW'
'\\xd9\\xfe?\\xe9`\\x92\\xe0\\x07\\x02\\x02@\\x1cn\\x0b4\\x8f\\xe7\\xf0?\\x'
'd9\\x12j\\xa3\\x8c\\xdd\\x05\\xc0y\\xfa\\x06\\xb7?8\\xf0\\xbfHz\\xb7\\x0f'
'\\x93\\xf2\\x01\\xc0\\xb8\\xebE}\\x0e\\x91\\xda?\\xca+\\xc3\\xb8\\xcc"\\xe'
'f?oY^VW\\xcd\\xd5?\\xe9`\\x92\\xe0\\x07\\x02\\x02@\\xb8C\\xe9\\xb9Z\\x83'
'\\xf4? s\\xf7\\x92,\\xd0\\xe2\\xbf\\xf2\\xcbM\\xc4\\xbf\\xfe\\xfc\\xbf\\xd'
'4K\\xba@\\x02$\\xbf\\xbf\\x8e\\x13\\xd6\\xe5<\\xdf\\xce\\xbf\\\\\\xe7\\xf1\\'
'x1d\\xc5\\x0e\\xdb?]\\xb9\\xc0\\x10\\x8a\\xb0\\xf4?\\x08\\xad\\xa4\\xbf\\x'
'b7\\xb1\\xfa?\\x82n\\x05P\\xc0h\\x01\\xc0\\x16(\\x92I\\x08\\x15\\xe6\\xbf'
'ay\\x97\\xdc\\xce\\x91\\t\\xc0Hz\\xb7\\x0f\\x93\\xf2\\x01\\xc0\\x06\\x9d\\x'
'08\\xe4\\xc0H\\xeb?X\\x16<q\\x06\\x0e\\x02@3?\\xf2o\\xbd:\\xf7?m\\xb7d\\'
'xc0\\xe3Z\\xff?\\x86D\\xdd\\xf1\\xbc\\x85\\xe5\\xbf)&\\xcc\\x99\\xa8\\xfe'
'\\x0b\\xc0\\x95\\xd2\\x1d\\xc3}z\\x01\\xc0\\xb5\\xd7\\xe0f\\x1a0\\x10\\xc0'
'\\xe3\\x8b\\x10\\xae\\x8b$\\xf8?_\\xa4\\xae\\xfb\\xe0t\\x06@\\xf06\\xbf4,'
'\\xc6\\x05@\\xb4h"\\xe1\\xe8\\x80\\x05@\\xb8C\\xe9\\xb9Z\\x83\\xf4?\\xd9\\'
'x12j\\xa3\\x8c\\xdd\\x05\\xc0\\xe2\\xb67[n\\xbc\\xee\\xbfHz\\xb7\\x0f\\x9'
'3\\xf2\\x01\\xc0q\\xc9U\\x11\\xa4H\\xe4?\\xf3\\xb1Fq\\x1b_\\xf8?/\\xee\\x'
'f4%\\xdf\\xf6\\x01@O^b\\xe0\\xbc\\xd7\\x00@\\x89\\xa1E\\xf4|b\\xcd?\\x06'
'ZU\\x88\\xde^\\x01\\xc0\\xb9\\xf2_\\x02\\x84\\x08\\xf7\\xbf\\xachA\\x0f\\x'
'aaW\\x05\\xc0\\x0e,\\xdb\\xde\\x08\\xb8\\xe5\\xbfkn*\\x00TD\\xe9?\\xb9OX'
'\\xe6\\xdf#\\xea?\\x18&\\x89{\\x93\\x15\\xee?\\xa4\\xbf|\\x8c/\\xca\\xf6\\'
'xbf\\xbch\\xd6\\xeb\\xbf\\\\\\x01@(w\\xe9p\\xd3)\\x00@\\xd2L$\\xb5\\tz\\xe'
'0?\\x06\\x9d\\x08\\xe4\\xc0H\\xeb?\\r\\xb1\\x91G\\xdbe\\xe3?2\\x840\\xb3\\'
'x13I\\xc7?\\x84\\xc8#\\xdf\\xdd\\x1e\\xcb?\\xbc\\xb88\\xbd\\x8c\\x1a\\x0e'
'@\\x89gwj\\xc8/\\xfc?\\x1d\\xb9\\x0f\\xac4\\xe2\\xd0?}LV\\x8e\\\\]\\xd5\\x'
'bf\\xb8\\xebE}\\x0e\\x91\\xda?>\\xd9\\x80Y\\xa7\\xa3\\xce?+\\xac"\\xcd\\x'
'c5\\xbb\\x97?\\xa0Vw\\x8d|\\xd8\\xe9\\xbf\\xd9\\x94`\\xf6\\xc9\\xe4\\t@\\x'
'1f\\xa3\\xf4!\\xa3\\x03\\xf4?\\x84\\x94\\xb2=\\x87g\\xdb\\xbf\\x96)8\\xb4'
'7D\\xe7?f\\xd3y\\xab/\\\\\\xf5\\xbf\\x029\\x7fS\\xaa\\x94\\xf6\\xbf\\xa7\\x'
'af\\xba\\xa0\\x12\\x07\\xf9\\xbf;L\\xb7\\x8cP/\\xe5\\xbf\\xb8C\\xe9\\xb9Z'
'\\x83\\xf4?\\x17\\x88\\x8cg\\xf5{\\xd8?\\xfc<\\xce\\x13-\\x1b\\t@\\xe8-\\x'
'a2\\xaa\\xc6\\xa3\\x07@\\x1b\\x89\\xc0\\xaf\\xa9!\\xc9?\\x8a|\\x9b\\xc4q\\'
'xaf\\xe5\\xbf\\xa20\\xe9G\\x8a\\xbc\\xed\\xbf\\xe7\\xba\\xdbG\\x00\\xea\\x'
'f3\\xbf=\\xbf\\x82p\\xfeH\\x06@\\xde\\xcc\\xcbDE\\x96\\xec?\\xd0\\xfd K\\'
'x1e\\xb3\\xf2?[\\x06L\\xb3e\\x0e\\xee?q\\xc9U\\x11\\xa4H\\xe4?\\xb7\\x11'
'Ek\\x1eP\\xc0\\xbfK\\x19#\\xcb\\xbc\\x93\\xe3\\xbf;L\\xb7\\x8cP/\\xe5\\xb'
'fc1Ux\\xb8\\x18\\x0b@)\\xeeYk\\xc7\\xbe\\xf0?\\x17\\x15\\xe5\\x1fk]\\xfa'
'?\\x14\\xc6V#R\\xda\\x8f\\xbfq\\xc9U\\x11\\xa4H\\xe4?\\x99\\x03\\xd4\\xa6'
'\\x00\\xe5\\xd3\\xbf@\\x1b\\x80\\x19\\xac\\xfe\\xdc\\xbf\\x04a7\\x8e\\xa8\\'
'x81\\xee\\xbfc1Ux\\xb8\\x18\\x0b@\\xa5H\\xa7F5a\\xf2?\\xefy\\xcdp@\\x1b'
'\\xf6?[\\x06L\\xb3e\\x0e\\xee?1=\\xd3\\x14>\\xdc\\xf8\\xbfa\\xf6\\x17\\x0'
'c#s\\xfc\\xbf\\xd2)\\x89A\\xdf|\\xf6\\xbf;L\\xb7\\x8cP/\\xe5\\xbf\\x9f=\\'
'x99"\\x1a\\xd6\\x0c\\xc07+\\xa5\\x7fre\\xf2\\xbf\\x1d\\xb9\\x0f\\xac4\\xe'
'2\\xd0?\\x14\\xc6V#R\\xda\\x8f\\xbfq\\xc9U\\x11\\xa4H\\xe4?\\xb7\\x11Ek\\'
"x1eP\\xc0\\xbf\\x0eD\\xf5\\'y\\x0f\\xe5?\\xe3nn\\x16\\xf1\\xb9\\xd7\\xbfs"
'\\x0b\\xf4\\xed\\xdf\\x1d\\xe3\\xbfOU\\x19\\x956\\x9f\\x99?\\xd2xaH]\\x08'
'\\xed\\xbf\\xe8-\\xa2\\xaa\\xc6\\xa3\\x07@f\\xd3y\\xab/\\\\\\xf5\\xbf\\xb1\\'
"x97\\xcb\\xaf\\xe6\\x83\\xf9\\xbf\\x94\\xa0\\'_\\xd6R\\x00\\xc0o\\xf7\\xcd"
'%\\xee\\x9b\\x05\\xc0\\x99}\\xc6\\xf5\\x99\\xed\\xe7\\xbf\\xfbFbm\\x9b\\xe'
'3\\xef\\xbf\\x81\\x19r@H\\x12\\xf1\\xbf}LV\\x8e\\\\]\\xd5\\xbf\\xfc\\xa6,~'
'L\\\\\\xfc\\xbf\\xe0YX\\xe2\\xcd(\\x01\\xc0R\\xbb\\x1d_y\\x1b\\xfe\\xbf\\xa'
'4\\xef=%\\r\\x1d\\x02\\xc0\\xd3(\\x82\\x011%\\xef\\xbf\\x92\\x82\\xdf$v\\x'
'b7\\xe7\\xbfr\\x04\\\\\\xd7\\xf6\\xcf\\xe2\\xbfM\\x9a\\xe4\\xb7m-\\xb7?\\x9'
'ci B!\\xdc\\xf1\\xbf\\x029\\x7fS\\xaa\\x94\\xf6\\xbf\\xa7\\xaf\\xba\\xa0\\'
'x12\\x07\\xf9\\xbf\\xb0\\xcf[IX<\\xfd\\xbfI\\xb8\\x97i;Z\\xcc\\xbfP|\\xb'
'9\\xb8\\xa1\\x16\\xdf\\xbf\\xf0P\\r\\xc7\\x80\\xa9\\xe2?\\x1f\\xbb\\xfaj\\x'
"f5\\x17\\x05@'\np78\ntp79\nbsS'_dual_coef_'\np80\ng31\nsS'degree'\np8"
"1\nI3\nsS'epsilon'\np82\nF0.0\nsS'max_iter'\np83\nI-1\nsS'fit_status"
"_'\np84\nI0\nsS'_intercept_'\np85\ng13\n(g14\n(I0\ntp86\ng16\ntp87\nRp8"
"8\n(I1\n(I1\ntp89\ng35\nI00\nS'N\\xb3CEkw(\\xc0'\np90\ntp91\nbsS'interc"
"ept_'\np92\ng13\n(g14\n(I0\ntp93\ng16\ntp94\nRp95\n(I1\n(I1\ntp96\ng35\nI"
"00\nS'N\\xb3CEkw(\\xc0'\np97\ntp98\nbsS'probB_'\np99\ng13\n(g14\n(I0\nt"
"p100\ng16\ntp101\nRp102\n(I1\n(I0\ntp103\ng35\nI00\ng48\ntp104\nbsS'cac"
"he_size'\np105\nI200\nsS'gamma'\np106\nF0.0\nsb.")

_PREC_STR = ('ccopy_reg\n_reconstructor\np0\n(csklearn.svm.classes\nOneClassSV'
"M\np1\nc__builtin__\nobject\np2\nNtp3\nRp4\n(dp5\nS'_impl'\np6\nS'one_"
"class'\np7\nsS'kernel'\np8\nS'rbf'\np9\nsS'verbose'\np10\nI00\nsS'pro"
"bability'\np11\nI00\nsS'support_'\np12\ncnumpy.core.multiarray\n_r"
"econstruct\np13\n(cnumpy\nndarray\np14\n(I0\ntp15\nS'b'\np16\ntp17\nRp"
"18\n(I1\n(I195\ntp19\ncnumpy\ndtype\np20\n(S'i4'\np21\nI0\nI1\ntp22\nRp2"
"3\n(I3\nS'<'\np24\nNNNI-1\nI-1\nI0\ntp25\nbI00\nS'\\x04\\x00\\x00\\x00\\x0"
'7\\x00\\x00\\x00\\x0b\\x00\\x00\\x00\\x0e\\x00\\x00\\x00\\x12\\x00\\x00\\x0'
'0\\x13\\x00\\x00\\x00o\\x00\\x00\\x00y\\x00\\x00\\x00\\x89\\x00\\x00\\x00\\'
'x8a\\x00\\x00\\x00\\x8b\\x00\\x00\\x00\\x8e\\x00\\x00\\x00\\xac\\x00\\x00\\'
'x00\\xb2\\x00\\x00\\x00\\xb3\\x00\\x00\\x00\\xc7\\x00\\x00\\x00\\xff\\x00\\'
'x00\\x00\\x03\\x01\\x00\\x00\\x07\\x01\\x00\\x00\\x0b\\x01\\x00\\x00\\x0f\\'
'x01\\x00\\x00\\x10\\x01\\x00\\x00\\x11\\x01\\x00\\x00\\x16\\x01\\x00\\x00\\'
'x17\\x01\\x00\\x00w\\x01\\x00\\x00\\x94\\x01\\x00\\x00\\x97\\x01\\x00\\x00'
'\\x9a\\x01\\x00\\x00\\x9d\\x01\\x00\\x00\\xa4\\x01\\x00\\x00\\xa5\\x01\\x00'
'\\x00\\xa6\\x01\\x00\\x00\\xa7\\x01\\x00\\x00\\xaa\\x01\\x00\\x00\\xac\\x01'
'\\x00\\x00\\xad\\x01\\x00\\x00\\xb3\\x01\\x00\\x00\\xb4\\x01\\x00\\x00\\xb9'
'\\x01\\x00\\x00\\xbe\\x01\\x00\\x00\\xc2\\x01\\x00\\x00\\xc7\\x01\\x00\\x00'
'\\xcb\\x01\\x00\\x00\\xce\\x01\\x00\\x004\\x02\\x00\\x00W\\x02\\x00\\x00X\\'
'x02\\x00\\x00\\xc8\\x02\\x00\\x00\\xe5\\x02\\x00\\x00\\xe6\\x02\\x00\\x00\\'
'xe7\\x02\\x00\\x00\\xe8\\x02\\x00\\x00\\xea\\x02\\x00\\x00\\xeb\\x02\\x00\\'
'x00\\xec\\x02\\x00\\x00\\xed\\x02\\x00\\x00\\xee\\x02\\x00\\x00\\xef\\x02\\'
'x00\\x00\\xf0\\x02\\x00\\x00\\xf1\\x02\\x00\\x00\\xf2\\x02\\x00\\x00\\xf3\\'
'x02\\x00\\x00\\xf4\\x02\\x00\\x00\\xf5\\x02\\x00\\x00\\xf7\\x02\\x00\\x00\\'
'xf8\\x02\\x00\\x00\\xf9\\x02\\x00\\x00\\xfa\\x02\\x00\\x00\\xfb\\x02\\x00\\'
'x00\\xfd\\x02\\x00\\x00\\xfe\\x02\\x00\\x00\\xff\\x02\\x00\\x00\\x00\\x03\\'
'x00\\x00\\x01\\x03\\x00\\x00\\x02\\x03\\x00\\x00!\\x03\\x00\\x00#\\x03\\x0'
'0\\x00%\\x03\\x00\\x00*\\x03\\x00\\x008\\x03\\x00\\x00;\\x03\\x00\\x00=\\x'
'03\\x00\\x00@\\x03\\x00\\x00R\\x03\\x00\\x00o\\x03\\x00\\x00\\x8f\\x03\\x0'
'0\\x00\\x95\\x03\\x00\\x00\\x96\\x03\\x00\\x00\\x99\\x03\\x00\\x00\\x9b\\x0'
'3\\x00\\x00\\x9f\\x03\\x00\\x00\\xa9\\x03\\x00\\x00\\xad\\x03\\x00\\x00\\xb'
'8\\x03\\x00\\x00\\xbd\\x03\\x00\\x00\\xbf\\x03\\x00\\x00\\xed\\x03\\x00\\x0'
'0\\xee\\x03\\x00\\x00\\xef\\x03\\x00\\x00\\xf0\\x03\\x00\\x00\\xf1\\x03\\x0'
'0\\x00\\xf2\\x03\\x00\\x00\\xf3\\x03\\x00\\x00\\xf4\\x03\\x00\\x00\\xf6\\x0'
'3\\x00\\x00\\xf8\\x03\\x00\\x00\\xf9\\x03\\x00\\x00\\xfb\\x03\\x00\\x00\\xf'
'c\\x03\\x00\\x00\\xfd\\x03\\x00\\x00\\xfe\\x03\\x00\\x00\\xff\\x03\\x00\\x0'
'0\\x00\\x04\\x00\\x00\\x02\\x04\\x00\\x00\\x04\\x04\\x00\\x00\\x05\\x04\\x0'
'0\\x00\\x06\\x04\\x00\\x00\\x08\\x04\\x00\\x00\\t\\x04\\x00\\x00\\x0c\\x04\\'
'x00\\x00\\x1c\\x04\\x00\\x00\\x1f\\x04\\x00\\x00H\\x04\\x00\\x00M\\x04\\x0'
'0\\x00[\\x04\\x00\\x00\\x89\\x04\\x00\\x00\\x96\\x04\\x00\\x00\\xf8\\x04\\x'
'00\\x00\\xfc\\x04\\x00\\x00\\x01\\x05\\x00\\x00\\x03\\x05\\x00\\x00\\n\\x05'
'\\x00\\x00\\x0b\\x05\\x00\\x00\\x10\\x05\\x00\\x00\\x1a\\x05\\x00\\x00+\\x0'
'5\\x00\\x00,\\x05\\x00\\x00-\\x05\\x00\\x00/\\x05\\x00\\x001\\x05\\x00\\x0'
'06\\x05\\x00\\x00:\\x05\\x00\\x00<\\x05\\x00\\x00H\\x05\\x00\\x00N\\x05\\x'
'00\\x00V\\x05\\x00\\x00d\\x05\\x00\\x00f\\x05\\x00\\x00m\\x05\\x00\\x00n\\'
'x05\\x00\\x00\\x92\\x05\\x00\\x00\\x93\\x05\\x00\\x00\\x94\\x05\\x00\\x00\\'
'x95\\x05\\x00\\x00\\x96\\x05\\x00\\x00\\x97\\x05\\x00\\x00\\x98\\x05\\x00\\'
'x00\\x99\\x05\\x00\\x00\\x9a\\x05\\x00\\x00\\x9b\\x05\\x00\\x00\\x9e\\x05\\'
'x00\\x00\\x9f\\x05\\x00\\x00\\xa0\\x05\\x00\\x00\\xa1\\x05\\x00\\x00\\xa2\\'
'x05\\x00\\x00\\xa3\\x05\\x00\\x00\\xa4\\x05\\x00\\x00\\xa5\\x05\\x00\\x00\\'
'xa8\\x05\\x00\\x00\\xa9\\x05\\x00\\x00\\xaa\\x05\\x00\\x00\\xab\\x05\\x00\\'
'x00\\xd1\\x05\\x00\\x00\\x1f\\x06\\x00\\x00$\\x06\\x00\\x00&\\x06\\x00\\x0'
'0*\\x06\\x00\\x003\\x06\\x00\\x006\\x06\\x00\\x00?\\x06\\x00\\x00A\\x06\\x'
'00\\x00B\\x06\\x00\\x00F\\x06\\x00\\x00I\\x06\\x00\\x00K\\x06\\x00\\x00M\\'
'x06\\x00\\x00N\\x06\\x00\\x00V\\x06\\x00\\x00W\\x06\\x00\\x00Y\\x06\\x00\\'
'x00l\\x06\\x00\\x00\\x86\\x06\\x00\\x00\\x99\\x06\\x00\\x00\\xdf\\x06\\x00'
"\\x00'\np26\ntp27\nbsS'dual_coef_'\np28\ng13\n(g14\n(I0\ntp29\ng16\ntp3"
"0\nRp31\n(I1\n(I1\nI195\ntp32\ng20\n(S'f8'\np33\nI0\nI1\ntp34\nRp35\n(I3\n"
"S'<'\np36\nNNNI-1\nI-1\nI0\ntp37\nbI00\nS'\\x00\\x00\\x00\\x00\\x00\\x00\\"
'xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\r\\x17p/\\xc5\\x1a\\xa8?\\x00\\x'
'00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00'
'\\x00\\x00\\x00\\x00\\xf0?\\xcc\\x0f/\\xd7\\xce\\xf9\\xcf?\\x01\\x9eao\\x1'
'0o\\xcc?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00'
'\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\x'
'f0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0'
'?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\'
'x02\\xd7`\\xccs\\xd4\\xd7?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\'
'x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x0'
'0\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\'
'x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x0'
'0\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\'
'x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x0'
'0\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\'
'x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x0'
'0\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x02^m\\x96]\\x14\\xe7?\\x90\\'
'xec4\\x8cz\\xcc\\xe6?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\'
'x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x0'
'0\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\'
'x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?K\\x1a\\xb3\\x97\\xc1\\x'
'8a\\xe3?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00'
'\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\x'
'f0?\\xe4%\\x9c\\xc8SQ\\xe3?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00'
'\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x'
'00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00'
'\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x'
'00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00'
'\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x'
'00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00'
'\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x'
'00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00'
'\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\xb8\\xe7#W"\\xff\\xee?\\x00\\x'
'00\\x00\\x00\\x00\\x00\\xf0?Ag\\xc3q\\x04 \\xbd?\\x00\\x00\\x00\\x00\\x00'
'\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x'
'00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00'
'\\xf0?\\xc8\\xd24\\xfb\\x90\\x1b~?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x0'
'0\\x00\\x00\\x00\\x00\\x00\\xf0?\\xc3h;4\\xb8\\xae\\xe9?\\x00\\x00\\x00\\x'
'00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00'
'\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x'
'00\\x00\\xf0?\\x9a\\xe7\\xfb\\x8cI\\x92\\xd0?\\x00\\x00\\x00\\x00\\x00\\x0'
'0\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\'
'xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\xfd\\xa1M\\x83\\xacE\\xe2?\\x00'
'\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x'
'00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00'
'\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x'
'00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00'
'\\x00\\x00\\x00\\xf0?\\x9a:\\x8e\\x9f\\xcd\\x1e\\xd2?\\x00\\x00\\x00\\x00\\'
'x00\\x00\\xf0?r\\xa6\\xba\\xf3\\x18\\xf4\\xef?\\x00\\x00\\x00\\x00\\x00\\x'
'00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00'
'\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\x'
'f0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0'
'?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\'
'x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?x~e'
'\\xaaH*\\xa9?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00'
'\\x00\\xf0?\\xcb\\xd7\\xd6QP\\x11\\xec?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0'
'?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x86\\n\\xa7(j\\xaa\\xe0?\\xaeF\\x16'
'H\\xee\\x02\\xcf?K\\xdc\\xcax=\\xcf\\xb6?\\x00\\x00\\x00\\x00\\x00\\x00\\x'
'f0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0'
'?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\'
'x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x0'
'0\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?[R\\xe'
'fvT\\x19\\xd3?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x0'
'0\\x00\\xf0?sC7\\xe1#X\\xe8?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x0'
'0\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?#r\\xd6\\xa'
'6%\\x93\\xe8?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00'
'\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x10\\t\\xb5\\x07,\\xa5\\xb'
'4?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?'
'\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?|\\'
'xfb\\xff\\x94\\x90r\\xc3?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x'
'00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00'
'\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x'
'00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00'
'\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x'
'00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00'
'\\x00\\xf0?(Y\\x92R\\x9a\\xc6\\xba?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x'
'00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00'
'\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x'
'00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00'
'\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x'
'00\\x00\\x00\\x00\\xf0?S\\xe4[v\\xfc\\xae\\xd5?\\x00\\x00\\x00\\x00\\x00\\'
'x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x0'
'0\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\'
'xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf'
'0?\\x89A\\xe1\\x8di\\xb4\\xed?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x'
'00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00'
'\\x00\\x00\\x00\\x00\\xf0?\\x89UW<\\xf6\\xc7\\xe8?\\x00\\x00\\x00\\x00\\x0'
'0\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\'
'x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?3\\xb8\\xdd\\xbd\\x12J\\xd6?'
'\\x9e\\xec\\xf2\\xfd\\xc9+\\xe7?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\'
"x00\\x00\\x00\\x00\\x00\\xf0?\\x00\\x00\\x00\\x00\\x00\\x00\\xf0?'\np38\nt"
"p39\nbsS'shrinking'\np40\nI01\nsS'class_weight'\np41\nNsS'_gamma'\n"
"p42\nF0.125\nsS'probA_'\np43\ng13\n(g14\n(I0\ntp44\ng16\ntp45\nRp46\n(I"
"1\n(I0\ntp47\ng35\nI00\nS''\np48\ntp49\nbsS'_sparse'\np50\nI00\nsS'clas"
"s_weight_'\np51\ng13\n(g14\n(I0\ntp52\ng16\ntp53\nRp54\n(I1\n(I0\ntp55\n"
"g35\nI00\ng48\ntp56\nbsS'random_state'\np57\nNsS'_sklearn_version'"
"\np58\nS'0.18'\np59\nsS'tol'\np60\nF0.001\nsS'coef0'\np61\nF0.0\nsS'nu"
"'\np62\nF0.1\nsS'n_support_'\np63\ng13\n(g14\n(I0\ntp64\ng16\ntp65\nRp6"
"6\n(I1\n(I2\ntp67\ng23\nI00\nS'\\x88\\xf7\\xac\\x1d\\xa1\\x7f\\x00\\x00'\np"
"68\ntp69\nbsS'shape_fit_'\np70\n(I1776\nI8\ntp71\nsS'C'\np72\nF0.0\nsS"
"'support_vectors_'\np73\ng13\n(g14\n(I0\ntp74\ng16\ntp75\nRp76\n(I1\n("
"I195\nI8\ntp77\ng35\nI00\nS'\\'\\xc3\\xde~\\'R\\x03@\\xef\\xa6\\xa2\\xa1E\\"
'xe8\\xf5?\\xf5\\xa1k\\xc2\\xa8\\x8e\\xe7??\\xecJ\\xf9\\xads\\xcc\\xbf\\xd'
'b\\x9b\\r\\x06\\xf0k\\x04@\\xcb>K\\xb5\\x83\\x16\\xf0?l\\xb8Z/R/\\xaf\\xb'
'f\\x9d\\x0b\\xb2\\x8fl\\x16\\xd1?w\\xb8\\x15\\x1e\\n\\xcb\\xec\\xbfoe\\ra$'
'o\\xfb\\xbf-\\xd7\\xf2I-\\x00\\xc0\\xbf\\xad~2A\\x06D\\xf2?\\x98N/9\\xf3'
'\\xb7\\xf0?X\\x11\\xf6\\x08\\xda\\x8b\\xd1\\xbf0\\xd0\\xd6\\x8fB\\x99\\x00'
"@9\\x1d;\\xd2\\x0f\\xd9\\xe3\\xbf\\'\\xc3\\xde~\\'R\\x03@\\x1b\\xa6R\\x7f-"
'\\x9e\\xf2?\\xf5\\xa1k\\xc2\\xa8\\x8e\\xe7?O\\xb4T\\xb2\\x9c\\xbd\\xcd?q)'
'\\x94\\xb6}\\x17\\x05@}\\xe9\\x87\\x15\\x08A\\xf7?=\\xaa\\x83\\xac\\x01\\x'
'c2\\xc5?\\xd5\\x98\\xf8\\x0e\\x87L\\xc0\\xbfj,\\x19Y]/\\xf2\\xbf\\xc9cm\\'
'x1c\\xf4\\xda\\xf4\\xbf\\xb5\\xe6f\\x93\\x99\\x11\\xed?\\xbfr\\xa6\\x96/\\'
'x8a\\xf9?~0*\\xeex\\x13\\xec?\\xd0_\\\\p\\xd9\\xab\\xf0\\xbf\\xf3\\xa2\\x9'
'9A\\xb1\\x85\\x01@p\\x0e\\xdf\\x98\\xc7\\x13\\xdb\\xbfw\\xb8\\x15\\x1e\\n\\'
'xcb\\xec\\xbf\\xc2\\xa7\\xf2\\xc3]2\\xf9?\\xdb\\xfc\\xa9k\\xae\\x8e\\xf9?'
'\\t\\xd0\\x03\\xb3\\xbe\\xb0\\xf4?\\x17\\xd8\\xc7y\\x97\\xda\\xe6\\xbf/\\x9'
'4\\xc4u\\x8ck\\xfe?R@\\x9a:w%\\xcd?g~k\\x10R\\xe0\\xb1?f\\xf5u\\x8b\\x1'
'5\\xb3\\xf3?=\\x8bJ\\xd83\\xd7\\xd5?\\x97~\\xe6\\xe8\\xd9\\x0b\\xde\\xbfO'
"\\xb4T\\xb2\\x9c\\xbd\\xcd?\\xfb\\'\\x98A\\xcep\\xfb?\\xe0z\\xc4\\xc8$T\\x"
'04@\\x03\\x92N|\\xc3\\t\\xf7?\\x9f$\\xe9\\xee\\xb3;\\xf0\\xbfoJ\\xcf\\x82'
'z\\xd2\\xef?\\x1b\\xa6R\\x7f-\\x9e\\xf2?\\xd4N\\xe7=Y\\x17\\xcc?O\\xb4T\\'
'xb2\\x9c\\xbd\\xcd?\\xa1\\xf1}\\x7f\\x97\\xc2\\xf8?R\\xf0\\xe4M\\x84\\xf6'
'\\xe0\\xbf\\x94\\x88Ofs\\xae\\xfc\\xbf\\x92\\x80l\\x7f\\xb6\\xca\\x00\\xc0'
'\\xc7\\xcc5\\xed\\r\\xc3\\xf9\\xbf\\xa0\\xc2\\xfajW\\xf9\\xe5\\xbf\\xea0\\x'
'ea@\\x8e\\x11\\xd9?\\xca\\xc2\\x80\\xd5\\xd4\\x1b\\xb2\\xbfn\\xc5L\\xc5&&'
'\\xf6\\xbf\\xa8\\xf5\\xd2p\\xbc(\\xf3?C\\xce#)\\x90^\\xd6\\xbf\\xd4\\xc7;'
'Y\\xbax\\xeb?j,\\x19Y]/\\xf2\\xbfl\\xed\\xd5\\xe84\\xe6\\xb1\\xbf\\xf5\\x'
'a1k\\xc2\\xa8\\x8e\\xe7?\\x94\\x9fo \\xb0\\x91\\xd8?\\x16\\x8f2\\x03\\xf0'
'w\\xf3\\xbf&\\xe4\\x99\\xf2\\xcf\\x0e\\xe5\\xbf\\xf4\\xa5Q!\\xbd\\x92\\x01'
'\\xc0\\x94\\x168\\x85\\xe2\\x19\\x07\\xc0\\xc3\\x95\\x92\\x1f\\xc6F\\xfb?Z'
'V\\xe9Hww\\x06@\\x8d\\xe5Mo\\xc0\\x0b\\x05@\\x16\\xdc\\xf5\\xae\\x88\\x9e'
'\\x01@\\x9b\\xe0Y&B}\\xf7\\xbfn\\x96\\x1eM\\xaf\\x17\\xbc?\\xde\\x03P[\\x'
'cb\\x80\\xef\\xbf:\\xb3\\x06\\xd8;(\\xea\\xbf\\xc3\\x95\\x92\\x1f\\xc6F\\x'
'fb?\\xc2\\xa7\\xf2\\xc3]2\\xf9?\\x1b\\xb8\\xae\\x9a\\xbd\\x0b\\xf4?O\\xb4'
'T\\xb2\\x9c\\xbd\\xcd?\\x1dW\\xc1\\xe5\\x9dZ\\xe1??x\\xa8G\\xcf\\xfa\\xfd'
"\\xbfW[\\x12\\x18\\xe2\\x9a\\xfd\\xbf!\\xa0\\'\\xf6\\xaa\\x1e\\xf8\\xbf\\x1"
'3\\xaa\\xb2\\xee\\xc9>\\xe8?\\xc2\\xa7\\xf2\\xc3]2\\xf9?\\xdb\\xfc\\xa9k\\'
'xae\\x8e\\xf9?d!\\xd5$w\\x1d\\xf7?\\xc8D\\xfc\\xfd\\x047\\xec\\xbf\\x15\\'
'x81f\\xc1\\xceN\\xd3?\\x94\\x88Ofs\\xae\\xfc\\xbf\\x92KR\\x82L\\xf2\\x03'
"\\xc0\\x99|\\'\\xa35\\xf9\\xf5\\xbfoe\\ra$o\\xfb\\xbf\\xb3\\xfb\\x90G\\xd5"
'\\x88\\x01\\xc0C\\xe8M\\xfcE\\xc0\\x02\\xc0\\xc8D\\xfc\\xfd\\x047\\xec\\xb'
'fR\\x8e\\x95t\\x01m\\xa7?\\xe2\\x00\\x06p\\xbf\\x17j\\xbf\\x86f\\x0b\\x1b'
"xq\\x9d\\xbf\\x99|\\'\\xa35\\xf9\\xf5\\xbf\\x8b\\xb3\\xd6R\\xaa\\x01\\x01\\"
'xc0\\xa6\\xb2&\\xbe\\xb9\\x8e\\xfd\\xbf\\x95?e\\xc3\\xe9\\x89\\x01\\xc0%h'
'\\xe5>\\xf3\\x9f\\xd2?R\\x8e\\x95t\\x01m\\xa7?F\\xda\\x03\\xa8\\xbf\\x92\\'
'xc6\\xbf\\xd3\\xfcUV$Q\\xe8?\\xf5\\x1cD7\\xe6\\x8c\\xfd\\xbf\\x8b\\xb3\\x'
'd6R\\xaa\\x01\\x01\\xc0\\xa6\\xb2&\\xbe\\xb9\\x8e\\xfd\\xbf\\xe7\\x96|\\x8'
'a\\x8dS\\x00\\xc0p\\x0e\\xe2;\\xce\\x88\\xe9\\xbf\\xa2 $b\\xb7\\xaa\\x92\\'
'xbf8\\x83\\x18b\\xd5\\xac\\xd2\\xbf\\xa17I\\x9b\\xc4\\xb4\\xdd?V\\x13\\xe'
'd\\xc8\\xff\\x1b\\x07@\\x95\\xa8B\\xe6u|\\xfc?\\xf5\\xa1k\\xc2\\xa8\\x8e\\'
'xe7?\\xca\\xc2\\x80\\xd5\\xd4\\x1b\\xb2\\xbf\\xcd\\xc3\\xf5i\\x0b\\xb7\\xe'
'6?&\\xe4\\x99\\xf2\\xcf\\x0e\\xe5\\xbf?\\xb6c+1\\xf6\\xd5?\\xd3\\xfcUV$Q'
'\\xe8?j,\\x19Y]/\\xf2\\xbfBf]\\x83<\\xb9\\xfe\\xbfF\\x10\\xa9UA\\xcd\\xf'
'a\\xbf\\x81r\\xda\\xf3H"\\xe1?\\x9b\\xe0Y&B}\\xf7\\xbf+\\x05\\xab\\xad%\\'
'xa4\\xd5\\xbf\\xda\\xf7o\\xdc\\x9bL\\xef?p\\x0e\\xdf\\x98\\xc7\\x13\\xdb\\'
'xbf\\xf1\\xe5\\xa0i\\x9e\\x10\\xff?\\xebHeu\\xfa\\x13\\xe8?\\xea0\\xea@\\'
'x8e\\x11\\xd9?\\x8c\\xbb\\xea\\xc3\\xb8\\xec\\xd7\\xbfn\\xc5L\\xc5&&\\xf6'
'\\xbfn\\x96\\x1eM\\xaf\\x17\\xbc?1\\xd3\\x8eoN\\xa6\\x00\\xc0\\x9d\\x0b\\x'
'b2\\x8fl\\x16\\xd1?\\xc3\\x95\\x92\\x1f\\xc6F\\xfb?\\x1b\\xa6R\\x7f-\\x9e'
'\\xf2?\\xd4N\\xe7=Y\\x17\\xcc?\\xeaR\\x94G\\xb2\\xaf\\xb4?%h\\xe5>\\xf3\\'
'x9f\\xd2?m\\xce\\x1f\\xbd\\xbc\\x98\\x03\\xc0=\\xaa\\x83\\xac\\x01\\xc2\\x'
"c5?R\\xf3%\\x00\\x9c\\x94\\x01\\xc0\\x99|\\'\\xa35\\xf9\\xf5\\xbfoe\\ra$o"
'\\xfb\\xbf-\\xd7\\xf2I-\\x00\\xc0\\xbf\\xeaE\\xdd\\x0c\\xaf\\x82\\xea\\xbf'
'\\xc8\\xfbf\\x87]\\xd4\\xf8\\xbf\\x08;\\x82\\xc8\\x1c\\xe7\\xca\\xbf\\xbc\\'
'x1fv&\\x8f^\\xf4\\xbf\\x94\\x168\\x85\\xe2\\x19\\x07\\xc0j,\\x19Y]/\\xf2'
'\\xbfBf]\\x83<\\xb9\\xfe\\xbf\\xea0\\xea@\\x8e\\x11\\xd9?\\x8c\\xbb\\xea\\'
'xc3\\xb8\\xec\\xd7\\xbf\\x16\\x8f2\\x03\\xf0w\\xf3\\xbf\\xbeh\\xd0\\nf\\x7'
'f\\xdb?W[\\x12\\x18\\xe2\\x9a\\xfd\\xbf<~\\xec\\xda\\xd1O\\xed\\xbf\\xc0\\'
'x8f\\xcc%\\x83;\\xa8?\\xa0\\xc2\\xfajW\\xf9\\xe5\\xbf\\x0b\\x84n\\xc5\\xd'
'd\\x88\\xe4\\xbf\\xbd9\\x85\\xbf\\xf1`\\xf9\\xbf\\xa9nA\\xac`\\\\\\xc0\\xbf'
'\\xab\\x03<\\x98\\xe1 \\xee?K\\x01o\\xf2\\xeb\\xa7\\xd9?9\\xab,!\\xa1\\xb'
'2\\x02@\\x83]\\xff\\x86\\xe1?\\xc8\\xbfF\\xc4\\x9a\\xaf\\x87\\x8d\\xec\\xb'
"f&)0\\x1c\\xd8\\x88\\xf2\\xbf\\xbd9\\x85\\xbf\\xf1`\\xf9\\xbf\\'C\\xa5\\xa"
'2\\xe9\\xc7\\xfc?\\xa3S\\x11\\x15%\\xc4\\xf4\\xbf\\xb0\\xcbB\\xc0\\xb0\\x8'
'5\\xe0?\\xa17I\\x9b\\xc4\\xb4\\xdd?w\\xb8\\x15\\x1e\\n\\xcb\\xec\\xbf\\xa0'
'\\xc2\\xfajW\\xf9\\xe5\\xbf\\x86\\xcb\\xad\\x84PJ\\xf5\\xbf\\xbd9\\x85\\xb'
'f\\xf1`\\xf9\\xbf\\xcd\\xc3\\xf5i\\x0b\\xb7\\xe6?\\xa3S\\x11\\x15%\\xc4\\x'
'f4\\xbfW[\\x12\\x18\\xe2\\x9a\\xfd\\xbfUT\\xd7\\x08^\\x0b\\x0b\\xc0\\x83]'
'\\xff\\x86\\xe1?\\xc8\\xbf\\xa0\\xc2\\xfajW\\xf9\\xe5\\xbf\\x0b\\x84n\\xc5'
'\\xdd\\x88\\xe4\\xbf\\xbd9\\x85\\xbf\\xf1`\\xf9\\xbf\\xe9s%\\xa2\\xd4 \\xf'
'2\\xbf+</4.\\xbd\\t\\xc0j\\xa5I\\x08jl\\t\\xc0\\x13\\xfc\\xaa\\x86\\xad\\x'
'ad\\x08\\xc0oJ\\xcf\\x82z\\xd2\\xef?=\\x8bJ\\xd83\\xd7\\xd5?\\x97~\\xe6\\'
'xe8\\xd9\\x0b\\xde\\xbf\\x8c\\xbb\\xea\\xc3\\xb8\\xec\\xd7\\xbf\\xa5\\x83\\'
'xa8l\\xe6\\x87\\xff\\xbfU\\xfeMu\\xa9\\xee\\xfb\\xbf\\xc8m9\\xcdU\\x1d\\x'
'e8\\xbf8RU\\xcfy\\xb1\\xe0\\xbf\\xc0\\x8f\\xcc%\\x83;\\xa8?\\xa7~u\\xc3\\'
'xed\\xa1\\xd1\\xbf\\xb5\\xe6f\\x93\\x99\\x11\\xed?\\xc4\\x84\\xde\\xe7\\xe'
'4\\xd4\\x02@\\nH\\xaa\\xb4;\\x15\\xcb\\xbf\\x82\\n\\x99\\xd0]\\xd6\\xf7\\xb'
'f\\x03\\x92N|\\xc3\\t\\xf7?\\x9f\\xa1}\\x95\\x98e\\xd7?\\x13\\xaa\\xb2\\xe'
'e\\xc9>\\xe8?EG\\xc50\\xca\\x7f\\xe1?5]p\\xf1\\xb7\\x0b\\xe2??\\xecJ\\xf'
'9\\xads\\xcc\\xbf\\x96\\xea\\xe7G\\x9c\\xc6\\x01\\xc0X\\x11\\xf6\\x08\\xda'
'\\x8b\\xd1\\xbfK\\x01o\\xf2\\xeb\\xa7\\xd9?\\xd31pS\\x8e)\\xe5?\\xf5\\x1c'
'D7\\xe6\\x8c\\xfd\\xbf\\xa7~u\\xc3\\xed\\xa1\\xd1\\xbf\\x0b\\x84n\\xc5\\xd'
'd\\x88\\xe4\\xbf?\\xecJ\\xf9\\xads\\xcc\\xbf\\x19j\\xf2f\\xe6\\x9f\\xdd\\x'
'bf\\xb6E^\\x0e\\x8dK\\xef\\xbf6t\\x80\\x03\\x9ak\\xf2?9\\xab,!\\xa1\\xb2'
'\\x02@\\xc0\\x8f\\xcc%\\x83;\\xa8?\\xa7~u\\xc3\\xed\\xa1\\xd1\\xbf\\xf5\\x'
'a1k\\xc2\\xa8\\x8e\\xe7?d!\\xd5$w\\x1d\\xf7?\\xa9nA\\xac`\\\\\\xc0\\xbf\\x'
'a3S\\x11\\x15%\\xc4\\xf4\\xbf\\x0e\\xddYC~\\xbb\\xfa?\\xecDO5\\x1f3\\xf7'
'?>\\xeb\\xe5\\xa3\\x137\\x05@\\xc3V\\x11Z\\x83\\x1c\\x08@\\x8d\\xe5Mo\\xc'
'0\\x0b\\x05@\\x1a\\xc4w\\x08\\xe8\\xf6\\xfb?\\xc8D\\xfc\\xfd\\x047\\xec\\x'
'bf\\xab\\x03<\\x98\\xe1 \\xee?l\\xb8Z/R/\\xaf\\xbf\\xecyi2\\x89\\x0b\\xf'
'4?\\xc3\\x95\\x92\\x1f\\xc6F\\xfb?\\x95\\xa8B\\xe6u|\\xfc?\\xdb\\xfc\\xa9'
'k\\xae\\x8e\\xf9?7\\x15}\\xd7\\xb9\\xfb\\xe5?ek\\x93\\xf5)~\\xe1\\xbf\\xa'
'8\\xf5\\xd2p\\xbc(\\xf3?\\x03\\x92N|\\xc3\\t\\xf7?o\\xc0\\x8d<\\x16\\x16\\'
'xff?\\x13\\xaa\\xb2\\xee\\xc9>\\xe8?\\xc2\\xa7\\xf2\\xc3]2\\xf9?\\x8d\\xe'
'5Mo\\xc0\\x0b\\x05@\\x1a\\xc4w\\x08\\xe8\\xf6\\xfb?,xn\\xf8)r\\x02\\xc0\\'
'xa2 $b\\xb7\\xaa\\x92\\xbfB\\xc5\\xfb\\x89l7\\xf6\\xbf8RU\\xcfy\\xb1\\xe'
'0\\xbfj,\\x19Y]/\\xf2\\xbfoe\\ra$o\\xfb\\xbfF\\x10\\xa9UA\\xcd\\xfa\\xbf'
'\\x95?e\\xc3\\xe9\\x89\\x01\\xc0\\xe5\\xf6au\\x0b\\x87\\xca?_S\\x18\\x7f\\'
'x85\\xb6\\xc2\\xbf\\x03\\x92N|\\xc3\\t\\xf7?9v\\x12$7\\xda\\x05@f\\xf5u\\'
'x8b\\x15\\xb3\\xf3?\\x91J\\x05\\xba*\\xa8\\xee?5]p\\xf1\\xb7\\x0b\\xe2?\\'
'x81r\\xda\\xf3H"\\xe1?p\\x0e\\xe2;\\xce\\x88\\xe9\\xbf\\x82\\n\\x99\\xd0]'
'\\xd6\\xf7\\xbf@\\xbc\\xd3\\xeaH\\x10\\xe6?m_\\xdc3T\\x9f\\xf5?{\\xef\\xb'
'8\\xebQG\\xdb\\xbfF\\xc4\\x9a\\xaf\\x87\\x8d\\xec\\xbf\\xea0\\xea@\\x8e\\x'
'11\\xd9?\\x8c\\xbb\\xea\\xc3\\xb8\\xec\\xd7\\xbf\\nH\\xaa\\xb4;\\x15\\xcb\\'
'xbf\\xc5\\x9c\\x89Y\\xec\\xb1\\xf1\\xbf}\\xec\\xc8\\x18\\xe60\\xf5?\\xb8\\'
'xc5\\xb9\\x1f\\xd6\\x1e\\x01@f\\xf5u\\x8b\\x15\\xb3\\xf3?\\x87U\\x99&_-\\'
'x03@\\x9bA\\xa5<\\x9f\\x11\\xff?\\xbfr\\xa6\\x96/\\x8a\\xf9? 2\\x81I\\x9'
'4\\x82\\xfb\\xbf\\x1c\\xa2w|$\\xe4\\xe3?Yd:\\xb7\\x05\\xc2\\xdd\\xbf<~\\x'
'ec\\xda\\xd1O\\xed\\xbf\\xb2\\xd2\\xf2\\x8c\\xd1.\\xd2?\\xef\\xa6\\xa2\\xa'
'1E\\xe8\\xf5?-C\\xd0\\x06HJ\\x02@\\x83!;vjQ\\x0b@\\xa1\\xf1}\\x7f\\x97\\'
'xc2\\xf8?\\xa8\\xf5\\xd2p\\xbc(\\xf3?\\xc8m9\\xcdU\\x1d\\xe8\\xbf:\\xb3\\'
'x06\\xd8;(\\xea\\xbfm;\\xf4\\xed\\xeb\\x00\\t@=\\xad<\\xfa\\xf7t\\x13@\\x'
'e7\\xd9\\x1fqI\\xca\\x12@\\x1d\\xe2\\xee\\xacm\\x15\\x10@E\\x0e\\x87Ub\\x'
'c0\\x03@\\xb9\\xd9\\xb6B\\xff\\xb7\\xf2\\xbf_\\x88\\xda3\\x94^\\xbe\\xbf\\'
'xd5\\x98\\xf8\\x0e\\x87L\\xc0\\xbf\\xc0\\x8f\\xcc%\\x83;\\xa8?\\xebHeu\\x'
"fa\\x13\\xe8?;\\x9f\\'\\xd4&P\\xfc?\\xcc~\\x98\\x92\\xf9w\\x06@\\x81y\\xb"
'fd v\\xff?\\xb5\\xb8\\xa5\\x87\\xa9"\\xf2?\\xc8m9\\xcdU\\x1d\\xe8\\xbf\\x'
'd7\\xc4\\x8f\\x1a\\xdf\\xea\\xcc\\xbfj,\\x19Y]/\\xf2\\xbf\\x9cd\\xbd>\\x0'
'c%\\xf8\\xbf\\x17\\xf5\\xefF\\xf8\\x05\\xd3\\xbf\\x1a\\xc4w\\x08\\xe8\\xf6'
'\\xfb?\\x1dW\\xc1\\xe5\\x9dZ\\xe1?R\\x8e\\x95t\\x01m\\xa7?0\\xd0\\xd6\\x8'
'fB\\x99\\x00@\\x9d\\x0b\\xb2\\x8fl\\x16\\xd1?\\x95E\\x84\\xd5\\xed|\\xf7?'
'\\x95\\xa8B\\xe6u|\\xfc?\\xfdq\\x91\\xd2\\x8b\\xe9\\x00@h3\\rv,h\\x00@\\x'
'19\\xf3y\\xf4Fi\\x02@"\\xd1\\xf1^\\x9fq\\xff?3kXdvD\\xd2?\\x9f\\xa1}\\x'
'95\\x98e\\xd7?oJ\\xcf\\x82z\\xd2\\xef?\\x91J\\x05\\xba*\\xa8\\xee?]\\x14'
'\\x0f;\\x04\\xab\\x03@\\xbfr\\xa6\\x96/\\x8a\\xf9?\\xec\\xd7l\\x93+\\x12\\'
'x01@:W\\x97\\x8cye\\xfd?F\\xda\\x03\\xa8\\xbf\\x92\\xc6\\xbf\\xd1f\\x8aP'
'\\xf8\\x01\\xe2?\\xf1\\xe5\\xa0i\\x9e\\x10\\xff?ZV\\xe9Hww\\x06@\\xfdq\\x'
'91\\xd2\\x8b\\xe9\\x00@\\t\\xd0\\x03\\xb3\\xbe\\xb0\\xf4?T^\\xb2\\x03\\x05'
'\\x1f\\xfe?\\x08;\\x82\\xc8\\x1c\\xe7\\xca\\xbft\\xa7-\\x11C\\x99\\xf1\\xb'
'f;\\xe8 \\xd5\\xa5\\x00\\xe7\\xbf\\xb2\\xd2\\xf2\\x8c\\xd1.\\xd2?\\xef\\xa'
"6\\xa2\\xa1E\\xe8\\xf5?;\\x9f\\'\\xd4&P\\xfc?\\xd5xR=\\x0e\\x1b\\n@\\x1d\\"
'xa0V\\\\E\\xbd\\xf4?\\xa8\\xf5\\xd2p\\xbc(\\xf3?\\xc8m9\\xcdU\\x1d\\xe8\\x'
'bf9\\x1d;\\xd2\\x0f\\xd9\\xe3\\xbfw\\xb8\\x15\\x1e\\n\\xcb\\xec\\xbf\\x91J'
'\\x05\\xba*\\xa8\\xee?\\xea0\\xea@\\x8e\\x11\\xd9?}\\x00\\x98E\\xcd\\xcf\\'
'xe0\\xbf\\x1dW\\xc1\\xe5\\x9dZ\\xe1?\\x14y\\x01\\x8dz\\x03\\x00\\xc0\\xde'
'\\x03P[\\xcb\\x80\\xef\\xbf\\x9f\\xa1}\\x95\\x98e\\xd7?j,\\x19Y]/\\xf2\\x'
'bf\\xc9cm\\x1c\\xf4\\xda\\xf4\\xbf\\x86\\xcb\\xad\\x84PJ\\xf5\\xbf\\xe7\\x'
'96|\\x8a\\x8dS\\x00\\xc0\\xbdX\\x18A\\xb9\\xc9\\xf0\\xbf\\n\\xb6.v\\x8d\\t'
'\\x01\\xc0\\xb2\\xd4j_\\xd4\\xac\\xf0\\xbf\\x1e\\n\\\\\\xf0~\\xcf\\xf1\\xbfw'
'\\xb8\\x15\\x1e\\n\\xcb\\xec\\xbf\\x9cd\\xbd>\\x0c%\\xf8\\xbf\\xa6\\xb2&\\x'
'be\\xb9\\x8e\\xfd\\xbf\\x95?e\\xc3\\xe9\\x89\\x01\\xc0\\xc8D\\xfc\\xfd\\x0'
'47\\xec\\xbf4\\xb5\\xd50\\xe2\\x00\\xff\\xbf\\x05\\x98\\xbe;\\xdb#\\xf7\\x'
"bf:\\xb3\\x06\\xd8;(\\xea\\xbf\\x99|\\'\\xa35\\xf9\\xf5\\xbfoe\\ra$o\\xfb"
'\\xbf\\xa6\\xb2&\\xbe\\xb9\\x8e\\xfd\\xbf\\x07\\x97\\xe2\\xdb\\x80\\x87\\xf'
'4\\xbfB\\xaa?d\\x0b\\xcf\\xf4\\xbfR\\xf0\\xe4M\\x84\\xf6\\xe0\\xbf\\xe2\\x'
'00\\x06p\\xbf\\x17j\\xbf\\xd1\\r\\xb3\\xfe\\xd0\\x00\\x00\\xc0\\x83]\\xff\\'
'x86\\xe1?\\xc8\\xbf\\xe2\\x0f\\x15\\x9e\\xa6]\\xc1?\\x0b\\x84n\\xc5\\xdd\\'
"x88\\xe4\\xbf\\xeaE\\xdd\\x0c\\xaf\\x82\\xea\\xbf\\'\\xfa\\x0f,Be\\xe9?j\\"
'xfa\\xbc\\x95\\x08\\x90\\x10@\\x15\\ri\\x92+\\xff\\x11@\\xffo\\xa4Y\\x055'
'\\x13@{\\xef\\xb8\\xebQG\\xdb\\xbf\\xf6b\\x1d\\xfa\\xdb\\x90\\xf1\\xbf\\xe'
'6m+\\xed\\xc8\\x0b\\xf8\\xbf\\xeaE\\xdd\\x0c\\xaf\\x82\\xea\\xbf\\x1dW\\xc'
'1\\xe5\\x9dZ\\xe1?\\x93o-C\\xe24\\xf5?\\x90g\\xe3.N&\\x10@^^\\x1b\\x17b'
"r\\x10@\\x99|\\'\\xa35\\xf9\\xf5\\xbf\\xf3\\x81\\xb5LN\\xca\\xde\\xbf&)0\\"
'x1c\\xd8\\x88\\xf2\\xbf}\\x00\\x98E\\xcd\\xcf\\xe0\\xbf\\xe9s%\\xa2\\xd4 '
'\\xf2\\xbfP\\xdd<\\xbaSY\\xfb?\\x88\\x94\\x9c\\xf0\\xb5&\\x13@^^\\x1b\\x1'
'7br\\x10@\\x1b\\x18\\xf9\\x89Y7\\xe5\\xbf\\xa7~u\\xc3\\xed\\xa1\\xd1\\xbf'
'-\\xd7\\xf2I-\\x00\\xc0\\xbf\\xeaE\\xdd\\x0c\\xaf\\x82\\xea\\xbf\\xe5\\xf6'
'au\\x0b\\x87\\xca?e\\x19\\xb6\\xcd\\xf4\\x96\\x0f@\\x1a(e\\n9m\\xfe?\\x1f'
'\\xd1\\xd4\\x97G<\\x11@\\xb2\\xd2\\xf2\\x8c\\xd1.\\xd2?F\\xc4\\x9a\\xaf\\x'
'87\\x8d\\xec\\xbf\\xe6m+\\xed\\xc8\\x0b\\xf8\\xbf\\xbd9\\x85\\xbf\\xf1`\\x'
'f9\\xbf\\xe6>\\\\L\\x94v\\x0c@\\x93o-C\\xe24\\xf5?vvJ\\xebbu\\x12@~\\x8a'
'1X:\\xa1\\x11@\\x83]\\xff\\x86\\xe1?\\xc8\\xbf\\xa7~u\\xc3\\xed\\xa1\\xd1'
'\\xbf\\x97~\\xe6\\xe8\\xd9\\x0b\\xde\\xbf\\xeaE\\xdd\\x0c\\xaf\\x82\\xea\\x'
'bf\\xe5\\xf6au\\x0b\\x87\\xca?\\xf0[&!\\xff\\x0c\\x10@\\xfb\\x92"S\\xf6\\'
'xe9\\x0e@\\x1f\\xd1\\xd4\\x97G<\\x11@\\xb2\\xd2\\xf2\\x8c\\xd1.\\xd2?l\\x'
'ed\\xd5\\xe84\\xe6\\xb1\\xbf\\xea0\\xea@\\x8e\\x11\\xd9??\\xecJ\\xf9\\xad'
's\\xcc\\xbf\\xf1\\x84I\\xfb)f\\xf3?e\\x19\\xb6\\xcd\\xf4\\x96\\x0f@\\xda\\'
"xfa\\'\\xbc\\xa7Y\\xff?ox\\x13\\x93\\x9b\\xc4\\xd4\\xbf\\xb6\\t\\x96Z\\x19"
'\\xab\\xe0?l\\xed\\xd5\\xe84\\xe6\\xb1\\xbf-\\xd7\\xf2I-\\x00\\xc0\\xbf3\\'
"xa3:)>\\xa9\\xe5\\xbf\\'\\xfa\\x0f,Be\\xe9?\\x9c\\x8659\\xa0\\xd7\\x11@\\"
'x1a(e\\n9m\\xfe?\\xfc~\\x97\\xaaH\\xf3\\x0c@\\xc7\\xcc5\\xed\\r\\xc3\\xf9'
'\\xbf\\xf6b\\x1d\\xfa\\xdb\\x90\\xf1\\xbf\\x8c\\reg\\xbf\\x8e\\xef\\xbf3\\x'
'a3:)>\\xa9\\xe5\\xbfLM\\x8e\\xaa\\xaf\\xd9\\xfc\\xbfR\\x8e\\x95t\\x01m\\x'
'a7?u\\xed\\x9c\\xef\\x18\\x11\\r@\\xbf\\x17x\\xd7T\\xd7\\x10@\\xb2\\xd2\\x'
'f2\\x8c\\xd1.\\xd2?\\xa0\\xc2\\xfajW\\xf9\\xe5\\xbf\\x0b\\x84n\\xc5\\xdd\\'
'x88\\xe4\\xbf\\xa0\\xe8\\x7f\\xf0\\x1f\\\\\\xef\\xbf\\xaf\\x80\\x00\\xa5\\xd'
'4\\x14\\x03@\\xeaz\\x1fY\\xeb\\x13\\x0f@8\\xc0_\\xa1\\x87\\xfd\\r@\\xdfC\\'
'x8e\\x18-\\x06\\x12@\\x1b\\x18\\xf9\\x89Y7\\xe5\\xbfF\\xc4\\x9a\\xaf\\x87'
'\\x8d\\xec\\xbf\\x86\\xcb\\xad\\x84PJ\\xf5\\xbf\\xeaE\\xdd\\x0c\\xaf\\x82\\'
'xea\\xbf\\xb6\\x90\\x89^\\x0b\\xe7\\xd2\\xbfec\\xe2\\xe7-M\\xf9?S:\\xa6\\'
'xe0\\xbc\\x12\\x11@;\\x0c\\xde)c)\\x0c@w\\xb8\\x15\\x1e\\n\\xcb\\xec\\xbf'
'\\xf6b\\x1d\\xfa\\xdb\\x90\\xf1\\xbf\\x86\\xcb\\xad\\x84PJ\\xf5\\xbfb\\xe8'
'\\xb3M9\\xf4\\xf6\\xbfw\\x10A\\x92UU\\xa4?\\x0b\\xc4\\x97\\x9d\\xb2\\x01\\'
'x0c@u\\xed\\x9c\\xef\\x18\\x11\\r@;\\x0c\\xde)c)\\x0c@\\xc7\\xcc5\\xed\\r'
'\\xc3\\xf9\\xbfF\\xc4\\x9a\\xaf\\x87\\x8d\\xec\\xbf\\x86\\xcb\\xad\\x84PJ\\'
'xf5\\xbf\\xabE\\x11j\\xc8\\x1a\\xf2\\xbf\\x16\\x8f2\\x03\\xf0w\\xf3\\xbf\\'
'xab\\x03<\\x98\\xe1 \\xee?\\xf1\\xd0\\xc4\\x87\\x85\\x9c\\x10@\\xfc~\\x97'
'\\xaaH\\xf3\\x0c@\\x83]\\xff\\x86\\xe1?\\xc8\\xbf\\xf6b\\x1d\\xfa\\xdb\\x9'
'0\\xf1\\xbf\\x86\\xcb\\xad\\x84PJ\\xf5\\xbf\\xbd9\\x85\\xbf\\xf1`\\xf9\\xb'
'f\\x82e\\xf3C\\xb9\\xbd\\x01@\\xf5=\\xf2o\\xd8\\r\\x0e@\\xb3\\x1a\\xda=\\x'
'aa$\\x0c@\\xbb&k(\\x98\\x95\\n@\\xb2\\xd2\\xf2\\x8c\\xd1.\\xd2?l\\xed\\xd'
'5\\xe84\\xe6\\xb1\\xbf\\x0b\\x84n\\xc5\\xdd\\x88\\xe4\\xbf?\\xecJ\\xf9\\xa'
'ds\\xcc\\xbf\\xd4\\xd4\\x19\\xc3`\\xfc\\xd7?r&\\xb5\\xfe\\x1aG\\xf8?S:\\x'
'a6\\xe0\\xbc\\x12\\x11@?\\xfd\\xea\\xd8\\x1fk\\x12@\\x83]\\xff\\x86\\xe1?'
'\\xc8\\xbfF\\xc4\\x9a\\xaf\\x87\\x8d\\xec\\xbf\\x97~\\xe6\\xe8\\xd9\\x0b\\x'
"de\\xbf\\xa0\\xe8\\x7f\\xf0\\x1f\\\\\\xef\\xbf\\'C\\xa5\\xa2\\xe9\\xc7\\xfc?"
'p\\xdc\\x88\\xe4\\xe1\\x90\\x0e@\\x887\\xd4\\xdf\\xa0\\xe2\\xf8?;\\x0c\\xd'
'e)c)\\x0c@\\xc0\\x8f\\xcc%\\x83;\\xa8?\\xa0\\xc2\\xfajW\\xf9\\xe5\\xbf\\x'
"8c\\reg\\xbf\\x8e\\xef\\xbfb\\xe8\\xb3M9\\xf4\\xf6\\xbf\\'C\\xa5\\xa2\\xe9"
'\\xc7\\xfc?\\x86b.\\x12\\xbc\\x84\\x0c@\\x13\\x84\\xbb\\x96\\xe1\\x9a\\x0c'
'@{d\\n\\xac\\x13\\x87\\x0e@\\xb6\\t\\x96Z\\x19\\xab\\xe0?l\\xed\\xd5\\xe84'
'\\xe6\\xb1\\xbf\\x8c\\reg\\xbf\\x8e\\xef\\xbf3\\xa3:)>\\xa9\\xe5\\xbf\\xfb'
"\\'\\x98A\\xcep\\xfb?\\xd4\\xb7\\xf1\\xb17Z\\x05@\\xf1\\xd0\\xc4\\x87\\x85"
'\\x9c\\x10@\\xbf\\x17x\\xd7T\\xd7\\x10@{\\xef\\xb8\\xebQG\\xdb\\xbf\\xf3\\'
'x81\\xb5LN\\xca\\xde\\xbf\\x8c\\reg\\xbf\\x8e\\xef\\xbfb\\xe8\\xb3M9\\xf4'
'\\xf6\\xbf\\xce\\x0c\\x8b\\xe0\\xb2\\x19\\xfa?\\xa7I\\x08P\\x8d\\xd1\\x10@'
'\\x90g\\xe3.N&\\x10@\\xffo\\xa4Y\\x055\\x13@\\xb2\\xd2\\xf2\\x8c\\xd1.\\x'
'd2?\\xe2\\x0f\\x15\\x9e\\xa6]\\xc1?\\x97~\\xe6\\xe8\\xd9\\x0b\\xde\\xbf3\\'
'xa3:)>\\xa9\\xe5\\xbf\\x86ANG\\xceX\\xdd?\\x9d\\xe8\\xd3?\\x96x\\n@\\\\\\x'
'fc\\x03\\xac-`\\x0f@~\\x8a1X:\\xa1\\x11@w\\xb8\\x15\\x1e\\n\\xcb\\xec\\xb'
'f\\xa0\\xc2\\xfajW\\xf9\\xe5\\xbf\\xcb\\xc8i\\x96\\xce\\x0b\\xea\\xbf}\\x0'
'0\\x98E\\xcd\\xcf\\xe0\\xbf\\xc8D\\xfc\\xfd\\x047\\xec\\xbf\\xcb>K\\xb5\\x'
'83\\x16\\xf0?dX\\xf8\\xe5\\x0f\\xc4\\x11@\\x1f\\x9c\\xba\\x9a\\xddc\\x14@'
'\\x1b\\x18\\xf9\\x89Y7\\xe5\\xbf\\xa7~u\\xc3\\xed\\xa1\\xd1\\xbf\\xcb\\xc8'
'i\\x96\\xce\\x0b\\xea\\xbf\\xa0\\xe8\\x7f\\xf0\\x1f\\\\\\xef\\xbf\\xa9nA\\xa'
'c`\\\\\\xc0\\xbf\\x90%\\x01)\\xa9~\\x0b@R\\xb1\\xf8\\xe4r\\xae\\x0b@{\\xce'
'>\\xa6\\xe77\\x08@\\x1b\\x18\\xf9\\x89Y7\\xe5\\xbf\\xa7~u\\xc3\\xed\\xa1\\'
'xd1\\xbf\\x8c\\reg\\xbf\\x8e\\xef\\xbf\\xeaE\\xdd\\x0c\\xaf\\x82\\xea\\xbf'
'\\nH\\xaa\\xb4;\\x15\\xcb\\xbfr&\\xb5\\xfe\\x1aG\\xf8?\\xf1\\xd0\\xc4\\x87'
'\\x85\\x9c\\x10@?\\xfd\\xea\\xd8\\x1fk\\x12@{\\xef\\xb8\\xebQG\\xdb\\xbf\\'
'xf6b\\x1d\\xfa\\xdb\\x90\\xf1\\xbf\\xe6m+\\xed\\xc8\\x0b\\xf8\\xbf\\xabE\\'
'x11j\\xc8\\x1a\\xf2\\xbf\\xcd\\xc3\\xf5i\\x0b\\xb7\\xe6?\\x93o-C\\xe24\\x'
'f5?\\x90g\\xe3.N&\\x10@\\xe0\\x0et\\x1b\\xc3-\\x15@f\\xf5u\\x8b\\x15\\xb'
'3\\xf3?l\\xed\\xd5\\xe84\\xe6\\xb1\\xbf\\x8c\\reg\\xbf\\x8e\\xef\\xbf}\\x0'
'0\\x98E\\xcd\\xcf\\xe0\\xbf#\\x96\\xc8:\\xebs\\n@\\xd7\\x0f\\x87\\xf3\\x95'
'\\x08\\xea?\\x13\\x84\\xbb\\x96\\xe1\\x9a\\x0c@o\\xc0\\x8d<\\x16\\x16\\xff'
'?{\\xef\\xb8\\xebQG\\xdb\\xbf\\xebHeu\\xfa\\x13\\xe8?N\\xef\\xe8\\xe7W.\\'
'xa8??\\xecJ\\xf9\\xads\\xcc\\xbf\\xc8D\\xfc\\xfd\\x047\\xec\\xbf\\x86b.\\'
'x12\\xbc\\x84\\x0c@\\x99)A\\xfa\\xbes\\x0e@;\\x0c\\xde)c)\\x0c@\\x1b\\x1'
'8\\xf9\\x89Y7\\xe5\\xbf\\xa7~u\\xc3\\xed\\xa1\\xd1\\xbf\\xcb\\xc8i\\x96\\x'
'ce\\x0b\\xea\\xbf\\xabE\\x11j\\xc8\\x1a\\xf2\\xbf\\xa9nA\\xac`\\\\\\xc0\\xb'
'f\\x86b.\\x12\\xbc\\x84\\x0c@\\x13\\x84\\xbb\\x96\\xe1\\x9a\\x0c@p\\x0e\\x'
"df\\x98\\xc7\\x13\\xdb\\xbf\\x99|\\'\\xa35\\xf9\\xf5\\xbf\\xf4\\xb3\\xfec\\"
'xb6\\xa6\\x02\\xc0\\xe6m+\\xed\\xc8\\x0b\\xf8\\xbf}\\x00\\x98E\\xcd\\xcf\\'
'xe0\\xbf\\xb6\\x90\\x89^\\x0b\\xe7\\xd2\\xbf\\xd0_\\\\p\\xd9\\xab\\xf0\\xbf'
'VR\\xeax\\xbes\\xed?m_\\xdc3T\\x9f\\xf5?\\xc7\\xcc5\\xed\\r\\xc3\\xf9\\xb'
'f\\xebHeu\\xfa\\x13\\xe8?\\x1b\\xb8\\xae\\x9a\\xbd\\x0b\\xf4?O\\xb4T\\xb2'
'\\x9c\\xbd\\xcd?\\x9b\\xe0Y&B}\\xf7\\xbf\\x1c\\xa2w|$\\xe4\\xe3?\\xf9FC\\'
'xb5\\x08X\\xf3?\\xb8\\xc5\\xb9\\x1f\\xd6\\x1e\\x01@\\xf5\\x1cD7\\xe6\\x8c'
'\\xfd\\xbf^\\xb4&u\\xc2K\\x04\\xc0\\x0b\\x84n\\xc5\\xdd\\x88\\xe4\\xbf}\\x'
'00\\x98E\\xcd\\xcf\\xe0\\xbf\\x9b\\xe0Y&B}\\xf7\\xbf\\xf9\\xd7N\\x97\\x1b'
"\\'\\xe9\\xbf\\xc6aYN&\\xe9\\xe7?m*\\xc26\\xea\\xc6\\xf8?\\xf5\\x1cD7\\xe"
'6\\x8c\\xfd\\xbfl\\xed\\xd5\\xe84\\xe6\\xb1\\xbf\\xcb\\xc8i\\x96\\xce\\x0b'
'\\xea\\xbf?\\xecJ\\xf9\\xads\\xcc\\xbfLM\\x8e\\xaa\\xaf\\xd9\\xfc\\xbf\\xa'
'2 $b\\xb7\\xaa\\x92\\xbf\\xe2\\x00\\x06p\\xbf\\x17j\\xbf\\xecyi2\\x89\\x0'
"b\\xf4?\\x99|\\'\\xa35\\xf9\\xf5\\xbf\\xf4\\xb3\\xfec\\xb6\\xa6\\x02\\xc0N"
'\\xef\\xe8\\xe7W.\\xa8?\\xad~2A\\x06D\\xf2?\\xc8\\xfbf\\x87]\\xd4\\xf8\\x'
'bf\\xe3Q\\xa9iA3\\xeb\\xbfJ\\x07\\xdf\\xb1\\x03\\xc2\\xe9?k\\x94\\xf60\\x'
'bew\\xf2?\\x95E\\x84\\xd5\\xed|\\xf7?\\xe2\\x0f\\x15\\x9e\\xa6]\\xc1?\\x9'
'7~\\xe6\\xe8\\xd9\\x0b\\xde\\xbf7\\x15}\\xd7\\xb9\\xfb\\xe5?\\xf1\\x84I\\x'
'fb)f\\xf3?\\xa2 $b\\xb7\\xaa\\x92\\xbf@\\xbf\\x8b\\xcaT\\x1d\\xf6?\\xb8\\'
'xc5\\xb9\\x1f\\xd6\\x1e\\x01@\\xc3\\x95\\x92\\x1f\\xc6F\\xfb?=\\x8bJ\\xd8'
'3\\xd7\\xd5?-\\xd7\\xf2I-\\x00\\xc0\\xbf3\\xa3:)>\\xa9\\xe5\\xbf\\xf1\\x8'
'4I\\xfb)f\\xf3?\\xa2 $b\\xb7\\xaa\\x92\\xbf\\xe2\\x00\\x06p\\xbf\\x17j\\x'
'bfo\\xf5\\xa79\\x80\\xee\\xfb?\\xc7\\xcc5\\xed\\r\\xc3\\xf9\\xbf\\xc2\\xa7'
'\\xf2\\xc3]2\\xf9?\\xb5\\xe6f\\x93\\x99\\x11\\xed?\\xee\\xb7\\x1f\\xbb*\\x'
'd5\\xea?LM\\x8e\\xaa\\xaf\\xd9\\xfc\\xbf\\xbeh\\xd0\\nf\\x7f\\xdb?\\\\p\\x1'
'a65\\xf6\\xcd\\xbf7\\xeb\\xcc\\x13\\x81\\x8e\\xc5?\\xc7\\xcc5\\xed\\r\\xc3'
'\\xf9\\xbf\\x95\\xa8B\\xe6u|\\xfc?\\xf5\\xa1k\\xc2\\xa8\\x8e\\xe7?7\\x15}'
'\\xd7\\xb9\\xfb\\xe5?n\\xc5L\\xc5&&\\xf6\\xbf\\x1c\\xa2w|$\\xe4\\xe3?\\xb'
"0\\xcbB\\xc0\\xb0\\x85\\xe0?m*\\xc26\\xea\\xc6\\xf8?\\x99|\\'\\xa35\\xf9\\"
'xf5\\xbf\\x8b\\xb3\\xd6R\\xaa\\x01\\x01\\xc0\\x83*R\\x13\\x19(\\x00\\xc0t'
"\\xdc\\'\\xa3b:\\xfe\\xbf\\xbe\\xa1\\xad\\xb7`,\\xe4\\xbf\\xb6E^\\x0e\\x8d"
'K\\xef\\xbfR@\\x9a:w%\\xcd?\\xd31pS\\x8e)\\xe5?\\xb1\\xb3\\t]\\xb0\\xaf\\'
'x0e@iX\\xb1\\x9e\\xb3\\xb0\\x0e@\\x7fB?\\x8b\\xa7z\\x13@\\x1d\\xe2\\xee\\'
'xacm\\x15\\x10@\\x0eUb\\x8f\\x16\\x8e\\xa6\\xbf\\xf0\\x95,!p\\xfc\\xe7?7'
'}\\xa8\\xa2\\xbd\\x92\\xe2\\xbf\\xd7\\xc4\\x8f\\x1a\\xdf\\xea\\xcc\\xbf>\\x'
'eb\\xe5\\xa3\\x137\\x05@\\x95\\xa8B\\xe6u|\\xfc?\\xbb\\x1512EJ\\xf1?\\x1'
'6\\xdc\\xf5\\xae\\x88\\x9e\\x01@w\\x10A\\x92UU\\xa4?\\x8f\\\\\\x85\\xaf\\xb'
'1\\x97\\xdf?}\\xec\\xc8\\x18\\xe60\\xf5?\\x9d\\x0b\\xb2\\x8fl\\x16\\xd1?\\'
'x84c\\xfb\\x12\\xd8\\xe5\\n@ZV\\xe9Hww\\x06@\\x1eY\\n\\x0c\\xf5-\\t@q-\\x'
'c7 A\\x0b\\x04@\\xd4\\xd4\\x19\\xc3`\\xfc\\xd7?\\xc0{x\\x9e\\x96\\x1c\\xf'
"1?\\xc6aYN&\\xe9\\xe7?\\xd31pS\\x8e)\\xe5?\\'\\xc3\\xde~\\'R\\x03@\\x87U"
'\\x99&_-\\x03@\\xde\\x9d\\x05\\xdd\\xe5\\xb0\\x0e@\\xd5xR=\\x0e\\x1b\\n@%'
'h\\xe5>\\xf3\\x9f\\xd2?\\xc0{x\\x9e\\x96\\x1c\\xf1?\\\\p\\x1a65\\xf6\\xcd\\'
'xbf\\xa17I\\x9b\\xc4\\xb4\\xdd?\\xf1\\xe5\\xa0i\\x9e\\x10\\xff?h\\xa9\\x9'
'2\\x08\\x8e\\xc6\\xff?~\\xfb\\x87tm\\xef\\x0b@\\x16\\xdc\\xf5\\xae\\x88\\x'
'9e\\x01@\\x0eUb\\x8f\\x16\\x8e\\xa6\\xbf\\xcb>K\\xb5\\x83\\x16\\xf0?8\\x8'
"3\\x18b\\xd5\\xac\\xd2\\xbf\\xd31pS\\x8e)\\xe5?\\'\\xc3\\xde~\\'R\\x03@h\\"
'xa9\\x92\\x08\\x8e\\xc6\\xff?~\\xfb\\x87tm\\xef\\x0b@\\xcc~\\x98\\x92\\xf'
'9w\\x06@w\\x10A\\x92UU\\xa4?\\x06\\x1c\\xd2NJ\\xf0\\xe5?B\\xc8\\xb3ixD\\'
"xe6\\xbfox\\x13\\x93\\x9b\\xc4\\xd4\\xbf\\'\\xc3\\xde~\\'R\\x03@\\x1dUq\\x"
'15S\\x88\\x01@\\xde\\x9d\\x05\\xdd\\xe5\\xb0\\x0e@\\xd5xR=\\x0e\\x1b\\n@%'
'h\\xe5>\\xf3\\x9f\\xd2?\\xcb>K\\xb5\\x83\\x16\\xf0?_\\x88\\xda3\\x94^\\xb'
'e\\xbf\\xd1f\\x8aP\\xf8\\x01\\xe2?f\\xf5u\\x8b\\x15\\xb3\\xf3?\\x91J\\x05'
'\\xba*\\xa8\\xee?N\\xef\\xe8\\xe7W.\\xa8?\\xca\\xc2\\x80\\xd5\\xd4\\x1b\\x'
'b2\\xbf\\xaf\\x80\\x00\\xa5\\xd4\\x14\\x03@\\xa8\\xf5\\xd2p\\xbc(\\xf3?L\\'
"x10\\x07Q\\'\\xe9\\xf9\\xbf\\x86f\\x0b\\x1bxq\\x9d\\xbff\\xf5u\\x8b\\x15\\"
'xb3\\xf3?\\xf0U\\xc17k\\xd2\\x04@\\x1b\\xb8\\xae\\x9a\\xbd\\x0b\\xf4?\\x9'
'4\\x9fo \\xb0\\x91\\xd8?T^\\xb2\\x03\\x05\\x1f\\xfe?\\x82\\n\\x99\\xd0]\\x'
'd6\\xf7\\xbf\\x919\\xb8\\xe8y\\x0f\\x01@k\\x94\\xf60\\xbew\\xf2?\\x10\\x9'
'b\\xd7Y;m\\x01@\\x1b\\xa6R\\x7f-\\x9e\\xf2?\\xea0\\xea@\\x8e\\x11\\xd9?\\'
'xca\\xc2\\x80\\xd5\\xd4\\x1b\\xb2\\xbf\\xce\\x0c\\x8b\\xe0\\xb2\\x19\\xfa?'
'\\x15\\x81f\\xc1\\xceN\\xd3?\\x887\\xd4\\xdf\\xa0\\xe2\\xf8?\\xf8R\\x00\\x'
'9f\\xf0T\\x00@w\\xb8\\x15\\x1e\\n\\xcb\\xec\\xbfBf]\\x83<\\xb9\\xfe\\xbf\\'
'x0b\\x84n\\xc5\\xdd\\x88\\xe4\\xbf\\x94\\x9fo \\xb0\\x91\\xd8?\\xc8D\\xfc'
'\\xfd\\x047\\xec\\xbf4\\xb5\\xd50\\xe2\\x00\\xff\\xbf@\\xbf\\x8b\\xcaT\\x1'
"d\\xf6?\\x9f\\xa1}\\x95\\x98e\\xd7?\\'\\xc3\\xde~\\'R\\x03@\\x1b\\xa6R\\x7"
'f-\\x9e\\xf2?\\xd4N\\xe7=Y\\x17\\xcc?\\xca\\xc2\\x80\\xd5\\xd4\\x1b\\xb2\\'
'xbf`\\xed4)Bq\\x08@\\xbeh\\xd0\\nf\\x7f\\xdb?@\\xbf\\x8b\\xcaT\\x1d\\xf6'
'?9\\x1d;\\xd2\\x0f\\xd9\\xe3\\xbf\\xc7\\xcc5\\xed\\r\\xc3\\xf9\\xbfl\\xed\\'
'xd5\\xe84\\xe6\\xb1\\xbf\\x97~\\xe6\\xe8\\xd9\\x0b\\xde\\xbf3\\xa3:)>\\xa'
'9\\xe5\\xbf\\xe9s%\\xa2\\xd4 \\xf2\\xbf\\x86%\\xa6\\x98\\xe2\\xbe\\x00@\\x'
"17\\xdf=LW\\xe8\\x02@o\\xf5\\xa79\\x80\\xee\\xfb?\\x99|\\'\\xa35\\xf9\\xf"
'5\\xbf\\x1b\\xa6R\\x7f-\\x9e\\xf2?\\xd4N\\xe7=Y\\x17\\xcc?\\x81r\\xda\\xf'
'3H"\\xe1?\\xc8\\xfbf\\x87]\\xd4\\xf8\\xbf/\\x94\\xc4u\\x8ck\\xfe?@\\xbc\\'
'xd3\\xeaH\\x10\\xe6?\\x9f\\xef\\xce\\xf1Ic\\xf3\\xbfoJ\\xcf\\x82z\\xd2\\x'
'ef?\\x91J\\x05\\xba*\\xa8\\xee?{Z,\\x036\\xcd\\xf6?\\x81r\\xda\\xf3H"\\x'
'e1?t\\x8d\\xdb\\xa7\\xd4\\x08\\xe4?{b\\xd3\\x81\\xf5\\xc4\\x01@M\\x13\\xb'
'f03\\xf6\\xe9\\xbf;\\xe8 \\xd5\\xa5\\x00\\xe7\\xbf\\x10\\x9b\\xd7Y;m\\x01'
'@\\x1b\\xa6R\\x7f-\\x9e\\xf2?\\xd4N\\xe7=Y\\x17\\xcc?3\\xa3:)>\\xa9\\xe5'
'\\xbf\\xb9#O\\xebx\\x1f\\x0b@\\x06\\x1c\\xd2NJ\\xf0\\xe5?}\\xec\\xc8\\x18'
'\\xe60\\xf5?m_\\xdc3T\\x9f\\xf5?\\x1b\\x18\\xf9\\x89Y7\\xe5\\xbfEG\\xc50'
'\\xca\\x7f\\xe1?\\x17\\xf5\\xefF\\xf8\\x05\\xd3\\xbf}\\x00\\x98E\\xcd\\xcf'
"\\xe0\\xbf\\'C\\xa5\\xa2\\xe9\\xc7\\xfc?p\\x9f\\x00k\\x08\\xcb\\x02@}\\xec"
'\\xc8\\x18\\xe60\\xf5?\\xee\\x0f58\\xb5Z\\xfa?\\x13\\xaa\\xb2\\xee\\xc9>\\'
'xe8?\\xf3\\x81\\xb5LN\\xca\\xde\\xbf\\x0b\\x84n\\xc5\\xdd\\x88\\xe4\\xbfO'
'\\xb4T\\xb2\\x9c\\xbd\\xcd?\\x1d\\xa0V\\\\E\\xbd\\xf4?/\\x94\\xc4u\\x8ck\\x'
"fe?\\x17\\xdf=LW\\xe8\\x02@\\x9f\\xef\\xce\\xf1Ic\\xf3\\xbf\\'\\xc3\\xde~"
"\\'R\\x03@\\x1b\\xa6R\\x7f-\\x9e\\xf2?\\xd4N\\xe7=Y\\x17\\xcc?O\\xb4T\\xb"
'2\\x9c\\xbd\\xcd?P\\xb1\\xd5\\x9b\\x06\\xcb\\x0b@\\xd7\\x0f\\x87\\xf3\\x95'
'\\x08\\xea?\\xd1\\xaf\\x1c\\xf5\\xec\\xa7\\xfb?8RU\\xcfy\\xb1\\xe0\\xbf\\x'
"99|\\'\\xa35\\xf9\\xf5\\xbf\\xc9cm\\x1c\\xf4\\xda\\xf4\\xbf\\x17\\xf5\\xef"
'F\\xf8\\x05\\xd3\\xbf\\x94\\x9fo \\xb0\\x91\\xd8?\\xe9s%\\xa2\\xd4 \\xf2\\'
'xbf\\x8f\\\\\\x85\\xaf\\xb1\\x97\\xdf?T\\x0c{\\x9a\\xe8\\xfb\\x01@\\xa0\\xb'
'a\\xb4\\xf4\\xdf\\x8a\\xf6\\xbf\\xb2\\xd2\\xf2\\x8c\\xd1.\\xd2?\\xa0\\xc2\\'
'xfajW\\xf9\\xe5\\xbf\\xd4N\\xe7=Y\\x17\\xcc?\\x8c\\xbb\\xea\\xc3\\xb8\\xe'
'c\\xd7\\xbf\\xfe:\\xf2\\xd9`\\x9c\\xbf?\\xc1\\x89\\xe1\\xc5\\xbb\\x14\\xec'
'?V\\x0f3z\\xf4\\x08\\x02\\xc0\\x92KR\\x82L\\xf2\\x03\\xc0\\xb2\\xd2\\xf2\\'
'x8c\\xd1.\\xd2?=\\x8bJ\\xd83\\xd7\\xd5?N\\xef\\xe8\\xe7W.\\xa8?\\xee\\xb'
'7\\x1f\\xbb*\\xd5\\xea?\\xc1\\x05\\xf5\\xa8\\xb7\\x1d\\x03\\xc0\\x9e2\\x00'
'Z\\xcf.\\xf4?\\xd1\\xaf\\x1c\\xf5\\xec\\xa7\\xfb?\\xa2P\\x80\\xfa\\x0b\\xd'
'a\\xfc\\xbf\\x13\\xaa\\xb2\\xee\\xc9>\\xe8?\\xa7~u\\xc3\\xed\\xa1\\xd1\\xb'
'f\\x97~\\xe6\\xe8\\xd9\\x0b\\xde\\xbf\\x94\\x9fo \\xb0\\x91\\xd8?%h\\xe5>'
'\\xf3\\x9f\\xd2?:W\\x97\\x8cye\\xfd?\\x9d\\x84\\xc3\\xaf4\\xc1\\x04@9\\x1'
'd;\\xd2\\x0f\\xd9\\xe3\\xbfw\\xb8\\x15\\x1e\\n\\xcb\\xec\\xbf\\x9cd\\xbd>\\'
'x0c%\\xf8\\xbf\\x0b\\x84n\\xc5\\xdd\\x88\\xe4\\xbf\\xeaR\\x94G\\xb2\\xaf\\'
"xb4?\\'C\\xa5\\xa2\\xe9\\xc7\\xfc?\\xbeh\\xd0\\nf\\x7f\\xdb?#*I\\x13\\x12"
'\\x9a\\x06@7\\xeb\\xcc\\x13\\x81\\x8e\\xc5?\\xc7\\xcc5\\xed\\r\\xc3\\xf9\\x'
'bf\\x8b\\xb3\\xd6R\\xaa\\x01\\x01\\xc0\\x0b\\x84n\\xc5\\xdd\\x88\\xe4\\xbf'
'\\xeaR\\x94G\\xb2\\xaf\\xb4?B\\xaa?d\\x0b\\xcf\\xf4\\xbfR\\x8e\\x95t\\x01'
"m\\xa7?T\\x0c{\\x9a\\xe8\\xfb\\x01@p\\x0e\\xdf\\x98\\xc7\\x13\\xdb\\xbf\\'"
"\\xc3\\xde~\\'R\\x03@\\x1b\\xa6R\\x7f-\\x9e\\xf2?\\xd4N\\xe7=Y\\x17\\xcc?"
'\\x81r\\xda\\xf3H"\\xe1?5\\xc8\\xc17\\xf3=\\x10@\\xcb>K\\xb5\\x83\\x16\\x'
'f0?WU\\xa2X\\xca\\x80\\xfd?ox\\x13\\x93\\x9b\\xc4\\xd4\\xbf\\xf1\\xe5\\xa'
'0i\\x9e\\x10\\xff?\\xebHeu\\xfa\\x13\\xe8?-\\xd7\\xf2I-\\x00\\xc0\\xbf\\x'
"94\\x9fo \\xb0\\x91\\xd8?4\\xd2\\'\\xc8&\\x1a\\x07@n\\x96\\x1eM\\xaf\\x17"
'\\xbc?6t\\x80\\x03\\x9ak\\xf2?\\xd1\\r\\xb3\\xfe\\xd0\\x00\\x00\\xc0\\x95E'
'\\x84\\xd5\\xed|\\xf7?\\xebHeu\\xfa\\x13\\xe8?\\xea0\\xea@\\x8e\\x11\\xd9'
'?\\xca\\xc2\\x80\\xd5\\xd4\\x1b\\xb2\\xbf\\x13Zi\\xad\\xaf\\xcd\\r@\\xe0z\\'
'xc4\\xc8$T\\x04@\\xb6u\\\\\\xf3\\x1fr\\x02@\\xec\\xae\\x83/\\xf3\\xe3\\xf0'
'?\\xb2\\xd2\\xf2\\x8c\\xd1.\\xd2?\\xa0\\xc2\\xfajW\\xf9\\xe5\\xbf\\x97~\\x'
"e6\\xe8\\xd9\\x0b\\xde\\xbf?\\xecJ\\xf9\\xads\\xcc\\xbf\\xfb\\'\\x98A\\xce"
'p\\xfb?p\\x9f\\x00k\\x08\\xcb\\x02@`W\\x86a\\xa3\\xad\\x05@\\x9f\\xa1}\\x'
"95\\x98e\\xd7?\\x99|\\'\\xa35\\xf9\\xf5\\xbf\\xa0\\xc2\\xfajW\\xf9\\xe5\\x"
'bf\\xcb\\xc8i\\x96\\xce\\x0b\\xea\\xbf}\\x00\\x98E\\xcd\\xcf\\xe0\\xbf\\xd'
'd\\xe4\\xa2|\\x97\\xce\\x07\\xc0\\x93o-C\\xe24\\xf5?\\x0e\\xddYC~\\xbb\\x'
'fa?\\xd4nd\\x07\\x93w\\t\\xc0w\\xb8\\x15\\x1e\\n\\xcb\\xec\\xbf\\x9cd\\xbd'
'>\\x0c%\\xf8\\xbf\\xcb\\xc8i\\x96\\xce\\x0b\\xea\\xbf}\\x00\\x98E\\xcd\\xc'
"f\\xe0\\xbf\\x81y\\xbfd v\\xff?\\xf9\\xd7N\\x97\\x1b\\'\\xe9\\xbf\\xcff\\x"
'f56\\x0b#\\x00@\\xa2\\x85\\x9a\\xf7u\\xb2\\xf9\\xbf\\x83]\\xff\\x86\\xe1?'
'\\xc8\\xbfF\\xc4\\x9a\\xaf\\x87\\x8d\\xec\\xbf\\xd4N\\xe7=Y\\x17\\xcc?\\xa'
'4Z\\xc2\\x9e\\x9b\\xae\\xef?i\\xfd\\xbd\\xe2xC\\xd8\\xbf\\xa8\\xf5\\xd2p\\'
'xbc(\\xf3?\\xc2\\xc0g\\xba\\xda#\\x06@7\\xeb\\xcc\\x13\\x81\\x8e\\xc5?oJ'
'\\xcf\\x82z\\xd2\\xef?\\xef\\xa6\\xa2\\xa1E\\xe8\\xf5?\\xea0\\xea@\\x8e\\x'
'11\\xd9?3\\xa3:)>\\xa9\\xe5\\xbfq)\\x94\\xb6}\\x17\\x05@\\xfe\\xf8_Rq\\x'
'bc\\xd9\\xbf\\x8b=D\\x9f\\xb8\\xfc\\xf8\\xbf\\x9f\\xef\\xce\\xf1Ic\\xf3\\x'
'bf\\xb6\\t\\x96Z\\x19\\xab\\xe0?\\xf3\\x81\\xb5LN\\xca\\xde\\xbf\\xcb\\xc8'
'i\\x96\\xce\\x0b\\xea\\xbf\\x8c\\xbb\\xea\\xc3\\xb8\\xec\\xd7\\xbf\\x98\\xa'
'1*@\\xce\\xf6\\x1a@\\xb9\\xd9\\xb6B\\xff\\xb7\\xf2\\xbf\\xb6u\\\\\\xf3\\x1f'
'r\\x02@\\xd31pS\\x8e)\\xe5?\\xb6\\t\\x96Z\\x19\\xab\\xe0?\\xf3\\x81\\xb5L'
'N\\xca\\xde\\xbf\\x97~\\xe6\\xe8\\xd9\\x0b\\xde\\xbf\\xabE\\x11j\\xc8\\x1a'
'\\xf2\\xbf\\x19\\xf3y\\xf4Fi\\x02@X\\x11\\xf6\\x08\\xda\\x8b\\xd1\\xbf\\xc'
'6aYN&\\xe9\\xe7?\\xa2P\\x80\\xfa\\x0b\\xda\\xfc\\xbf\\xb6\\t\\x96Z\\x19\\x'
'ab\\xe0?\\xebHeu\\xfa\\x13\\xe8?-\\xd7\\xf2I-\\x00\\xc0\\xbf\\xabE\\x11j'
'\\xc8\\x1a\\xf2\\xbfE\\x0e\\x87Ub\\xc0\\x03@\\xeat\\x1bf\\x1ag\\xd7?7z\\x'
'f0\\xc2\\xb1\\x85\\xf2\\xbfp\\x0e\\xdf\\x98\\xc7\\x13\\xdb\\xbfw\\xb8\\x15'
'\\x1e\\n\\xcb\\xec\\xbf\\x9cd\\xbd>\\x0c%\\xf8\\xbfF\\x10\\xa9UA\\xcd\\xfa'
'\\xbf3\\xa3:)>\\xa9\\xe5\\xbf\\x17\\xd8\\xc7y\\x97\\xda\\xe6\\xbf\\xf0\\x9'
'5,!p\\xfc\\xe7?0\\xd0\\xd6\\x8fB\\x99\\x00@:\\xb3\\x06\\xd8;(\\xea\\xbf\\'
'xc7\\xcc5\\xed\\r\\xc3\\xf9\\xbf\\x9cd\\xbd>\\x0c%\\xf8\\xbf\\xa6\\xb2&\\x'
'be\\xb9\\x8e\\xfd\\xbf\\xeaE\\xdd\\x0c\\xaf\\x82\\xea\\xbf\\x0eUb\\x8f\\x1'
'6\\x8e\\xa6\\xbf\\xcd\\xcb\\x03<g?\\xed\\xbfs\\xa1\\xbdQ+\\x7f\\xf1?o\\xc'
'0\\x8d<\\x16\\x16\\xff?j,\\x19Y]/\\xf2\\xbfF\\xc4\\x9a\\xaf\\x87\\x8d\\xe'
'c\\xbf\\xea0\\xea@\\x8e\\x11\\xd9?O\\xb4T\\xb2\\x9c\\xbd\\xcd?\\xc1\\x05\\'
"xf5\\xa8\\xb7\\x1d\\x03\\xc0\\xf9\\xd7N\\x97\\x1b\\'\\xe9\\xbfl\\xb8Z/R/\\"
"xaf\\xbf#6\\xf3\\xfb\\xd6m\\xfe\\xbf\\x99|\\'\\xa35\\xf9\\xf5\\xbf\\xf3\\x"
'81\\xb5LN\\xca\\xde\\xbf\\x8c\\reg\\xbf\\x8e\\xef\\xbfb\\xe8\\xb3M9\\xf4\\'
'xf6\\xbf\\xff\\\\a\\x97\\x0e\\x1b\\x01\\xc0\\xc5\\x9c\\x89Y\\xec\\xb1\\xf1\\'
'xbfR@\\x9a:w%\\xcd?k\\x94\\xf60\\xbew\\xf2?w\\xb8\\x15\\x1e\\n\\xcb\\xec'
'\\xbf\\x9cd\\xbd>\\x0c%\\xf8\\xbf\\x8c\\reg\\xbf\\x8e\\xef\\xbf\\x8c\\xbb\\'
'xea\\xc3\\xb8\\xec\\xd7\\xbf\\xb6\\x90\\x89^\\x0b\\xe7\\xd2\\xbf\\xf2/\\x8'
'9H\\xb3\\x15\\x03\\xc0\\xd2\\xb8D\\x94\\x10\\xcf\\xeb\\xbf\\x94\\xe1\\x1d\\'
'x88xA\\n\\xc0\\x83]\\xff\\x86\\xe1?\\xc8\\xbf\\xa0\\xc2\\xfajW\\xf9\\xe5\\'
'xbf\\x17\\xf5\\xefF\\xf8\\x05\\xd3\\xbf\\xca\\xc2\\x80\\xd5\\xd4\\x1b\\xb2'
'\\xbf\\xe5\\xf6au\\x0b\\x87\\xca?\\x83T\\xc5\\xea\\x96\\x8c\\x01\\xc0\\xd2'
'\\xb8D\\x94\\x10\\xcf\\xeb\\xbfS\\xbe\\x0b\\x032\\xbc\\x04\\xc0\\xf1\\xe5\\'
'xa0i\\x9e\\x10\\xff?\\xc2\\xa7\\xf2\\xc3]2\\xf9?~\\xfb\\x87tm\\xef\\x0b@'
"q-\\xc7 A\\x0b\\x04@\\'C\\xa5\\xa2\\xe9\\xc7\\xfc?\\xb6E^\\x0e\\x8dK\\xef"
'\\xbf7z\\xf0\\xc2\\xb1\\x85\\xf2\\xbf\\x8d\\xf9\\x93\\xca\\xcd9\\x14\\xc0w'
'\\xb8\\x15\\x1e\\n\\xcb\\xec\\xbf\\x9cd\\xbd>\\x0c%\\xf8\\xbf\\xea0\\xea@\\'
'x8e\\x11\\xd9?}\\x00\\x98E\\xcd\\xcf\\xe0\\xbf\\xb9#O\\xebx\\x1f\\x0b@\\x'
"eb=\\x97\\xdf\\x11N\\x03@\\xdc\\x00\\x98{\\xbfs\\xff\\xbf!\\xa0\\'\\xf6\\x"
'aa\\x1e\\xf8\\xbfw\\xb8\\x15\\x1e\\n\\xcb\\xec\\xbfEG\\xc50\\xca\\x7f\\xe1'
'?\\xea0\\xea@\\x8e\\x11\\xd9?\\xca\\xc2\\x80\\xd5\\xd4\\x1b\\xb2\\xbf\\xe5'
'\\xf6au\\x0b\\x87\\xca?\\xa4\\xda\\xc5\\xa87@\\n\\xc0\\xfaL\\xb3t r\\xf3\\'
'xbf\\x9f$\\xe9\\xee\\xb3;\\xf0\\xbff\\xf5u\\x8b\\x15\\xb3\\xf3?\\x1dUq\\x'
"15S\\x88\\x01@\\xbb\\x1512EJ\\xf1?v\\x15Iz\\xa0c\\xfe?\\'\\xfa\\x0f,Be\\"
"xe9?\\xd1\\xe6\\x10\\x04\\xec\\'\\x06\\xc0\\xb2\\xd4j_\\xd4\\xac\\xf0\\xbf"
'\\xd1f\\x8aP\\xf8\\x01\\xe2?\\xb6\\t\\x96Z\\x19\\xab\\xe0?\\xc2\\xa7\\xf2\\'
"xc3]2\\xf9?\\xf5\\xa1k\\xc2\\xa8\\x8e\\xe7?\\xad~2A\\x06D\\xf2?\\'\\xfa\\"
'x0f,Be\\xe9?b\\x0bM\\xa6\\xcf\\x9e\\x04\\xc0t\\xa7-\\x11C\\x99\\xf1\\xbf'
'\\x9f\\xa1}\\x95\\x98e\\xd7?\\xb6\\t\\x96Z\\x19\\xab\\xe0?\\xef\\xa6\\xa2\\'
'xa1E\\xe8\\xf5?\\x1b\\xb8\\xae\\x9a\\xbd\\x0b\\xf4?\\x94\\x9fo \\xb0\\x91'
'\\xd8?t\\x8d\\xdb\\xa7\\xd4\\x08\\xe4?WHz\\x8f\\xe2\\xa4\\x05\\xc0L\\x10\\'
"x07Q\\'\\xe9\\xf9\\xbf\\xd7\\xc4\\x8f\\x1a\\xdf\\xea\\xcc\\xbf{\\xef\\xb8\\"
'xebQG\\xdb\\xbf\\xe2\\x0f\\x15\\x9e\\xa6]\\xc1?\\x0b\\x84n\\xc5\\xdd\\x88'
'\\xe4\\xbf7\\x15}\\xd7\\xb9\\xfb\\xe5?\\x0eUb\\x8f\\x16\\x8e\\xa6\\xbf\\x8'
'f\\x17\\x98\\x01\\x84\\x86\\x00\\xc0\\x93<p\\xc8\\x85\\x1c\\x01\\xc0<~\\xe'
'c\\xda\\xd1O\\xed\\xbf\\x13\\xaa\\xb2\\xee\\xc9>\\xe8?\\xf0U\\xc17k\\xd2\\'
"x04@;\\x9f\\'\\xd4&P\\xfc?7\\x15}\\xd7\\xb9\\xfb\\xe5?\\xf1\\x84I\\xfb)f"
'\\xf3?4\\xb5\\xd50\\xe2\\x00\\xff\\xbf7z\\xf0\\xc2\\xb1\\x85\\xf2\\xbf\\x9'
'f\\xa1}\\x95\\x98e\\xd7?\\xfc\\x95\\x0ff:/\\x12@\\x00X\\x89\\x8d\\xa7\\x0'
'b\\r@\\x1eY\\n\\x0c\\xf5-\\t@\\x16\\xdc\\xf5\\xae\\x88\\x9e\\x01@\\x98N/9\\'
'xf3\\xb7\\xf0?\\x01\\xc4<\\r\\xecA\\x01@C\\xce#)\\x90^\\xd6\\xbf8RU\\xcf'
'y\\xb1\\xe0\\xbfV\\x13\\xed\\xc8\\xff\\x1b\\x07@\\x95\\xa8B\\xe6u|\\xfc?\\'
'x1b\\xb8\\xae\\x9a\\xbd\\x0b\\xf4?\\xad~2A\\x06D\\xf2?\\x98N/9\\xf3\\xb7'
'\\xf0?n\\x96\\x1eM\\xaf\\x17\\xbc?\\x1a(e\\n9m\\xfe?8RU\\xcfy\\xb1\\xe0\\'
'xbf\\xc0\\x8f\\xcc%\\x83;\\xa8?\\xf0U\\xc17k\\xd2\\x04@\\x9bA\\xa5<\\x9f'
'\\x11\\xff?\\xbfr\\xa6\\x96/\\x8a\\xf9?\\x86ANG\\xceX\\xdd?\\x15\\x81f\\x'
'c1\\xceN\\xd3?#*I\\x13\\x12\\x9a\\x06@\\xf9\\x1d\\xe6\\xa1\\x86|\\x03@f\\'
'xf5u\\x8b\\x15\\xb3\\xf3?=\\x8bJ\\xd83\\xd7\\xd5?\\x97~\\xe6\\xe8\\xd9\\x'
'0b\\xde\\xbf}\\x00\\x98E\\xcd\\xcf\\xe0\\xbf\\xb6\\x90\\x89^\\x0b\\xe7\\xd'
'2\\xbf\\x06\\x1c\\xd2NJ\\xf0\\xe5?VR\\xeax\\xbes\\xed?\\xa2P\\x80\\xfa\\x'
"0b\\xda\\xfc\\xbf\\'\\xc3\\xde~\\'R\\x03@\\xc2\\xa7\\xf2\\xc3]2\\xf9?\\xf5"
'\\xa1k\\xc2\\xa8\\x8e\\xe7?O\\xb4T\\xb2\\x9c\\xbd\\xcd?\\xc8D\\xfc\\xfd\\x'
'047\\xec\\xbf\\x06\\x1c\\xd2NJ\\xf0\\xe5?\\xc6aYN&\\xe9\\xe7?\\x1e\\n\\\\\\'
'xf0~\\xcf\\xf1\\xbf\\x95E\\x84\\xd5\\xed|\\xf7?\\x91J\\x05\\xba*\\xa8\\xe'
'e?\\xb5\\xe6f\\x93\\x99\\x11\\xed?\\x81r\\xda\\xf3H"\\xe1?\\x1dW\\xc1\\xe'
'5\\x9dZ\\xe1?\\xc9\\xf4\\x1e\\x9bJ`\\x06@TLz\\xb9\\xa6Y\\xdd?m_\\xdc3T\\'
'x9f\\xf5?j,\\x19Y]/\\xf2\\xbf\\x9cd\\xbd>\\x0c%\\xf8\\xbf\\x83*R\\x13\\x'
'19(\\x00\\xc0C\\xe8M\\xfcE\\xc0\\x02\\xc0J\\xbbc\\xbd`\\x14\\xf6?\\xd7\\x'
'0f\\x87\\xf3\\x95\\x08\\xea?\\xb0\\xce\\xfa\\x9f\\xbc\\x92\\xf0?\\x9f\\xa1'
"}\\x95\\x98e\\xd7?\\x99|\\'\\xa35\\xf9\\xf5\\xbfBf]\\x83<\\xb9\\xfe\\xbf\\"
'xa6\\xb2&\\xbe\\xb9\\x8e\\xfd\\xbfC\\xe8M\\xfcE\\xc0\\x02\\xc0i\\xfd\\xbd'
'\\xe2xC\\xd8\\xbf\\x1c\\xa2w|$\\xe4\\xe3?@\\xbc\\xd3\\xeaH\\x10\\xe6?\\xa'
'17I\\x9b\\xc4\\xb4\\xdd?\\xc7\\xcc5\\xed\\r\\xc3\\xf9\\xbfBf]\\x83<\\xb9\\'
'xfe\\xbf\\xb3\\xfb\\x90G\\xd5\\x88\\x01\\xc0\\x18\\x8bV1\\xaa\\xcd\\xfb\\x'
'bfp\\x0e\\xe2;\\xce\\x88\\xe9\\xbf\\x8f\\\\\\x85\\xaf\\xb1\\x97\\xdf?J\\x07'
'\\xdf\\xb1\\x03\\xc2\\xe9?R\\xf3%\\x00\\x9c\\x94\\x01\\xc0\\xc0\\x8f\\xcc%'
'\\x83;\\xa8?\\xa7~u\\xc3\\xed\\xa1\\xd1\\xbf\\x97~\\xe6\\xe8\\xd9\\x0b\\xd'
'e\\xbf\\x07\\x97\\xe2\\xdb\\x80\\x87\\xf4\\xbf\\xa1\\xf1}\\x7f\\x97\\xc2\\x'
'f8?`\\xc1 \\x8c\\x96\\xe8\\xfa\\xbfV\\x0f3z\\xf4\\x08\\x02\\xc0\\x92KR\\x'
"82L\\xf2\\x03\\xc0\\x99|\\'\\xa35\\xf9\\xf5\\xbf\\xc2\\xa7\\xf2\\xc3]2\\xf"
'9?\\x9bA\\xa5<\\x9f\\x11\\xff?d!\\xd5$w\\x1d\\xf7?\\x9b\\xe0Y&B}\\xf7\\x'
'bf\\xd2\\xec\\x14\\xf7\\xbc\\xd4\\xdd\\xbf\\x05\\x98\\xbe;\\xdb#\\xf7\\xbf'
'!k\\r\\xf9@F\\xfb\\xbf\\x83]\\xff\\x86\\xe1?\\xc8\\xbfZV\\xe9Hww\\x06@\\x'
'fdq\\x91\\xd2\\x8b\\xe9\\x00@\\x1a\\xc4w\\x08\\xe8\\xf6\\xfb?\\xd4\\xd4\\x'
'19\\xc3`\\xfc\\xd7?\\xf0\\x95,!p\\xfc\\xe7??\\xb6c+1\\xf6\\xd5?o\\xc0\\x'
'8d<\\x16\\x16\\xff?m;\\xf4\\xed\\xeb\\x00\\t@\\x1dUq\\x15S\\x88\\x01@;\\x'
"9f\\'\\xd4&P\\xfc?\\x16\\xdc\\xf5\\xae\\x88\\x9e\\x01@\\xfe:\\xf2\\xd9`\\x"
'9c\\xbf?\\x15\\x81f\\xc1\\xceN\\xd3?D\\xf8YAZ\\xec\\xab?\\xa17I\\x9b\\xc'
'4\\xb4\\xdd?\\xc3\\x95\\x92\\x1f\\xc6F\\xfb?\\x87U\\x99&_-\\x03@\\xdb\\xf'
'c\\xa9k\\xae\\x8e\\xf9?\\t\\xd0\\x03\\xb3\\xbe\\xb0\\xf4?\\x17\\xd8\\xc7y\\'
'x97\\xda\\xe6\\xbfwG\\xc6\\xb9p\\xdc\\xf8\\xbf\\x1a.\\xd5\\xc9P\\x87\\xfe'
'\\xbfT\\x89\\xf1\\x05\\xc8\\xe3\\x07\\xc0j,\\x19Y]/\\xf2\\xbf\\x87U\\x99&'
'_-\\x03@\\x9bA\\xa5<\\x9f\\x11\\xff?\\x1a\\xc4w\\x08\\xe8\\xf6\\xfb?\\x16'
'\\x8f2\\x03\\xf0w\\xf3\\xbf_S\\x18\\x7f\\x85\\xb6\\xc2\\xbfYd:\\xb7\\x05\\'
'xc2\\xdd\\xbf7\\xeb\\xcc\\x13\\x81\\x8e\\xc5?\\xc7\\xcc5\\xed\\r\\xc3\\xf9'
'\\xbf\\xf6b\\x1d\\xfa\\xdb\\x90\\xf1\\xbf&)0\\x1c\\xd8\\x88\\xf2\\xbf\\xa0'
'\\xe8\\x7f\\xf0\\x1f\\\\\\xef\\xbfek\\x93\\xf5)~\\xe1\\xbfp\\x9f\\x00k\\x08'
'\\xcb\\x02@_\\x88\\xda3\\x94^\\xbe\\xbf<~\\xec\\xda\\xd1O\\xed\\xbf\\xc0\\'
'x8f\\xcc%\\x83;\\xa8?\\xf6b\\x1d\\xfa\\xdb\\x90\\xf1\\xbf&)0\\x1c\\xd8\\x'
'88\\xf2\\xbf\\xa0\\xe8\\x7f\\xf0\\x1f\\\\\\xef\\xbf\\x00\\x0f\\x05\\x10\\xba'
'\\x93\\x10@Y\\x19[=.\\xd7\\x04@\\xe2\\x00\\x06p\\xbf\\x17j\\xbfp\\x0e\\xd'
'f\\x98\\xc7\\x13\\xdb\\xbf\\x83]\\xff\\x86\\xe1?\\xc8\\xbf\\xa0\\xc2\\xfaj'
'W\\xf9\\xe5\\xbf\\x86\\xcb\\xad\\x84PJ\\xf5\\xbf\\x07\\x97\\xe2\\xdb\\x80\\'
'x87\\xf4\\xbf\\x13Zi\\xad\\xaf\\xcd\\r@R\\x8e\\x95t\\x01m\\xa7?_\\x88\\xd'
'a3\\x94^\\xbe\\xbfox\\x13\\x93\\x9b\\xc4\\xd4\\xbf\\x1b\\x18\\xf9\\x89Y7\\'
"xe5\\xbfoe\\ra$o\\xfb\\xbf\\xa6\\xb2&\\xbe\\xb9\\x8e\\xfd\\xbft\\xdc\\'\\x"
'a3b:\\xfe\\xbf\\xca_\\xaex\\xb4\\xc5\\x07@"\\xd1\\xf1^\\x9fq\\xff?3kXdv'
'D\\xd2?\\xd5\\x92!\\\\P\\xa0\\xee?\\x1b\\x18\\xf9\\x89Y7\\xe5\\xbf\\x9cd\\x'
'bd>\\x0c%\\xf8\\xbf\\xe6m+\\xed\\xc8\\x0b\\xf8\\xbfb\\xe8\\xb3M9\\xf4\\xf'
'6\\xbf\\xca_\\xaex\\xb4\\xc5\\x07@e\\xdc-T\\x1b\\xd1\\x03@\\xb0\\xcbB\\xc'
'0\\xb0\\x85\\xe0?\\xecyi2\\x89\\x0b\\xf4?{\\xef\\xb8\\xebQG\\xdb\\xbf\\x9'
'cd\\xbd>\\x0c%\\xf8\\xbfF\\x10\\xa9UA\\xcd\\xfa\\xbf\\xa0\\xe8\\x7f\\xf0\\'
'x1f\\\\\\xef\\xbfP\\xb1\\xd5\\x9b\\x06\\xcb\\x0b@{b\\xd3\\x81\\xf5\\xc4\\x0'
'1@6q\\xc8#\\x8e^\\xe2?\\xd31pS\\x8e)\\xe5?\\x83]\\xff\\x86\\xe1?\\xc8\\x'
'bf\\xc9cm\\x1c\\xf4\\xda\\xf4\\xbf\\xe6m+\\xed\\xc8\\x0b\\xf8\\xbf\\xeaE\\'
'xdd\\x0c\\xaf\\x82\\xea\\xbf=uv\\x0e\\xcb$\\x0f@\\xeb=\\x97\\xdf\\x11N\\x'
'03@\\xba\\x16N\\x87k7\\xe4?k\\x94\\xf60\\xbew\\xf2?j,\\x19Y]/\\xf2\\xbf'
'\\xf6b\\x1d\\xfa\\xdb\\x90\\xf1\\xbf\\x8c\\reg\\xbf\\x8e\\xef\\xbf3\\xa3:)'
'>\\xa9\\xe5\\xbf\\xd4\\xd4\\x19\\xc3`\\xfc\\xd7?"J=\\xcb\\x8c\\xf5\\t@\\xc'
'6aYN&\\xe9\\xe7?\\xec\\xae\\x83/\\xf3\\xe3\\xf0?\\xc0\\x8f\\xcc%\\x83;\\x'
'a8?F\\xc4\\x9a\\xaf\\x87\\x8d\\xec\\xbf&)0\\x1c\\xd8\\x88\\xf2\\xbf\\xeaE'
'\\xdd\\x0c\\xaf\\x82\\xea\\xbf5\\xc8\\xc17\\xf3=\\x10@\\x0c\\x87\\x0f$\\xd'
'9;\\x00@F\\xda\\x03\\xa8\\xbf\\x92\\xc6\\xbf;\\xe8 \\xd5\\xa5\\x00\\xe7\\x'
'bf\\xb2\\xd2\\xf2\\x8c\\xd1.\\xd2?F\\xc4\\x9a\\xaf\\x87\\x8d\\xec\\xbf\\x8'
'c\\reg\\xbf\\x8e\\xef\\xbf\\xeaE\\xdd\\x0c\\xaf\\x82\\xea\\xbf|\\xcc\\xe2\\'
'xfc!"\\r@{b\\xd3\\x81\\xf5\\xc4\\x01@N\\x19/\\xf0J\\x10\\xda\\xbf:\\xb3\\'
'x06\\xd8;(\\xea\\xbf{\\xef\\xb8\\xebQG\\xdb\\xbf\\xc9cm\\x1c\\xf4\\xda\\x'
'f4\\xbf\\xe6m+\\xed\\xc8\\x0b\\xf8\\xbf\\x07\\x97\\xe2\\xdb\\x80\\x87\\xf4'
'\\xbf\\xdb\\x9b\\r\\x06\\xf0k\\x04@\\xd7\\x0f\\x87\\xf3\\x95\\x08\\xea?\\xf'
'aL\\xb3t r\\xf3\\xbf\\x12f\\xdf\\x80\\x81^\\x02\\xc0w\\xb8\\x15\\x1e\\n\\x'
'cb\\xec\\xbfoe\\ra$o\\xfb\\xbfF\\x10\\xa9UA\\xcd\\xfa\\xbf\\xbd9\\x85\\xb'
'f\\xf1`\\xf9\\xbf\\xfb\\\'\\x98A\\xcep\\xfb?\\xb5\\xb8\\xa5\\x87\\xa9"\\xf2'
"?L\\x10\\x07Q\\'\\xe9\\xf9\\xbf\\x94\\xe1\\x1d\\x88xA\\n\\xc0\\x99|\\'\\xa3"
'5\\xf9\\xf5\\xbf\\xc9cm\\x1c\\xf4\\xda\\xf4\\xbf\\x86\\xcb\\xad\\x84PJ\\xf'
'5\\xbf\\x07\\x97\\xe2\\xdb\\x80\\x87\\xf4\\xbfi\\xfd\\xbd\\xe2xC\\xd8\\xbf'
'}\\xe9\\x87\\x15\\x08A\\xf7?B\\xc5\\xfb\\x89l7\\xf6\\xbf\\x131\\xc5\\x83\\'
'x17\\x86\\x05\\xc0{\\xef\\xb8\\xebQG\\xdb\\xbf\\x9cd\\xbd>\\x0c%\\xf8\\xb'
"f\\xe6m+\\xed\\xc8\\x0b\\xf8\\xbft\\xdc\\'\\xa3b:\\xfe\\xbf\\xec\\xd7l\\x9"
'3+\\x12\\x01@\\xab\\x03<\\x98\\xe1 \\xee?\\xdc\\x00\\x98{\\xbfs\\xff\\xbf'
"\\x94\\x168\\x85\\xe2\\x19\\x07\\xc0\\x99|\\'\\xa35\\xf9\\xf5\\xbf\\xf6b\\x"
'1d\\xfa\\xdb\\x90\\xf1\\xbf\\x8c\\reg\\xbf\\x8e\\xef\\xbf\\xeaE\\xdd\\x0c\\'
'xaf\\x82\\xea\\xbfp\\x0e\\xe2;\\xce\\x88\\xe9\\xbf}\\xe9\\x87\\x15\\x08A\\'
'xf7?\\x94\\x88Ofs\\xae\\xfc\\xbfU\\x1f\\xbd\\x0b\\xf42\\x0e\\xc0\\x1b\\x1'
'8\\xf9\\x89Y7\\xe5\\xbfoe\\ra$o\\xfb\\xbfF\\x10\\xa9UA\\xcd\\xfa\\xbfb\\x'
"e8\\xb3M9\\xf4\\xf6\\xbf\\'C\\xa5\\xa2\\xe9\\xc7\\xfc?\\xc1\\x89\\xe1\\xc5"
'\\xbb\\x14\\xec?\\x0f\\xe3\\xc9\\x02\\x96\\xd5\\xfa\\xbfT\\x89\\xf1\\x05\\x'
'c8\\xe3\\x07\\xc0\\x1b\\x18\\xf9\\x89Y7\\xe5\\xbfoe\\ra$o\\xfb\\xbf\\xa6\\'
'xb2&\\xbe\\xb9\\x8e\\xfd\\xbf\\x07\\x97\\xe2\\xdb\\x80\\x87\\xf4\\xbfT^\\x'
"b2\\x03\\x05\\x1f\\xfe?\\xf0\\x95,!p\\xfc\\xe7?L\\x10\\x07Q\\'\\xe9\\xf9\\"
"xbfU\\x1f\\xbd\\x0b\\xf42\\x0e\\xc0\\x99|\\'\\xa35\\xf9\\xf5\\xbf\\xc9cm\\"
'x1c\\xf4\\xda\\xf4\\xbf\\x86\\xcb\\xad\\x84PJ\\xf5\\xbf3\\xa3:)>\\xa9\\xe'
'5\\xbf\\x19j\\xf2f\\xe6\\x9f\\xdd\\xbfr&\\xb5\\xfe\\x1aG\\xf8?\\x7f\\xf28'
'\\xd8\\xfdJ\\xf5\\xbf\\x15\\xc7\\x90\\x89C\\xd5\\x0b\\xc0{\\xef\\xb8\\xebQ'
'G\\xdb\\xbf\\xc9cm\\x1c\\xf4\\xda\\xf4\\xbf\\x86\\xcb\\xad\\x84PJ\\xf5\\xb'
'f\\xabE\\x11j\\xc8\\x1a\\xf2\\xbfq)\\x94\\xb6}\\x17\\x05@P\\xdd<\\xbaSY\\'
"xfb?t\\xa7-\\x11C\\x99\\xf1\\xbf\\x12f\\xdf\\x80\\x81^\\x02\\xc0\\x99|\\'"
'\\xa35\\xf9\\xf5\\xbf\\xf6b\\x1d\\xfa\\xdb\\x90\\xf1\\xbf&)0\\x1c\\xd8\\x8'
'8\\xf2\\xbf\\xa0\\xe8\\x7f\\xf0\\x1f\\\\\\xef\\xbf\\xb6\\x90\\x89^\\x0b\\xe7'
'\\xd2\\xbf/\\x94\\xc4u\\x8ck\\xfe?M\\x13\\xbf03\\xf6\\xe9\\xbf\\x92KR\\x8'
'2L\\xf2\\x03\\xc0\\x83]\\xff\\x86\\xe1?\\xc8\\xbf\\xa0\\xc2\\xfajW\\xf9\\x'
'e5\\xbf\\x86\\xcb\\xad\\x84PJ\\xf5\\xbf\\x07\\x97\\xe2\\xdb\\x80\\x87\\xf4'
'\\xbf\\xdb\\x9b\\r\\x06\\xf0k\\x04@\\xd0_\\\\p\\xd9\\xab\\xf0\\xbft\\xa7-\\x'
'11C\\x99\\xf1\\xbf!k\\r\\xf9@F\\xfb\\xbf\\x83]\\xff\\x86\\xe1?\\xc8\\xbf\\'
'xc9cm\\x1c\\xf4\\xda\\xf4\\xbf&)0\\x1c\\xd8\\x88\\xf2\\xbf\\xabE\\x11j\\x'
'c8\\x1a\\xf2\\xbf\\x07\\xb7\\x1ag\\x0b\\xc3\\x05@Z\\xa0\\x0f\\xd1@S\\xfa?'
't\\xa7-\\x11C\\x99\\xf1\\xbf\\xa2P\\x80\\xfa\\x0b\\xda\\xfc\\xbf\\xb6\\t\\x'
'96Z\\x19\\xab\\xe0?\\xc2\\xa7\\xf2\\xc3]2\\xf9?{Z,\\x036\\xcd\\xf6?\\x94'
'\\x9fo \\xb0\\x91\\xd8?\\xa1\\xf1}\\x7f\\x97\\xc2\\xf8?J;{^\\xbc\\xf4\\xf'
'c\\xbfB\\xc5\\xfb\\x89l7\\xf6\\xbfR\\xf3%\\x00\\x9c\\x94\\x01\\xc0oJ\\xcf'
'\\x82z\\xd2\\xef?\\xc3V\\x11Z\\x83\\x1c\\x08@\\xfdq\\x91\\xd2\\x8b\\xe9\\x'
'00@\\x16\\xdc\\xf5\\xae\\x88\\x9e\\x01@\\xd4\\xd4\\x19\\xc3`\\xfc\\xd7?\\x'
'14y\\x01\\x8dz\\x03\\x00\\xc07}\\xa8\\xa2\\xbd\\x92\\xe2\\xbf\\xd5\\x92!\\'
'\\P\\xa0\\xee?\\xb6\\t\\x96Z\\x19\\xab\\xe0?\\xf0U\\xc17k\\xd2\\x04@\\xbd\\'
'xb6\\x8c\\xa3|l\\x06@\\x16\\xdc\\xf5\\xae\\x88\\x9e\\x01@\\xc4i<\\x9a\\x0'
"e\\x0f\\xf2?\\xe7l\\xb61\\xc6\\x1b\\x04\\xc0L\\x10\\x07Q\\'\\xe9\\xf9\\xbf"
'\\x94\\x168\\x85\\xe2\\x19\\x07\\xc0\\xc3\\x95\\x92\\x1f\\xc6F\\xfb?\\x00X'
"\\x89\\x8d\\xa7\\x0b\\r@\\xae\\xcc\\xc6\\xa8)P\\r@\\x83!;vjQ\\x0b@\\xfb\\'"
'\\x98A\\xcep\\xfb?U\\xfeMu\\xa9\\xee\\xfb\\xbfN\\x19/\\xf0J\\x10\\xda\\xb'
'f\\xd5\\x98\\xf8\\x0e\\x87L\\xc0\\xbf\\x83]\\xff\\x86\\xe1?\\xc8\\xbfl\\xe'
"d\\xd5\\xe84\\xe6\\xb1\\xbf;\\x9f\\'\\xd4&P\\xfc?\\x83!;vjQ\\x0b@\\x19j\\"
'xf2f\\xe6\\x9f\\xdd\\xbf2(\\x1d\\xaa\\xfe\\xd7\\xe1?\\x8b=D\\x9f\\xb8\\xf'
'c\\xf8\\xbf9\\x1d;\\xd2\\x0f\\xd9\\xe3\\xbf\\x13\\xaa\\xb2\\xee\\xc9>\\xe8'
'?\\xc3V\\x11Z\\x83\\x1c\\x08@-C\\xd0\\x06HJ\\x02@\\x1f\\xd6\\xafY\\x9dA\\'
'x05@\\xf1\\x84I\\xfb)f\\xf3?\\n\\xb6.v\\x8d\\t\\x01\\xc0\\xde\\x03P[\\xcb'
'\\x80\\xef\\xbf\\x9f$\\xe9\\xee\\xb3;\\xf0\\xbf{\\xef\\xb8\\xebQG\\xdb\\xb'
'fF\\xc4\\x9a\\xaf\\x87\\x8d\\xec\\xbf\\xb5\\xe6f\\x93\\x99\\x11\\xed?v\\x1'
'5Iz\\xa0c\\xfe?\\x86ANG\\xceX\\xdd?\\xa2 $b\\xb7\\xaa\\x92\\xbfV\\x0f3z'
'\\xf4\\x08\\x02\\xc0R\\xf3%\\x00\\x9c\\x94\\x01\\xc0\\x1b\\x18\\xf9\\x89Y7'
"\\xe5\\xbf\\xa7~u\\xc3\\xed\\xa1\\xd1\\xbf{Z,\\x036\\xcd\\xf6?z\\'\\x81\\x"
'cbU\\xae\\x07@p\\x0e\\xe2;\\xce\\x88\\xe9\\xbf\\xdf2\\xf9\\xefn<\\xc6?\\x'
'7f\\xf28\\xd8\\xfdJ\\xf5\\xbf8RU\\xcfy\\xb1\\xe0\\xbff\\xf5u\\x8b\\x15\\x'
'b3\\xf3?\\xf0U\\xc17k\\xd2\\x04@\\x1eY\\n\\x0c\\xf5-\\t@q-\\xc7 A\\x0b\\x'
'04@\\xf1\\x84I\\xfb)f\\xf3?\\x8f\\x17\\x98\\x01\\x84\\x86\\x00\\xc0\\xc8m'
'9\\xcdU\\x1d\\xe8\\xbf<~\\xec\\xda\\xd1O\\xed\\xbf\\x95E\\x84\\xd5\\xed|\\'
'xf7?\\xc3V\\x11Z\\x83\\x1c\\x08@~\\xfb\\x87tm\\xef\\x0b@q-\\xc7 A\\x0b\\'
'x04@~0*\\xeex\\x13\\xec?U\\xfeMu\\xa9\\xee\\xfb\\xbf8\\x83\\x18b\\xd5\\x'
'ac\\xd2\\xbf7\\xeb\\xcc\\x13\\x81\\x8e\\xc5?\\x95E\\x84\\xd5\\xed|\\xf7?\\'
'x00X\\x89\\x8d\\xa7\\x0b\\r@\\x8d\\xe5Mo\\xc0\\x0b\\x05@(\\xd0i\\x04\\xb2'
'\\xe4\\x08@\\x82e\\xf3C\\xb9\\xbd\\x01@\\xcd\\xcb\\x03<g?\\xed\\xbf3kXdv'
'D\\xd2?\\xecyi2\\x89\\x0b\\xf4?\\x95E\\x84\\xd5\\xed|\\xf7?\\x00X\\x89\\x'
'8d\\xa7\\x0b\\r@]\\x14\\x0f;\\x04\\xab\\x03@\\xc4\\x84\\xde\\xe7\\xe4\\xd4'
'\\x02@\\x98N/9\\xf3\\xb7\\xf0?wG\\xc6\\xb9p\\xdc\\xf8\\xbf\\xb2\\xd7"?\\x'
'e0\\xb9\\xe0\\xbf;\\xe8 \\xd5\\xa5\\x00\\xe7\\xbf\\x1b\\x18\\xf9\\x89Y7\\x'
'e5\\xbfl\\xed\\xd5\\xe84\\xe6\\xb1\\xbf\\xfdq\\x91\\xd2\\x8b\\xe9\\x00@v\\'
'x15Iz\\xa0c\\xfe?\\x17\\xd8\\xc7y\\x97\\xda\\xe6\\xbf\\x8f\\\\\\x85\\xaf\\x'
'b1\\x97\\xdf?\\x1a.\\xd5\\xc9P\\x87\\xfe\\xbf\\xa0\\xba\\xb4\\xf4\\xdf\\x8'
'a\\xf6\\xbf\\x13\\xaa\\xb2\\xee\\xc9>\\xe8?h\\xa9\\x92\\x08\\x8e\\xc6\\xff'
'?\\x8d\\xe5Mo\\xc0\\x0b\\x05@\\xc4\\x84\\xde\\xe7\\xe4\\xd4\\x02@\\xcd\\xc'
'3\\xf5i\\x0b\\xb7\\xe6?\\n\\xb6.v\\x8d\\t\\x01\\xc0t\\xa7-\\x11C\\x99\\xf1'
'\\xbf\\xa2\\x85\\x9a\\xf7u\\xb2\\xf9\\xbf\\x83]\\xff\\x86\\xe1?\\xc8\\xbfF'
'\\xc4\\x9a\\xaf\\x87\\x8d\\xec\\xbf\\xf5\\xa1k\\xc2\\xa8\\x8e\\xe7?h3\\rv,'
'h\\x00@\\x98N/9\\xf3\\xb7\\xf0?2(\\x1d\\xaa\\xfe\\xd7\\xe1?B\\xc5\\xfb\\x'
'89l7\\xf6\\xbf\\xa0\\xba\\xb4\\xf4\\xdf\\x8a\\xf6\\xbfj,\\x19Y]/\\xf2\\xb'
'f=\\x8bJ\\xd83\\xd7\\xd5?\\xdb\\xfc\\xa9k\\xae\\x8e\\xf9?\\xcc~\\x98\\x92'
'\\xf9w\\x06@\\xbdX\\x18A\\xb9\\xc9\\xf0\\xbf\\x06\\x1c\\xd2NJ\\xf0\\xe5?\\'
'xfaL\\xb3t r\\xf3\\xbf;\\xe8 \\xd5\\xa5\\x00\\xe7\\xbf\\x83]\\xff\\x86\\x'
'e1?\\xc8\\xbf\\xa7~u\\xc3\\xed\\xa1\\xd1\\xbf{Z,\\x036\\xcd\\xf6?\\x16\\x'
'dc\\xf5\\xae\\x88\\x9e\\x01@LM\\x8e\\xaa\\xaf\\xd9\\xfc\\xbf<j? \\xaa\\x0'
"2\\xe3\\xbfL\\x10\\x07Q\\'\\xe9\\xf9\\xbf!\\xa0\\'\\xf6\\xaa\\x1e\\xf8\\xbf"
'\\x13\\xaa\\xb2\\xee\\xc9>\\xe8?\\xef\\xa6\\xa2\\xa1E\\xe8\\xf5?5]p\\xf1\\'
'xb7\\x0b\\xe2?}\\x00\\x98E\\xcd\\xcf\\xe0\\xbfp\\x0e\\xe2;\\xce\\x88\\xe9'
'\\xbf\\x8f\\x17\\x98\\x01\\x84\\x86\\x00\\xc0\\xbc\\x1fv&\\x8f^\\xf4\\xbf;'
'\\xe8 \\xd5\\xa5\\x00\\xe7\\xbfw\\xb8\\x15\\x1e\\n\\xcb\\xec\\xbf\\xe2\\x0f'
'\\x15\\x9e\\xa6]\\xc1?\\x0b\\x84n\\xc5\\xdd\\x88\\xe4\\xbf\\x07\\x97\\xe2\\'
'xdb\\x80\\x87\\xf4\\xbf\\xbe\\xa1\\xad\\xb7`,\\xe4\\xbf\\x8f\\x17\\x98\\x0'
'1\\x84\\x86\\x00\\xc0\\xbc\\x1fv&\\x8f^\\xf4\\xbfox\\x13\\x93\\x9b\\xc4\\x'
'd4\\xbf\\x1b\\x18\\xf9\\x89Y7\\xe5\\xbf\\xf3\\x81\\xb5LN\\xca\\xde\\xbf\\x'
'97~\\xe6\\xe8\\xd9\\x0b\\xde\\xbf(\\xd0i\\x04\\xb2\\xe4\\x08@\\x0eUb\\x8f'
'\\x16\\x8e\\xa6\\xbfz\\xd7\\\\k\\xdc\\x0b\\xb5\\xbf7}\\xa8\\xa2\\xbd\\x92\\x'
'e2\\xbfg~k\\x10R\\xe0\\xb1?\\xc7\\xcc5\\xed\\r\\xc3\\xf9\\xbf\\x8b\\xb3\\x'
'd6R\\xaa\\x01\\x01\\xc0\\xe6m+\\xed\\xc8\\x0b\\xf8\\xbf3\\xa3:)>\\xa9\\xe'
'5\\xbf,xn\\xf8)r\\x02\\xc0\\xc5\\x9c\\x89Y\\xec\\xb1\\xf1\\xbf=\\xaa\\x83'
'\\xac\\x01\\xc2\\xc5?\\xd1f\\x8aP\\xf8\\x01\\xe2?j,\\x19Y]/\\xf2\\xbfBf]'
'\\x83<\\xb9\\xfe\\xbf\\xe3\\xcc\\xcf{\\x91\\xe9\\x02\\xc0L\\xe2\\x07\\xa7Z'
'c\\x06\\xc0\\xb6\\x90\\x89^\\x0b\\xe7\\xd2\\xbfR\\xf0\\xe4M\\x84\\xf6\\xe0'
"\\xbfl\\xb8Z/R/\\xaf\\xbf7\\xeb\\xcc\\x13\\x81\\x8e\\xc5?'\np78\ntp79\nbs"
"S'_dual_coef_'\np80\ng31\nsS'degree'\np81\nI3\nsS'epsilon'\np82\nF0."
"0\nsS'max_iter'\np83\nI-1\nsS'fit_status_'\np84\nI0\nsS'_intercept_"
"'\np85\ng13\n(g14\n(I0\ntp86\ng16\ntp87\nRp88\n(I1\n(I1\ntp89\ng35\nI00\nS"
"'>,\\x9e\\xaf\\x87P(\\xc0'\np90\ntp91\nbsS'intercept_'\np92\ng13\n(g14"
"\n(I0\ntp93\ng16\ntp94\nRp95\n(I1\n(I1\ntp96\ng35\nI00\nS'>,\\x9e\\xaf\\x8"
"7P(\\xc0'\np97\ntp98\nbsS'probB_'\np99\ng13\n(g14\n(I0\ntp100\ng16\ntp1"
"01\nRp102\n(I1\n(I0\ntp103\ng35\nI00\ng48\ntp104\nbsS'cache_size'\np10"
"5\nI200\nsS'gamma'\np106\nF0.0\nsb.")


_CLASSIFIERS = [pickle.loads(_LIMB_STR), pickle.loads(_PREC_STR)]
